Package: sitadela
Type: Package
Title: An R package for the easy provision of simple but complete
        tab-delimited genomic annotation from a variety of sources and
        organisms
Authors@R: c(person(given="Panagiotis", family="Moulos",
    email="moulos@fleming.gr", role=c("aut", "cre")))
Depends: R (>= 4.1.0)
Imports: Biobase, BiocGenerics, biomaRt, Biostrings, Seqinfo,
        GenomicFeatures, GenomicRanges, IRanges, methods, parallel,
        Rsamtools, RSQLite, rtracklayer, S4Vectors, tools, txdbmaker,
        utils
Suggests: GenomeInfoDb, BiocStyle, BSgenome, knitr, rmarkdown, RMySQL,
        RUnit
Description: Provides an interface to build a unified database of
    genomic annotations and their coordinates (gene, transcript
    and exon levels). It is aimed to be used when simple 
    tab-delimited annotations (or simple GRanges objects) are 
    required instead of the more complex annotation Bioconductor 
    packages. Also useful when combinatorial annotation elements
    are reuired, such as RefSeq coordinates with Ensembl biotypes.
    Finally, it can download, construct and handle annotations with
    versioned genes and transcripts (where available, e.g. RefSeq
    and latest Ensembl). This is particularly useful in precision
    medicine applications where the latter must be reported.
License: Artistic-2.0
Encoding: UTF-8
URL: https://github.com/pmoulos/sitadela
biocViews: Software, WorkflowStep, RNASeq, Transcription, Sequencing,
        Transcriptomics, BiomedicalInformatics, FunctionalGenomics,
        SystemsBiology, AlternativeSplicing, DataImport, ChIPSeq
VignetteBuilder: knitr
BugReports: https://github.com/pmoulos/sitadela/issues
Version: 1.19.0
Date: 2025-07-22
Collate: 'argcheck.R' 'fromtxdb.R' 'query-ensembl.R' 'query-ncbi.R'
        'query-refseq.R' 'query-ucsc.R' 'query-utils.R' 'reduceops.R'
        'sitadela.R' 'testfuns.R' 'zzz.R'
git_url: https://git.bioconductor.org/packages/sitadela
git_branch: devel
git_last_commit: 45b2af3
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
NeedsCompilation: no
Packaged: 2025-11-05 01:04:18 UTC; biocbuild
Author: Panagiotis Moulos [aut, cre]
Maintainer: Panagiotis Moulos <moulos@fleming.gr>

import(Biobase)
import(biomaRt)
import(GenomicRanges)
import(methods)
import(parallel)
import(Rsamtools)
import(RSQLite)

importFrom("BiocGenerics",match)
importFrom("BiocGenerics",unlist)
importFrom("BiocGenerics",which)

importFrom("Biostrings",alphabetFrequency)

importFrom("Seqinfo",seqinfo)
importFrom("Seqinfo",Seqinfo)
importFrom("Seqinfo",seqlevels)
importFrom("Seqinfo",seqlengths)
importFrom("Seqinfo","seqlengths<-")

importFrom("txdbmaker",makeTxDbFromGFF)
#importFrom("GenomicFeatures",makeTxDbFromGFF)
importFrom("GenomicFeatures",exons)
importFrom("GenomicFeatures",exonsBy)
importFrom("GenomicFeatures",genes)
importFrom("GenomicFeatures","seqlevels<-")
importFrom("GenomicFeatures",transcripts)
importFrom("GenomicFeatures",threeUTRsByTranscript)

importFrom("IRanges",IRanges)
importFrom("IRanges",slice)

importFrom("S4Vectors",DataFrame)
importFrom("S4Vectors",Rle)
importFrom("S4Vectors",queryHits)
importFrom("S4Vectors",runValue)
importFrom("S4Vectors",subjectHits)

importFrom("rtracklayer",import)

importFrom("tools",R_user_dir)

importFrom("utils",download.file)
importFrom("utils",head)
importFrom("utils","menu")
importFrom("utils",packageVersion)
importFrom("utils",read.delim)
importFrom("utils",read.table)
importFrom("utils",write.table)
importFrom("utils",zip)

export(addAnnotation)
export(addCustomAnnotation)
export(getAnnotation)
export(getDbPath)
export(getInstalledAnnotations)
export(getSeqInfo)
export(getUserAnnotations)
export(importCustomAnnotation)
export(loadAnnotation)
export(removeAnnotation)
export(setDbPath)
export(testCustomBuild)
export(testCustomGtf)
export(testKnownBuild)
export(testEnsembl)
export(testEnsemblSimple)
export(testUcsc)
export(testUcscAll)
export(testUcscUtr)
export(testUcscUtrAll)

SiTaDelA - Simple, flexible and reusable tab-delimited genome annotations
================================================================================

Next Generation Sequencing has introduced a massive need for working with 
integer interval data which correspond to actual chromosomal regions, depicted
in linear representations. As a result, previously under-developed algorithms
for working with such data have tremendously evolved. Maybe the most common
application where genomic intervals are used is overlapping a set of query 
intervals with a set of reference intervals. One typical example is counting
the reads produced e.g. from an RNA-Seq experiment and assigning them to genes
of interest through overlapping their mapped coordinates with those of the genes
over a reference genome. As a result, collections of such reference genomic 
regions for several reference organisms are essential for the quick 
interrogation of the latter.

The generation of genomic coordinate systems are nowadays mainstream. Typical
ways of reference genomic region representations are:

* [BED](https://genome.ucsc.edu/FAQ/FAQformat.html#format1) files, which are
simple tab-delimited files with at least 3 columns including the main reference
sequence name (e.g. a chromosome), its start and its end.
* More complex structured files such as 
[GTF](https://genome.ucsc.edu/FAQ/FAQformat.html#format4) and
[GFF](https://genome.ucsc.edu/FAQ/FAQformat.html#format3) which also contain
structures such as exons, different transcripts anf untranslated regions.

Bioconductor offers great infrastructures for fast genomic interval calculations
which are now very mature, high-level and cover most needs. It also offers
many comprehensive and centrally maintained genomic interval annotation packages
as well as tools to quickly create custom annotation packages, such as
[AnnotationForge](https://bioconductor.org/packages/release/bioc/html/AnnotationForge.html).
These packages, are primarily designed to capture genomic structures (genes,
transcripts, exons etc.) accurately and place them in a genomic interval content
suitable for fast calculations. While this is more than sufficient for many 
users and work out-of-the-box, especially for less experienced R users, they may
miss certain characteristics which may be also useful for many users. Such
additional elements are often required by tools that report e.g. transcript
biotypes (such as those in Ensembl) and do not gather mappings between elements
of the same annotation (e.g. gene, transcript, exon ids) in one place in a more
straightforward manner. More specifically, some elements which are not 
directly achievable with standard Bioconductor annotation packages include:

* Simple tab-delimited (or in GRanges objects) genomic interval annotations
capturing several characteristics of these annotations (biotype, GC content).
* Centralization of simple tab-delimited annotations for many organisms and
several genomic interval types in one package.
* Versioning of these annotations under the same database instead of many,
dispersed packages which may be difficult to track and upgrade, especially when
transitioning between Bioconductor versions.
* Gene and transcript versioning (when available, e.g. in NCBI annotations)
which is essential for applications related to precision medicine and diagnostic
procedures.
* A unified interface to several genomic interval annotation sources.

SiTaDelA (**Si**mple **Ta**b **Del**imited **A**nnotations), through efficient
and extensive usage of Bioconductor facilites offers these additional 
functionalities along with certain levels of automation. More specifically, the
`sitadela` package offers:

* Simple tab-delimited (easily output also as GRanges objects) genomic interval
annotations for several transcription unit types with additional characteristics
(gene GC content, biotypes).
* A centralized annotation building and retrieval system, supporting several
organisms, versions and annotation resources as well as custom user annotations
coming in GTF/GFF format.
* Versioning of the annotation builds to improve reproducibility and tracking.
* A unified interface to several genomic interval annotation sources which
automates database build but also fetches annotations on-the-fly if not already
present in the build.
* Centralized gene and transcript versioning where available (e.g. NCBI),
especially useful for genomics precision medicine appplications and the
respective diagnostic processes.
* Additional portability from Bioconductor to other applications through the
simple database schema adopted.
* Additional attributes such as corrected feature lengths (i.e. corrected
gene lengths based on sum of lengths of coding regions, to be used e.g. for
RNA abundance estimation and normalization).

The `sitadela` annotation database building is extremely simple. The user 
defines a list of desired annotations (organisms, sources, versions) and 
supplies them to the `addAnnotation` function which in turn creates a new or
updates a current database. A custom, non-directly supported organism annotation
can be imported through the `addCustomAnnotation` function and annotations not
needed anymore can be removed with the `removeAnnotation` function. Finally, as
the built can require some time, especially if many organisms and sources are
required for a local database, we maintain pre-built databases which are built
periodically (e.g. upon a new Ensembl release).

# Supported organisms

The following organisms (essentially genome versions) are supported for 
automatic database builds:

* Human (*Homo sapiens*) genome version **hg38** (or **GRCh38**)
* Human (*Homo sapiens*) genome version **hg19** (or **GRCh37**)
* Human (*Homo sapiens*) genome version **hg18**
* Mouse (*Mus musculus*) genome version **mm10** (or **GRCm37**)
* Mouse (*Mus musculus*) genome version **mm9**
* Rat (*Rattus norvegicus*) genome version **rn6**
* Rat (*Rattus norvegicus*) genome version **rn5**
* Fruitfly (*Drosophila melanogaster*) genome version **dm6**
* Fruitfly (*Drosophila melanogaster*) genome version **dm3**
* Zebrafish (*Danio rerio*) genome version **danRer7**
* Zebrafish (*Danio rerio*) genome version **danRer10**
* Zebrafish (*Danio rerio*) genome version **danRer11**
* Chimpanzee (*Pan troglodytes*) genome version **panTro4**
* Chimpanzee (*Pan troglodytes*) genome version **panTro5**
* Pig (*Sus scrofa*) genome version **susScr3**
* Pig (*Sus scrofa*) genome version **susScr11**
* Horse (*Equus cabalus*) genome version **equCab2**
* Arabidopsis (*Arabidobsis thaliana*) genome version **TAIR10**

Please note that if genomic annotations from UCSC, RefSeq or NCBI are required,
the following `BSgenome` packages are required (depending on the organisms to
be installed) in order to calculate GC content for gene annotations. Also note 
that there is no `BSgenome` package for some of the `sitadela` supported
organisms and therefore GC contents will not be available anyway.

* BSgenome.Hsapiens.UCSC.hg18
* BSgenome.Hsapiens.UCSC.hg19
* BSgenome.Hsapiens.UCSC.hg38
* BSgenome.Mmusculus.UCSC.mm9
* BSgenome.Mmusculus.UCSC.mm10
* BSgenome.Rnorvegicus.UCSC.rn5
* BSgenome.Rnorvegicus.UCSC.rn6
* BSgenome.Dmelanogaster.UCSC.dm3
* BSgenome.Dmelanogaster.UCSC.dm6
* BSgenome.Drerio.UCSC.danRer7
* BSgenome.Drerio.UCSC.danRer10

Is is therefore advised to install these `BSgenome` packages in advance.

# Building and using a SiTaDelA local database

## Installation of sitadela

To install the sitadela package, one should start R and enter:

```
if(!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("sitadela")
```

To install the latest (perhaps non-stable) version:

```
library(remotes)
remotes::install_github("pmoulos/sitadela")
```

## Setup the database

By default, the database file will be written in the
`tools::R_user_dir("sitadela","data")` directory and the file is called 
`"annotation.sqlite"`. By default the build function will ask for the path where
the database will be installed. You can specify another prefered destination for 
it using the `db` argument in the function call, but if you do that, you will 
have to supply an argument pointing to the SQLite database file you created to 
every sitadela package function call you perform, or any other function that 
uses sitadela annotations, otherwise, the annotation will be downloaded and
formatted on-the-fly instead of using the local database. Upon loading 
`sitadela`, an option is added to the R environment pointing to the default
`sitadela` annotation database. If you wish to change that location and do not
wish to supply the database to other function calls, you can change the default
location of the annotation to your preferred location with the `setDbPath`
function in the beginning of your script/function that uses the annotation 
database.

In this example, we will build a minimal database comprising only the mouse
*mm9* genome version from Ensembl. The database will be built in a temporary
directory inside the current R session's `tempdir()`.

**Important note**: As the annotation build function makes use of 
[Kent](http://hgdownload.soe.ucsc.edu/admin/exe/) utilities for creating 3'UTR
annotations from RefSeq and UCSC, the latter cannot be built in Windows. 
Therefore it is advised to either build the annotation database in a Linux 
system or use our pre-built databases.

```
library(sitadela)

buildDir <- file.path(tempdir(),"test_anndb")
dir.create(buildDir)

# The location of the custom database
myDb <- file.path(buildDir,"testann.sqlite")

# Since we are using Ensembl, we can also ask for a version
organisms <- list(mm9=67)
sources <- ifelse(.Platform$OS.type=="unix",c("ensembl","refseq"),"ensembl")

# If the example is not running in a multicore system, rc is ignored
addAnnotation(organisms,sources,forceDownload=FALSE,db=myDb,rc=0.5)

## Alternatively
# setDbPath(myDb)
# addAnnotation(organisms,sources,forceDownload=FALSE,rc=0.5)
```

## Use the database

Now, that a small database is in place, let's retrieve some data. Remember that
since the built database is not in the default location, we need to pass the
database file in each data retrieval function. The annotation is retrieved as
a `GRanges` object by default.

```
# Load standard annotation based on gene body coordinates
genes <- loadAnnotation(genome="mm9",refdb="ensembl",type="gene",db=myDb)
genes

# Load standard annotation based on 3' UTR coordinates
utrs <- loadAnnotation(genome="mm9",refdb="ensembl",type="utr",db=myDb)
utrs

# Load summarized exon annotation based used with RNA-Seq analysis
sumEx <- loadAnnotation(genome="mm9",refdb="ensembl",type="exon",
    summarized=TRUE,db=myDb)
sumEx

# Load standard annotation based on gene body coordinates from RefSeq
if (.Platform$OS.type=="unix") {
    refGenes <- loadAnnotation(genome="mm9",refdb="refseq",type="gene",
        db=myDb)
    refGenes
}
```

Or as a data frame if you prefer using `asdf=TRUE`. The data frame however does 
not contain metadata like `Seqinfo` to be used for any susequent validations:

```
# Load standard annotation based on gene body coordinates
genes <- loadAnnotation(genome="mm9",refdb="ensembl",type="gene",db=myDb,
    asdf=TRUE)
head(genes)
```

## Add a custom annotation

Apart from the supported organisms and databases, you can add a custom 
annotation. Such an annotation can be: 

* A non-supported organism (e.g. an insect or another mammal e.g. dog)
* A modification or further curation you have done to existing/supported
annotations
* A supported organism but from a different source
* Any other case where the provided annotations are not adequate

This can be achieved through the usage of
[GTF/GFF](https://www.ensembl.org/info/website/upload/gff.html) files, along 
with some simple metadata that you have to provide for proper import to the
annotation database. This can be achieved through the usage of the
`addCustomAnnotation` function. Details on required metadata can be found
in the function's help page.

**Important note:** Please note that importing a custom genome annotation 
directly from UCSC (UCSC SQL database dumps) is not supported in Windows as the
process involves using the `genePredToGtf` program which is not available for
Windows.

Let's try a couple of examples. The first one uses example GTF files shipped
with the package. These are sample chromosomes from:

* Atlantic cod (*Gadus morhua*), sequence HE567025
* Armadillo (*Dasypus novemcinctus*), sequence JH569334
* European bass (*Dicentrarchus labrax*), chromosome LG3

Below, we test custom building with reference sequence HE567025 of Atlantic cod:

```
gtf <- system.file(package="sitadela","extdata",
    "gadMor1_HE567025.gtf.gz")
chrom <- system.file(package="sitadela","extdata",
    "gadMor1_HE567025.txt.gz")
chromInfo <- read.delim(chrom,header=FALSE,row.names=1)
names(chromInfo) <- "length"
metadata <- list(
    organism="gadMor1_HE567025",
    source="sitadela_package",
    chromInfo=chromInfo
)
tmpdb <- tempfile()

addCustomAnnotation(gtfFile=gtf,metadata=metadata,db=tmpdb)

# Try to retrieve some data
g <- loadAnnotation(genome="gadMor1_HE567025",refdb="sitadela_package",
    type="gene",db=tmpdb)
g

# Delete the temporary database
unlink(tmpdb)
```

The next one is part of a custom annotation for the Ebola virus from UCSC:

```{r example-5, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
gtf <- system.file(package="sitadela","extdata",
    "eboVir3_KM034562v1.gtf.gz")
chrom <- system.file(package="sitadela","extdata",
    "eboVir3_KM034562v1.txt.gz")
chromInfo <- read.delim(chrom,header=FALSE,row.names=1)
names(chromInfo) <- "length"
metadata <- list(
    organism="gadMor1_HE567025",
    source="sitadela_package",
    chromInfo=chromInfo
)
tmpdb <- tempfile()

addCustomAnnotation(gtfFile=gtf,metadata=metadata,db=tmpdb)

# Try to retrieve some data
g <- loadAnnotation(genome="gadMor1_HE567025",refdb="sitadela_package",
    type="gene",db=tmpdb)
g

# Delete the temporary database
unlink(tmpdb)
```

Again, please note that complete annotations from UCSC require the 
`genePredToGtf` tool from the UCSC tools base and runs only on Linux. The tool 
is required only for building 3' UTR annotations from UCSC, RefSeq and NCBI, all
of which are being retrieved from the UCSC databases. The next example (full 
EBOLA virus annotation) demonstrates how this is done in a Unix based machine:

```
# Setup a temporary directory to download files etc.
customDir <- file.path(tempdir(),"test_custom")
dir.create(customDir)

# Convert from GenePred to GTF - Unix/Linux only!
if (.Platform$OS.type == "unix" && !grepl("^darwin",R.version$os)) {
    # Download data from UCSC
    goldenPath="http://hgdownload.cse.ucsc.edu/goldenPath/"
    # Gene annotation dump
    download.file(paste0(goldenPath,"eboVir3/database/ncbiGene.txt.gz"),
        file.path(customDir,"eboVir3_ncbiGene.txt.gz"))
    # Chromosome information
    download.file(paste0(goldenPath,"eboVir3/database/chromInfo.txt.gz"),
        file.path(customDir,"eboVir3_chromInfo.txt.gz"))

    # Prepare the build
    chromInfo <- read.delim(file.path(customDir,"eboVir3_chromInfo.txt.gz"),
        header=FALSE)
    chromInfo <- chromInfo[,1:2]
    rownames(chromInfo) <- as.character(chromInfo[,1])
    chromInfo <- chromInfo[,2,drop=FALSE]
    
    # Coversion from genePred to GTF
    genePredToGtf <- file.path(customDir,"genePredToGtf")
    if (!file.exists(genePredToGtf)) {
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
            genePredToGtf
        )
        system(paste("chmod 775",genePredToGtf))
    }
    gtfFile <- file.path(customDir,"eboVir3.gtf")
    tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgtf",sep="."))
    command <- paste0(
        "zcat ",file.path(customDir,"eboVir3_ncbiGene.txt.gz"),
        " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
        " -source=eboVir3"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
        gtfFile
    )
    system(command)

    # Build with the metadata list filled (you can also provide a version)
    addCustomAnnotation(
        gtfFile=gtfFile,
        metadata=list(
            organism="eboVir3_test",
            source="ucsc_test",
            chromInfo=chromInfo
        ),
        db=myDb
    )

    # Try to retrieve some data
    eboGenes <- loadAnnotation(genome="eboVir3_test",refdb="ucsc_test",
        type="gene",db=myDb)
    eboGenes
}
```

Another example, the full Atlantic cod genome annotation from UCSC. The same 
things apply for the operating system.

```
if (.Platform$OS.type == "unix") {
    # Gene annotation dump
    download.file(paste0(goldenPath,"gadMor1/database/augustusGene.txt.gz"),
        file.path(customDir,"gadMori1_augustusGene.txt.gz"))
    # Chromosome information
    download.file(paste(goldenPath,"gadMor1/database/chromInfo.txt.gz",sep=""),
        file.path(customDir,"gadMori1_chromInfo.txt.gz"))

    # Prepare the build
    chromInfo <- read.delim(file.path(customDir,"gadMori1_chromInfo.txt.gz"),
        header=FALSE)
    chromInfo <- chromInfo[,1:2]
    rownames(chromInfo) <- as.character(chromInfo[,1])
    chromInfo <- chromInfo[,2,drop=FALSE]
    
    # Coversion from genePred to GTF
    genePredToGtf <- file.path(customDir,"genePredToGtf")
    if (!file.exists(genePredToGtf)) {
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
            genePredToGtf
        )
        system(paste("chmod 775",genePredToGtf))
    }
    gtfFile <- file.path(customDir,"gadMori1.gtf")
    tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgtf",sep="."))
    command <- paste0(
        "zcat ",file.path(customDir,"gadMori1_augustusGene.txt.gz"),
        " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
        " -source=gadMori1"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
        gtfFile
    )
    system(command)

    # Build with the metadata list filled (you can also provide a version)
    addCustomAnnotation(
        gtfFile=gtfFile,
        metadata=list(
            organism="gadMor1_test",
            source="ucsc_test",
            chromInfo=chromInfo
        ),
        db=myDb
    )

    # Try to retrieve some data
    gadGenes <- loadAnnotation(genome="gadMor1_test",refdb="ucsc_test",
        type="gene",db=myDb)
    gadGenes
}
```

Another example, Armadillo from Ensembl. This should work irrespectively of 
operating system. We are downloading chromosomal information from UCSC. Again,
a small dataset included in the package is included in this vignette. See the
commented code below for the full annotation case.

```
## Gene annotation dump from Ensembl
#download.file(paste0("ftp://ftp.ensembl.org/pub/release-98/gtf/",
#    "dasypus_novemcinctus/Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#    file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"))
#
# gtfFile <- file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz")
#
## Chromosome information will be provided from the following BAM file
## available from Ensembl. We have noticed that when using Windows as the OS,
## a remote BAM file cannot be opened by scanBamParam, so for this example,
## chromosome length information will not be available when running in Windows.
#chromInfo <- NULL
#if (.Platform$OS.type == "unix")
#    chromInfo <- paste0("ftp://ftp.ensembl.org/pub/release-98/bamcov/",
#        "dasypus_novemcinctus/genebuild/Dasnov3.broad.Ascending_Colon_5.1.bam")

gtfFile <- system.file(package="sitadela","extdata",
    "dasNov3_JH569334.gtf.gz")
chromInfo <- read.delim(system.file(package="sitadela",
    "extdata","dasNov3_JH569334.txt.gz"),header=FALSE)

# Build with the metadata list filled (you can also provide a version)
addCustomAnnotation(
    gtfFile=gtfFile,
    metadata=list(
        organism="dasNov3_test",
        source="ensembl_test",
        chromInfo=chromInfo
    ),
    db=myDb
)

# Try to retrieve some data
dasGenes <- loadAnnotation(genome="dasNov3_test",refdb="ensembl_test",
    type="gene",db=myDb)
dasGenes
```

## A complete build

A quite complete build (with latest versions of Ensembl annotations) would look
like (supposing the default annotation database location):

```
organisms <- list(
    hg18=67,
    hg19=75,
    hg38=101:102,
    mm9=67,
    mm10=101:102,
    rn5=77,
    rn6=101:102,
    dm3=77,
    dm6=101:102,
    danrer7=77,
    danrer10=91,
    danrer11=101:102,
    pantro4=90,
    pantro5=101:102,
    susscr3=89,
    susscr11=101:102,
    equcab2=94,
    equcab3=101:102
)

sources <- c("ensembl","ucsc","refseq","ncbi")

addAnnotation(organisms,sources,forceDownload=FALSE,rc=0.5)
```

The aforementioned complete built can be found
[here](https://drive.google.com/drive/folders/14vIQBL2iNlVtHkhhbjSMwt04-ZuDAuuR?usp=sharing)
Complete builts will become available from time to time (e.g. with every new
Ensembl release) for users who do not wish to create annotation databases on
their own. Root access may be required (depending on the sitadela library
location) to place it in the default location where it can be found 
automatically.

# Annotations on-the-fly

If for some reason you do not want to build and use an annotation database but
you wish to benefit from the sitadela simple formats nonetheless, or even to
work with an organism that does not yet exist in the database, the 
`loadAnnotation` function will perform all required actions (download and create
a `GRanges` object) on-the-fly as long as there is an internet connection. 
However, the aforementioned function does not handle custom annotations in GTF 
files. In that case, you should use the `importCustomAnnotation` function with
a list describing the GTF file, that is:

```{r pseudo-1, eval=TRUE, echo=TRUE, message=TRUE, warning=FALSE}
metadata <- list(
    organism="ORGANISM_NAME",
    source="SOURCE_NAME",
    chromInfo="CHROM_INFO"
)
```

The above argument can be passed to the `importCustomAnnotation` call in the
respective position.

For further details about custom annotations on the fly, please check
`addCustomAnnotation` and `importCustomAnnotation` functions.

Apart from an artificial GTF file called "dummy.gtf.gz" which serves testing
of custom annotation building, there are three more examples of GTF files
accompanying the sitadela package. These are:

1. gadMor1_HE567025.gtf.gz and gadMor1_HE567025.txt.gz
A GTF file with the reference sequence HE567025 of the Atlantic cod (Gadus 
morhua) retrieved from UCSC, along with the length of the sequence, required
for a more complete custom annotation build.

2. dasNov3_JH569334.gtf.gz and dasNov3_JH569334.txt.gz
A GTF file with the reference sequence JH569334 of the Armadillo (Dasypus  
novemcinctus) retrieved from UCSC, along with the length of the sequence, 
required for a more complete custom annotation build.

3. dicLab1_LG3.gtf.gz and dicLab1_LG3.txt.gz
A GTF file with the reference sequence LG3 of the European bass (Dicentrarchus  
labrax) retrieved from Max Planck Institute for Molecular Genetics, Berlin
(http://seabass.mpipz.mpg.de/), along with the length of the sequence, required
for a more complete custom annotation build.

4. eboVir3_ncbiGene.gtf.gz and eboVir3_ncbiGene.txt.gz
A GTF file with the genome sequence of the EBOLA virus retrieved from UCSC, 
along with the length of the sequence, required for a more complete custom 
annotation build.

# ann <- .makeGeneGeneFromTxDb(txdb,map,FALSE)
.makeGeneGeneFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- genes(txdb)
        # We need to add gc_content, gene_name, biotype from the map
        # Remove the exon_id and transcript_id columns from the map
        # for the gene case
        smap <- map[,!names(map) %in% c("exon_id", "transcript_id"),drop=FALSE]
        smap <- smap[!duplicated(smap),]
        
        # Add metadata from the map
        rownames(smap) <- smap$gene_id
        smap <- smap[names(gr),,drop=FALSE]
        gr$gene_name <- smap$gene_name
        gr$biotype <- smap$biotype
    }
    else {
        gr <- genes(txdb,columns=c("gene_id","tx_type"))
        gr$gene_id <- as.character(gr$gene_id)
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(gr$tx_type)
        gr$tx_type <- NULL
    }
    
    gr$gc_content <- rep(50,length(gr))
    ann <- as.data.frame(gr)
    ann <- ann[,c(1,2,3,6,9,5,7,8)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeGeneExonFromTxDb(txdb,map,FALSE)
.makeGeneExonFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- exons(txdb,columns="exon_name")
        
        # There are cases where exons are unnamed and just the structure 
        useMap <- TRUE
        if (any(is.na(gr$exon_name)))
            useMap <- FALSE
        
        if (useMap)
            gr <- .useMapTr(gr,map) 
        else
            # In order to create the exon annotation, we need to manually 
            # overlap and assign to genes
            gr <- .notUseMapEx(gr,txdb)
    }
    else
        gr <- .formatTxDb2Exon(gr,txdb)
    
    ann <- .formatFromGr(gr,"exon_id")
    if (length(unique(ann$exon_id)) == length(ann$exon_id))
        # Should always be TRUE
        rownames(ann) <- as.character(ann$exon_id)
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumGeneExonFromTxDb(txdb,map,FALSE)
.makeSumGeneExonFromTxDb <- function(txdb,map,asdf) {
    ann <- .makeGeneExonFromTxDb(txdb,map,asdf)
    
    annList <- reduceExons(GRanges(ann))
    sexon <- annList$model
    names(sexon) <- as.character(sexon$exon_id)
    activeLength <- annList$length
    names(activeLength) <- unique(sexon$gene_id)
    attr(sexon,"activeLength") <- activeLength
    
    if (asdf) {
        eann <- as.data.frame(sexon)
        eann <- eann[,c(1,2,3,6,7,5,8,9)]
        names(eann)[c(1,4)] <- c("chromosome","exon_id")
        attr(eann,"activeLength") <- activeLength
        return(eann)
    }
    else
        return(sexon)
}

# ann <- .makeGeneUtrFromTxDb(txdb,map,FALSE)
.makeGeneUtrFromTxDb <- function(txdb,map,asdf) {
    utrList <- threeUTRsByTranscript(txdb,use.names=TRUE)
    utrGr <- unlist(utrList)
    
    utrGr <- .checkUTRPossible(utrGr)
    if (length(utrGr) == 0)
        return(.dfOgr(utrGr,asdf))
    
    utrGr$transcript_id <- names(utrGr)
    utrTmp <- as.data.frame(unname(utrGr))
    keep <- c("seqnames","start","end","transcript_id",
        "exon_rank","strand","exon_name")
    utr <- utrTmp[,keep]
    
    useMap <- TRUE
    if (any(is.na(utr$exon_name)) || is.null(map))
        # Impossible to take info from map...
        useMap <- FALSE
    else
        rownames(utr) <- paste(utr$exon_name,utr$transcript_id,sep="_")
        
    if (useMap)
        utr <- .useMapUtrEx(utr,map)
    else
        # Add metadata from the utr frame
        utr <- .notUseMapUtr(utr,utrGr,txdb)
    
    ann <- utr[,c(1,2,3,4,8,6,9,10)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumGeneUtrFromTxDb(txdb,map,FALSE)
.makeSumGeneUtrFromTxDb <- function(txdb,map,asdf) {
    ann <- .makeGeneUtrFromTxDb(txdb,map,FALSE)
    if (length(ann) == 0)
        return(.dfOgr(ann,asdf))
    
    annList <- reduceTranscripts(ann)
    s3utr <- annList$model
    names(s3utr) <- as.character(s3utr$transcript_id)
    activeLength <- annList$length
    names(activeLength) <- unique(as.character(s3utr$gene_id))
    
    if (asdf) {
        sann <- as.data.frame(s3utr)
        sann <- sann[,c(1,2,3,6,8,5,7,9)]
        #if (useMap)
        #    names(sann)[c(1,4)] <- c("chromosome","gene_id")
        #else {
        #    sann <- sann[,c(1,2,3,7,5,6,4,8)]
            names(sann)[c(1,4,7)] <- c("chromosome","gene_id","transcript_id")
        #}
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(s3utr)
}

# ann <- .makeTranscriptGeneFromTxDb(txdb,map,FALSE)
.makeTranscriptGeneFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map))
        gr <- .useMapEx(gr,txdb,map)
    else {
        gr <- transcripts(txdb,columns=c("tx_name","gene_id","tx_type"))
        if (is(gr$gene_id,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
        else
            gtmp <- as.character(gr$gene_id)
        gr$gene_id <- gtmp
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
        gr$tx_type <- NULL
    }
    
    ann <- .formatFromGr(gr,"transcript_id")
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumTranscriptGeneFromTxDb(txdb,map,FALSE)
.makeSumTranscriptGeneFromTxDb <- function(txdb,map,asdf) {
    ann <- .makeTranscriptGeneFromTxDb(txdb,map,asdf)
    
    annList <- reduceTranscripts(GRanges(ann))
    stranscript <- annList$model
    names(stranscript) <- as.character(stranscript$transcript_id)
    activeLength <- annList$length
    names(activeLength) <- unique(stranscript$gene_id)
    
    if (asdf) {
        sann <- as.data.frame(stranscript)
        sann <- sann[,c(1,2,3,6,8,5,7,9)]
        names(sann)[c(1,4)] <- c("chromosome","transcript_id")
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(stranscript)
}

# ann <- .makeTranscriptExonFromTxDb(txdb,map,FALSE)
.makeTranscriptExonFromTxDb <- function(txdb,map,asdf) {
    gr <- exonsBy(txdb,by="tx")
    tr <- transcripts(txdb,columns=c("tx_name","gene_id"))
    if (is(tr$gene_id,"CharacterList"))
        gtmp <- unlist(lapply(tr$gene_id,function(x) x[1]))
    else
        gtmp <- tr$gene_id
    names(gr) <- names(tr) <- tr$tx_name
        
    if (!is.null(map)) {
        grTmp <- as.data.frame(gr)
        grTmp$gene_id <- rep(gtmp,lengths(gr))
        names(grTmp)[2] <- "transcript_id"
        keep <- c("seqnames","start","end","exon_name","transcript_id","strand")
        ann <- grTmp[,keep]
        ann$transcript_id <- as.character(ann$transcript_id)
        rownames(ann) <- paste(ann$exon_name,ann$transcript_id,sep="_")
        
        # There are cases where exons are unnamed and just the structure 
        useMap <- TRUE
        if (any(is.na(gr$exon_name)))
            useMap <- FALSE
        
        if (useMap) {
            rownames(map) <- paste(map$exon_id,map$transcript_id,sep="_")

            # Different case with map here
            smap <- map[intersect(rownames(ann),rownames(map)),]
            ann <- ann[intersect(rownames(ann),rownames(map)),]
            
            # We need to add gene_id, gene_name, biotype from the map
            # Remove the exon_id column from the map for the gene case
            smap <- smap[,!names(map)=="gene_id",drop=FALSE]
            
            # Add metadata from the map
            ann$gene_name <- smap$gene_name
            ann$biotype <- smap$biotype
            
            names(ann)[c(1,4)] <- c("chromosome","exon_id")
            ann$chromosome <- as.character(ann$chromosome)
            ord <- order(ann$chromosome,ann$start)
            ann <- ann[ord,]
            smap <- smap[ord,]
            if (length(unique(ann$exon_id)) == length(ann$exon_id))
                rownames(ann) <- as.character(ann$exon_id)
            else
                rownames(ann) <- rownames(smap)
        }
        else {
            # In order to create the exon annotation, we need to manually 
            # overlap and assign to transcripts
            gr <- unlist(gr)
            ov <- findOverlaps(gr,tr)
            exonPos <- queryHits(ov)
            tranPos <- subjectHits(ov)
            levs <- unique(tranPos)
            
            dup <- which(duplicated(exonPos))
            if (length(dup) > 0) {
                exonPos <- exonPos[-dup]
                tranPos <- tranPos[-dup]
            }
            
            S <- split(exonPos,factor(tranPos,levels=levs))
            gr$transcript_id <- rep(names(tr)[levs],lengths(S))
            gr$gene_name <- rep(tr$gene_id[levs],lengths(S))
            gr$biotype <- rep("transcript",length(gr))
            if (is(gr$gene_name,"CharacterList"))
                gtmp <- unlist(lapply(gr$gene_name,function(x) x[1]))
            else
                gtmp <- gr$gene_name
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gr$transcript_id,sep="")
            gr$exon_name <- names(gr)
            gr$gene_name <- gtmp
            
            
            ann <- .formatFromGrE(gr)
            
            if (length(unique(ann$exon_id)) == length(ann$exon_id))
                # Should always be TRUE
                rownames(ann) <- as.character(ann$exon_id)
        }
    }
    else {
        # The same as above... some repetition, we check later
        gr <- unlist(gr)
        ov <- findOverlaps(gr,tr)
        exonPos <- queryHits(ov)
        tranPos <- subjectHits(ov)
        levs <- unique(tranPos)
        
        dup <- duplicated(exonPos)
        exonPos <- exonPos[!dup]
        tranPos <- tranPos[!dup]
        
        S <- split(exonPos,factor(tranPos,levels=levs))
        gr$transcript_id <- rep(names(tr)[levs],lengths(S))
        gr$gene_name <- rep(tr$gene_id[levs],lengths(S))
        gr$biotype <- rep("transcript",length(gr))
        if (is(gr$gene_name,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_name,function(x) x[1]))
        else
            gtmp <- gr$gene_name
        names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
            gr$transcript_id,sep="")
        gr$exon_name <- names(gr)
        gr$gene_name <- gtmp
        
        ann <- .formatFromGrE(gr)        
        if (length(unique(ann$exon_id)) == length(ann$exon_id))
            # Should always be TRUE
            rownames(ann) <- as.character(ann$exon_id)
    }
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumTranscriptExonFromTxDb(txdb,map,FALSE)
.makeSumTranscriptExonFromTxDb <- function(txdb,map,asdf) {
    # Test code rewrapping
    annGr <- .makeTranscriptExonFromTxDb(txdb,map,asdf=FALSE)
    
    # Do the rest
    annList <- reduceTranscriptsExons(annGr)
    strexon <- annList$model
    names(strexon) <- as.character(strexon$exon_id)
    activeLength <- annList$length
    names(activeLength) <- unique(strexon$transcript_id)
    
    if (asdf) {
        sann <- as.data.frame(strexon)
        sann <- sann[,c(1,2,3,6,7,5,8,9)]
        names(sann)[c(1,4)] <- c("chromosome","exon_id")
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(strexon)   
}

# ann <- .makeTranscriptUtrFromTxDb(txdb,map,FALSE)
.makeTranscriptUtrFromTxDb <- function(txdb,map,asdf) {
    utrList <- threeUTRsByTranscript(txdb,use.names=TRUE)
    utrGr <- unlist(utrList)
    
    utrGr <- .checkUTRPossible(utrGr)
    if (length(utrGr) == 0)
        return(.dfOgr(utrGr,asdf))
    
    utrGr$transcript_id <- names(utrGr)
    utrTmp <- as.data.frame(unname(utrGr))
    keep <- c("seqnames","start","end","transcript_id",
        "exon_rank","strand","exon_name")
    utr <- utrTmp[,keep]
    
    useMap <- TRUE
    if (any(is.na(utr$exon_name)) || is.null(map))
        useMap <- FALSE
    else
        rownames(utr) <- paste(utr$exon_name,utr$transcript_id,sep="_")
    
    if (useMap)
        utr <- .useMapUtrEx(utr,map)
    else
        utr <- .notUseMapUtr(utr,utrGr,txdb)

    ann <- utr[,c(1,2,3,4,8,6,9,10)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumTranscriptUtrFromTxDb(txdb,map,FALSE)
.makeSumTranscriptUtrFromTxDb <- function(txdb,map,asdf) {
    ann <- .makeTranscriptUtrFromTxDb(txdb,map,FALSE)
    if (length(ann) == 0)
        return(.dfOgr(ann,asdf))
    
    #message("  summarizing UTRs per gene for imported GTF")
    annList <- reduceTranscriptsUtr(ann)
    s3utrTranscript <- annList$model
    names(s3utrTranscript) <- 
        as.character(s3utrTranscript$transcript_id)
    activeLength <- annList$length
    # Remember, gene_id in UTR's per transcript is the transcript
    names(activeLength) <- unique(as.character(s3utrTranscript$gene_id))
    
    if (asdf) {
        sann <- as.data.frame(s3utrTranscript)
        sann <- sann[,c(1,2,3,6,8,5,7,9)]
        names(sann)[c(1,4)] <- c("chromosome","transcript_id")
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(s3utrTranscript)
}

# ann <- .makeExonExonFromTxDb(txdb,map,FALSE)
.makeExonExonFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- exons(txdb,columns="exon_name")
        
        # There are cases where exons are unnamed and just the structure 
        useMap <- TRUE
        if (any(is.na(gr$exon_name)))
            useMap <- FALSE
        
        if (useMap) {
            grs <- .useMapTr(gr,map,TRUE) 
            gr <- grs$gr
            smap <- grs$smap
            ann <- .formatFromGr(gr,"exon_id")
            if (length(unique(ann$exon_id)) == length(ann$exon_id))
                rownames(ann) <- as.character(ann$exon_id)
            else {
                ord <- order(ann$chromosome,ann$start)
                smap <- smap[ord,]
                rownames(ann) <- rownames(smap)
            }
        }
        else {
            # In order to create the exon annotation, we need to manually 
            # overlap and assign to genes
            ge <- genes(txdb)
            ov <- findOverlaps(gr,ge)
            exonPos <- queryHits(ov)
            genePos <- subjectHits(ov)
            levs <- unique(genePos)
            
            dup <- which(duplicated(exonPos))
            if (length(dup) > 0) {
                exonPos <- exonPos[-dup]
                genePos <- genePos[-dup]
            }
            
            S <- split(exonPos,factor(genePos,levels=levs))
            gr$gene_id <- gr$gene_name <- rep(names(ge)[levs],lengths(S))
            gr$biotype <- rep("gene",length(gr))
            if (is(gr$gene_id,"CharacterList"))
                gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
            else
                gtmp <- gr$gene_id
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
            gr$exon_name <- names(gr)
            
            ann <- .formatFromGr(ann,"exon_id")
            
            if (length(unique(ann$exon_id)) == length(ann$exon_id))
                # Should always be TRUE
                rownames(ann) <- as.character(ann$exon_id)
        }
    }
    else {
        gr <- .formatTxDb2Exon(gr,txdb)
        ann <- .formatFromGr(gr,"exon_id")
        if (length(unique(ann$exon_id)) == length(ann$exon_id))
            # Should always be TRUE
            rownames(ann) <- as.character(ann$exon_id)
    }
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

.makeIdMap <- function(grdf) {
    # Some basic names must be present in order to make a map
    if (!any(c("exon_id","transcript_id","gene_id") %in% names(grdf))
        || any(is.na(grdf$exon_id))
        || any(is.na(grdf$transcript_id))
        || any(is.na(grdf$gene_id)))
        return(NULL)

    #hasGeneName <- hasBiotype <- FALSE
    if (!all(is.na(grdf$gene_name)) && !all(is.na(grdf$gene_biotype))) {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_name",
            "gene_biotype")]
        names(map)[5] <- "biotype"
        #hasGeneName <- hasBiotype <- TRUE
    }
    else if (all(is.na(grdf$gene_name)) && !all(is.na(grdf$gene_biotype))) {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_id",
            "gene_biotype")]
        names(map)[4:5] <- c("gene_name","biotype")
        #hasBiotype <- TRUE
    }
    else if (!all(is.na(grdf$gene_name)) && all(is.na(grdf$gene_biotype))) {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_name")]
        map$gene_biotype <- rep("gene",nrow(map))
        names(map)[5] <- "biotype"
        #hasGeneName <- TRUE
    }
    else {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_id")]
        map$gene_biotype <- rep("gene",nrow(map))
        names(map)[4:5] <- c("gene_name","biotype")
    }
    nag <- which(is.na(map$gene_name))
    if (length(nag) > 0)
        map$gene_name[nag] <- map$gene_id[nag]
    nab <- which(is.na(map$biotype))
    if (length(nab) > 0)
        map$biotype[nag] <- "gene"
    return(map)
}

.useMapTr <- function(gr,map,.sm=FALSE) {
    names(gr) <- gr$exon_name

    # We need to add gene_id, gene_name, biotype from the map
    # Remove the transcript_id column from the map for the gene
    # case
    smap <- map[,names(map)!="transcript_id",drop=FALSE]
    smap <- smap[!duplicated(smap),,drop=FALSE]
    
    # Add metadata from the map
    rownames(smap) <- smap$exon_id
    smap <- smap[names(gr),,drop=FALSE]
    gr$gene_id <- smap$gene_id
    gr$gene_name <- smap$gene_name
    gr$biotype <- smap$biotype
    
    if (.sm)
        return(list(gr=gr,smap=smap))
    else
        return(gr)
}

.useMapEx <- function(gr,txdb,map) {
    gr <- transcripts(txdb,columns="tx_name")
    names(gr) <- gr$tx_name
    
    # We need to add gene_id, gene_name, biotype from the map
    # Remove the exon_id column from the map for the transcript 
    # case
    smap <- map[,names(map)!="exon_id",drop=FALSE]
    smap <- smap[!duplicated(smap),,drop=FALSE]
    
    # Add metadata from the map
    rownames(smap) <- smap$transcript_id
    smap <- smap[names(gr),,drop=FALSE]
    gr$gene_id <- smap$gene_name
    gr$gene_name <- smap$gene_name
    gr$biotype <- smap$biotype
    
    return(gr)
}

.useMapUtrEx <- function(utr,map) {
    rownames(map) <- paste(map$exon_id,map$transcript_id,sep="_")
    
    # Different case with map here
    smap <- map[rownames(utr),]
    
    # We need to add gene_id, gene_name, biotype from the map
    # Remove the exon_id column from the map for the gene case
    smap <- smap[,names(map)!="exon_id",drop=FALSE]
    
    # Add metadata from the map
    utr$gene_id <- smap$gene_id
    utr$gene_name <- smap$gene_name
    utr$biotype <- smap$biotype
    
    return(utr)
}

.notUseMapEx <- function(gr,txdb) {
    ge <- genes(txdb,columns=c("gene_id","tx_type"))
    ov <- findOverlaps(gr,ge)
    exonPos <- queryHits(ov)
    genePos <- subjectHits(ov)
    levs <- unique(genePos)

    dup <- duplicated(exonPos)
    exonPos <- exonPos[!dup]
    genePos <- genePos[!dup]

    S <- split(exonPos,factor(genePos,levels=levs))
    gr$gene_id <- gr$gene_name <- rep(names(ge)[levs],lengths(S))
    gr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],
        lengths(S))
    if (is(gr$gene_id,"CharacterList"))
        gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
    else
        gtmp <- gr$gene_id
    names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
        gtmp,sep="")
    gr$exon_name <- names(gr)
    
    return(gr)
}

.notUseMapUtr <- function(utr,utrGr,txdb) {
    utr$gene_id <- utr$transcript_id
    utr$gene_name <- utr$gene_id
    
    # To get biotypes
    ge <- genes(txdb,columns=c("gene_id","tx_type"))
    ov <- findOverlaps(utrGr,ge)
    utrPos <- queryHits(ov)
    genePos <- subjectHits(ov)
    levs <- unique(genePos)
    
    dup <- duplicated(utrPos)
    utrPos <- utrPos[!dup]
    genePos <- genePos[!dup]
    
    S <- split(utrPos,factor(genePos,levels=levs))
    utr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],lengths(S))
    
    return(utr)
}

.formatFromGr <- function(gr,what) {
    ann <- as.data.frame(unname(gr))
    ann <- ann[,c(1,2,3,6,7,5,8,9)]
    names(ann)[c(1,4)] <- c("chromosome",what)
    ann$chromosome <- as.character(ann$chromosome)
    return(ann[order(ann$chromosome,ann$start),])
}

.formatFromGrE <- function(gr) {
    ann <- as.data.frame(unname(gr))
    ann <- ann[,c(1,2,3,7,9,5,10,11)]
    names(ann)[c(1,4)] <- c("chromosome","exon_id")
    ann$chromosome <- as.character(ann$chromosome)
    return(ann[order(ann$chromosome,ann$start),])
}

.formatTxDb2Exon <- function(gr,txdb) {
    gr <- exons(txdb,columns=c("exon_name","gene_id","tx_type"))
    if (is(gr$gene_id,"CharacterList"))
        gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
    else
        gtmp <- gr$gene_id
    if (any(is.na(gr$exon_name)))            
        names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
            gtmp,sep="")
    else
        names(gr) <- gr$exon_name
    gr$gene_id <- gtmp
    gr$gene_name <- gr$gene_id
    gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
    gr$tx_type <- NULL
    return(gr)
}

.checkUTRPossible <- function(utrGr) {
    if (length(utrGr) == 0) {
        warning("No UTR information was found in the provided GTF file! Will ",
            "return an empty object...",immediate.=TRUE)
        ann <- data.frame(chromosome=1,start=1,end=1,
            transcript_id=1,gene_id=1,strand=1,gene_name=1,
            biotype=1,row.names=1)
        # Strange but required to be compatible with GRanges
        return(GRanges(ann[-1,]))
    }
    else
        return(utrGr)
}

.dfOgr <- function(ch,asdf) {
    if (is.data.frame(ch) && asdf)
        return(ch)
    else if (is.data.frame(ch) && !asdf)
        return(GRanges(ch))
    else if (is(ch,"GRanges") && asdf)
        return(as.data.frame(ch))
    else if (is(ch,"GRanges") && !asdf)
        return(ch)
}

.GENE_TYPES <- c("gene","pseudogene","transposable_element_gene")
.TX_TYPES <- c("transcript","pseudogenic_transcript","primary_transcript",
    "mRNA","ncRNA","rRNA","snoRNA","snRNA","tRNA","tmRNA","miRNA",
    "miRNA_primary_transcript","RNase_P_RNA","RNase_MRP_RNA","SRP_RNA",
    "misc_RNA","antisense_RNA", "antisense","lnc_RNA","antisense_lncRNA",
    "transcript_region","pseudogenic_tRNA","scRNA","guide_RNA",
    "telomerase_RNA","vault_RNA","Y_RNA")
.EXON_TYPES <- c("exon","pseudogenic_exon","coding_exon",
    "five_prime_coding_exon","interior_coding_exon",
    "three_prime_coding_exon","exon_of_single_exon_gene","interior_exon",
    "noncoding_exon","five_prime_noncoding_exon",
    "three_prime_noncoding_exon")
.CDS_TYPES <- c("CDS","transposable_element_CDS","CDS_predicted",
    "edited_CDS")
.STOP_CODON_TYPES <- "stop_codon"
.GFF_FEATURE_TYPES <- c(.GENE_TYPES,.TX_TYPES,.EXON_TYPES,.CDS_TYPES,
    .STOP_CODON_TYPES)

addAnnotation <- function(organisms,sources,db=getDbPath(),versioned=FALSE,
    forceDownload=TRUE,retries=5,rc=NULL,stopIfNotBS=FALSE) {
    # Firstly, throw a warning if RefSeq, NCBI or UCSC is in sources and the
    # respective BSgenome package (if available) is not installed as there will
    # be no GC content available in gene annotations.
    if (any(sources %in% c("ucsc","refseq","ncbi"))) {
        if (is.list(organisms))
            orgToCheck <- names(organisms)
        else
            orgToCheck <- organisms
        .checkIfBSgenomesInstalled(orgToCheck,stopIfNotBS)
    }
        
    complete <- FALSE
    times <- 1
    pv <- tryCatch(packageVersion("sitadela"),error=function(e) return(""),
        finally="")
    
    message("\n********************************************************")
    message("This is sitadela ",pv," genomic region annotation builder")
    message("********************************************************")
    
    # Check database path
    .checkDbPath(db)
    
    while(!complete && times<=retries) {
        message("\n========================================================")
        message(format(Sys.time(),"%Y-%m-%d %H:%M:%S")," - Try ",times)
        message("========================================================\n")

        buildResult <- .annotationWorker(organisms=organisms,sources=sources,
            db=db,versioned=versioned,forceDownload=forceDownload,rc=rc)

        complete <- buildResult$complete

        if (!complete) {
            message("-------------------------------------------------------")
            message("Failed annotation builds:")
            for (i in seq_along(buildResult$failed))
                message("  -> Organism: ",buildResult$failed[[i]]$org," - ",
                    "Source: ",buildResult$failed[[i]]$refdb," - ",
                    "Version: ",buildResult$failed[[i]]$ver," - ",
                    "Versioned IDs: ",buildResult$failed[[i]]$tv)
            message("Will retry ",retries-times," times...")
            newRun <- .collapseFailures(buildResult$failed)
            organisms <- newRun$organisms
            sources <- newRun$sources
            message("-------------------------------------------------------\n")
        }
        else {
            message("\n-------------------------------------------------------")
            message("Building process complete!")
            message("-------------------------------------------------------\n")
        }

        times <- times + 1
    }
}

.annotationWorker <- function(organisms,sources,db=getDbPath(),versioned=FALSE,
    forceDownload=TRUE,rc=NULL) {
    if (missing(organisms))
        organisms <- getSupportedOrganisms()
    if (missing(sources))
        sources <- getSupportedRefDbs()
    
    orgIsList <- FALSE
    if (!is.list(organisms) && is.character(organisms)
        && "ensembl" %in% sources)
        warning("When ensembl is in the annotation sources to download, it is ",
            "advised to provide organisms as\na named list with names the ",
            "requested organisms and list members the Ensembl versions.\n",
            "Otherwise, the latest Ensembl version for each organism will be ",
            "used.",immediate.=TRUE)
    if (is.list(organisms)) {
        if (is.null(names(organisms)))
            stop("When organisms is a list, it must be named!")
        orgList <- organisms
        organisms <- names(organisms)
        orgIsList <- TRUE
    }
    
    .checkTextArgs("organisms",organisms,getSupportedOrganisms(),multiarg=TRUE)
    .checkTextArgs("sources",sources,getSupportedRefDbs(),multiarg=TRUE)
    
    if (!is.logical(versioned))
        stop("The versioned argument must be TRUE or FALSE")
    if (!is.logical(forceDownload))
        stop("The forceDownload argument must be TRUE or FALSE")
    
    if (!requireNamespace("GenomeInfoDb"))
        stop("R package GenomeInfoDb is required to construct annotation ",
            "stores!")
    
    # Initialize or open the annotation SQLite datatabase
    message("Opening sitadela SQLite database ",db)
    con <- initDatabase(db)
    
    # Main fault (connection or other) marker
    completed <- TRUE
    failures <- 0
    failed <- list()
    
    for (s in sources) {
        for (o in organisms) {
            # Retrieving genome info. We will be inserting the seqinfo to the
            # sqlite multiple times for the sake of simplicity regarding later
            # deletion by foreign key cascade.
            message("Retrieving genome information for ",o," from ",s)
            sf <- getSeqInfo(o)
            
            # Now, we must derive versioning. For Ensembl it's obvious as it has
            # official versions. For UCSC we need to keep a date track. Then 
            # inside metaseqR, if date not provided, it will automatically
            # detect the latest one (as in Ensembl with versions, if not
            # provided).
            if (s == "ensembl") {
                if (orgIsList)
                    vs <- orgList[[o]]
                else {
                    vss <- .getUcscToEnsembl(o)
                    vs <- vss[length(vss)]
                }
                vs <- .validateEnsemblVersions(o,vs)
                if (is.null(vs)) { # Something went wrong... Get latest again
                    vss <- .getUcscToEnsembl(o)
                    vs <- vss[length(vss)]
                }
            }
            else if (s %in% getSupportedUcscDbs())
                vs <- format(Sys.Date(),"%Y%m%d")
                        
            for (v in vs) {
                # Retrieve gene annotations
                if (.annotationExists(con,o,s,v,"gene",versioned) 
                    && !forceDownload)
                    message("Gene annotation for ",o," from ",s," version ",v,
                        " has already been created and will be skipped.\nIf ",
                        "you wish to recreate it choose forceDownload = TRUE.")
                else {
                    message("Retrieving gene annotation for ",o," from ",s,
                        " version ",v)
                    tryCatch({
                        ann <- getAnnotation(o,"gene",refdb=s,ver=v,
                            tv=versioned,rc=rc)
                    
                        # First drop if previously exists
                        nr <- .dropAnnotation(con,o,s,v,"gene",versioned)
                        # Then insert to the contents table so as to get the 
                        # content id to attach in the annotation table
                        nr <- .insertContent(con,o,s,v,"gene",versioned)
                        nid <- .annotationExists(con,o,s,v,"gene",versioned,
                            out="id")
                        # If something happens, the whole procedure will break
                        # anyway
                        # Add content_id
                        ann$content_id <- rep(nid,nrow(ann))
                        sfGene <- sf
                        sfGene$content_id <- rep(nid,nrow(sfGene))
                        # Write genes and seqinfo
                        dbWriteTable(con,"gene",ann,row.names=FALSE,append=TRUE)
                        dbWriteTable(con,"seqinfo",sfGene,row.names=FALSE,
                            append=TRUE)
                    },error=function(e) {
                        message("Possible connection failure! Marking...")
                        message("Caught error: ",e$message)
                        completed <- FALSE
                        failures <- failures + 1
                        failed[[failures]] <- 
                            list(org=o,refdb=s,ver=v,tv=versioned)
                    },finally="")
                }
                
                # Retrieve transcript annotations
                if (.annotationExists(con,o,s,v,"transcript",versioned) 
                    && !forceDownload)
                    message("Transcript annotation for ",o," from ",s,
                        " version ",v," has already been created and will be ",
                        "skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    message("Retrieving transcript annotation for ",o," from ",
                        s," version ",v)
                    tryCatch({
                        ann <- getAnnotation(o,"transcript",refdb=s,ver=v,
                            tv=versioned,rc=rc)
                        nr <- .dropAnnotation(con,o,s,v,"transcript",versioned)
                        nr <- .insertContent(con,o,s,v,"transcript",versioned)
                        nid <- .annotationExists(con,o,s,v,"transcript",
                            versioned,out="id")
                        ann$content_id <- rep(nid,nrow(ann))
                        sfTranscript <- sf
                        sfTranscript$content_id <- rep(nid,nrow(sfTranscript))
                        dbWriteTable(con,"transcript",ann,row.names=FALSE,
                            append=TRUE)
                        dbWriteTable(con,"seqinfo",sfTranscript,row.names=FALSE,
                            append=TRUE)
                    },error=function(e) {
                        message("Possible connection failure! Marking...")
                        message("Caught error: ",e$message)
                        completed <- FALSE
                        failures <- failures + 1
                        failed[[failures]] <- 
                            list(org=o,refdb=s,ver=v,tv=versioned)
                    },finally="")
                }
                
                # Then summarize the transcripts and write again with type 
                # sum_transcript
                if (.annotationExists(con,o,s,v,"summarized_transcript",
                    versioned) && !forceDownload)
                    message("Summarized transcript annotation for ",o," from ",
                        s," version ",v," has already been created and will ",
                        "be skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"transcript",versioned)) 
                        stop("Transcript annotation for ",o," from ",s,
                            " version ",v," is required in order to build ",
                            "predefined merged transcript regions for read ",
                            "counting.\nPlease rerun the addAnnotation ",
                            "function with appropriate parameters.")
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"transcript",
                        versioned)
                    message("Merging transcripts for ",o," from ",s,
                        " version ",v)
                    annList <- reduceTranscripts(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    
                    nr <- .dropAnnotation(con,o,s,v,"summarized_transcript",
                        versioned)
                    nr <- .insertContent(con,o,s,v,"summarized_transcript",
                        versioned)
                    nid <- .annotationExists(con,o,s,v,"summarized_transcript",
                        versioned,out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumTranscript <- sf
                    sfSumTranscript$content_id <- rep(nid,nrow(sfSumTranscript))
                    dbWriteTable(con,"summarized_transcript",ann,
                        row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumTranscript,row.names=FALSE,
                        append=TRUE)
                }
                
                # Retrieve 3' UTR annotations
                if (.annotationExists(con,o,s,v,"utr",versioned) 
                    && !forceDownload)
                    message("3' UTR annotation for ",o," from ",s," version ",
                        v," has already been created and will be skipped.\nIf ",
                        "you wish to recreate it choose forceDownload = TRUE.")
                else {
                    message("Retrieving 3' UTR annotation for ",o," from ",s,
                        " version ",v)
                    tryCatch({
                        ann <- getAnnotation(o,"utr",refdb=s,ver=v,tv=versioned,
                            rc=rc)
                        nr <- .dropAnnotation(con,o,s,v,"utr",versioned)
                        nr <- .insertContent(con,o,s,v,"utr",versioned)
                        nid <- .annotationExists(con,o,s,v,"utr",versioned,
                            out="id")
                        ann$content_id <- rep(nid,nrow(ann))
                        sfUtr <- sf
                        sfUtr$content_id <- rep(nid,nrow(sfUtr))
                        dbWriteTable(con,"utr",ann,row.names=FALSE,
                            append=TRUE)
                        dbWriteTable(con,"seqinfo",sfUtr,row.names=FALSE,
                            append=TRUE)
                    },error=function(e) {
                        message("Possible connection failure! Marking...")
                        message("Caught error: ",e$message)
                        completed <- FALSE
                        failures <- failures + 1
                        failed[[failures]] <- 
                            list(org=o,refdb=s,ver=v,tv=versioned)
                    },finally="")
                }
                
                # Then summarize the 3'utrs per gene and write again with type 
                # sum_utr
                if (.annotationExists(con,o,s,v,"summarized_3utr",versioned)
                    && !forceDownload)
                    message("Summarized 3' UTR annotation for ",o," from ",
                        s," version ",v," has already been created and will ",
                        "be skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"utr",versioned)) 
                        stop("3' UTR annotation for ",o," from ",s," version ",
                            vs," is required in order to build predefined ",
                            "merged 3' UTR regions for read counting.\nPlease ",
                            "rerun the addAnnotation function with ",
                            "appropriate parameters.")
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"utr",versioned)
                    message("Merging gene 3' UTRs for ",o," from ",s,
                        " version ",v)
                    annList <- reduceTranscripts(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,"summarized_3utr",versioned)
                    nr <- .insertContent(con,o,s,v,"summarized_3utr",versioned)
                    nid <- .annotationExists(con,o,s,v,"summarized_3utr",
                        versioned,out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumUtr <- sf
                    sfSumUtr$content_id <- rep(nid,nrow(sfSumUtr))
                    dbWriteTable(con,"summarized_3utr",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumUtr,row.names=FALSE,
                        append=TRUE)
                    
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_utr_length",
                        versioned)
                    nr <- .insertContent(con,o,s,v,"active_utr_length",
                        versioned)
                    nid <- .annotationExists(con,o,s,v,"active_utr_length",
                        versioned,out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_utr_length",active,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the 3'utrs per transcript and write again with 
                # type sum_utr
                if (.annotationExists(con,o,s,v,"summarized_3utr_transcript",
                    versioned) && !forceDownload)
                    message("Summarized 3' UTR annotation per transcript for ",
                        o," from ",s," version ",v," has already been created ",
                        "and will be skipped.\nIf you wish to recreate it ",
                        "choose forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"utr",versioned))
                        stop("3' UTR annotation per transcript for ",o," from ",
                            s," version ",v," is required in order to build ",
                            "predefined merged 3'UTR regions for read ",
                            "counting.\nPlease rerun the ",
                            "addAnnotation function with ",
                            "appropriate parameters.")
                    annGr <- 
                        .loadPrebuiltAnnotation(con,o,s,v,"utr",versioned)
                    message("Merging transcript 3' UTRs for ",o," from ",s,
                        " version ",v)
                    annList <- reduceTranscriptsUtr(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,
                        "summarized_3utr_transcript",versioned)
                    nr <- .insertContent(con,o,s,v,"summarized_3utr_transcript",
                        versioned)
                    nid <- .annotationExists(con,o,s,v,
                        "summarized_3utr_transcript",versioned,out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumUtrTranscript <- sf
                    sfSumUtrTranscript$content_id <- 
                        rep(nid,nrow(sfSumUtrTranscript))
                    dbWriteTable(con,"summarized_3utr_transcript",ann,
                        row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumUtrTranscript,
                        row.names=FALSE,append=TRUE)
                    
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_trans_utr_length",
                        versioned)
                    nr <- .insertContent(con,o,s,v,"active_trans_utr_length",
                        versioned)
                    nid <- .annotationExists(con,o,s,v,
                        "active_trans_utr_length",versioned,out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_trans_utr_length",active,
                        row.names=FALSE,append=TRUE)
                }
                
                # Retrieve exon annotations
                if (.annotationExists(con,o,s,v,"exon",versioned) 
                    && !forceDownload)
                    message("Exon annotation for ",o," from ",s," version ",v,
                        " has already been created and will be skipped.\nIf ",
                        "you wish to recreate it choose forceDownload = TRUE.")
                else {
                    message("Retrieving exon annotation for ",o," from ",s,
                        " version ",v)
                    tryCatch({
                        ann <- getAnnotation(o,"exon",refdb=s,ver=v,
                            tv=versioned,rc=rc)
                        nr <- .dropAnnotation(con,o,s,v,"exon",versioned)
                        nr <- .insertContent(con,o,s,v,"exon",versioned)
                        nid <- .annotationExists(con,o,s,v,"exon",versioned,
                            out="id")
                        ann$content_id <- rep(nid,nrow(ann))
                        sfExon <- sf
                        sfExon$content_id <- rep(nid,nrow(sfExon))
                        dbWriteTable(con,"exon",ann,row.names=FALSE,
                            append=TRUE)
                        dbWriteTable(con,"seqinfo",sfExon,row.names=FALSE,
                            append=TRUE)
                    },error=function(e) {
                        message("Possible connection failure! Marking...")
                        message("Caught error: ",e$message)
                        completed <- FALSE
                        failures <- failures + 1
                        failed[[failures]] <- 
                            list(org=o,refdb=s,ver=v,tv=versioned)
                    },finally="")
                }
                
                # Retrieve extended annotations
                if (.annotationExists(con,o,s,v,"transexon",versioned) 
                    && !forceDownload)
                    message("Extended exon annotation for ",o," from ",s,
                        " version ",v," has already been created and will be ",
                        "skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    message("Retrieving extended exon annotation for ",o,
                        " from ",s," version ",v)
                    tryCatch({
                        ann <- getAnnotation(o,"transexon",refdb=s,ver=v,
                            tv=versioned,rc=rc)
                        nr <- .dropAnnotation(con,o,s,v,"transexon",versioned)
                        nr <- .insertContent(con,o,s,v,"transexon",versioned)
                        nid <- .annotationExists(con,o,s,v,"transexon",
                            versioned,out="id")
                        ann$content_id <- rep(nid,nrow(ann))
                        sfTrExon <- sf
                        sfTrExon$content_id <- rep(nid,nrow(sfExon))
                        dbWriteTable(con,"transexon",ann,row.names=FALSE,
                            append=TRUE)
                        dbWriteTable(con,"seqinfo",sfTrExon,row.names=FALSE,
                            append=TRUE)
                    },error=function(e) {
                        message("Possible connection failure! Marking...")
                        message("Caught error: ",e$message)
                        completed <- FALSE
                        failures <- failures + 1
                        failed[[failures]] <- 
                            list(org=o,refdb=s,ver=v,tv=versioned)
                    },finally="")
                }
                
                # Then summarize the exons per gene and write again with type 
                # summarized_exon
                if (.annotationExists(con,o,s,v,"summarized_exon",versioned)
                    && !forceDownload)
                    message("Summarized exon annotation for ",o," from ",s,
                        " version ",v," has already been created and will be ",
                        "skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"exon",versioned)) 
                        stop("Exon annotation for ",o," from ",s," version ",v,
                            " is required in order to build predefined merged ",
                            "exon regions for RNA-Seq (exon) coverage ",
                            "calculations.\nPlease rerun the ",
                            "addAnnotation function with ",
                            "appropriate parameters.")
                            
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"exon",versioned)
                    message("Merging exons for ",o," from ",s," version ",v)
                    annList <- reduceExons(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,"summarized_exon",versioned)
                    nr <- .insertContent(con,o,s,v,"summarized_exon",versioned)
                    nid <- .annotationExists(con,o,s,v,"summarized_exon",
                        versioned,out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumExon <- sf
                    sfSumExon$content_id <- rep(nid,nrow(sfSumExon))
                    dbWriteTable(con,"summarized_exon",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumExon,row.names=FALSE,
                        append=TRUE)        
                            
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_length",versioned)
                    nr <- .insertContent(con,o,s,v,"active_length",versioned)
                    nid <- .annotationExists(con,o,s,v,"active_length",
                        versioned,out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_length",active,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the exons per transcript and write again with 
                # type summarized_transcript_exon
                if (.annotationExists(con,o,s,v,"summarized_transcript_exon",
                    versioned) && !forceDownload)
                    message("Summarized exon annotation per transcript for ",o,
                        " from ",s," version ",v," has already been created ",
                        "and will be skipped.\nIf you wish to recreate it ",
                        "choose forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"transexon",versioned)) 
                        stop("Extended exon annotation for ",o," from ",s,
                            " version ",v," is required in order to build ",
                            "summarized exon annotation per transcript\n",
                            "Please rerun the addAnnotation ",
                            "function with appropriate parameters.")
                
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"transexon",
                        versioned)
                    message("Merging exons for ",o," from ",s," version ",v)
                    annList <- reduceTranscriptsExons(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,
                        "summarized_transcript_exon",versioned)
                    nr <- .insertContent(con,o,s,v,"summarized_transcript_exon",
                        versioned)
                    nid <- .annotationExists(con,o,s,v,
                        "summarized_transcript_exon",versioned,out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumTrExon <- sf
                    sfSumTrExon$content_id <- rep(nid,nrow(sfSumTrExon))
                    dbWriteTable(con,"summarized_transcript_exon",ann,
                        row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumTrExon,row.names=FALSE,
                        append=TRUE)
                            
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_trans_exon_length",
                        versioned)
                    nr <- .insertContent(con,o,s,v,"active_trans_exon_length",
                        versioned)
                    nid <- .annotationExists(con,o,s,v,
                        "active_trans_exon_length",versioned,out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_trans_exon_length",active,
                        row.names=FALSE,append=TRUE)
                }
            }
        }
    }
    
    dbDisconnect(con)
    
    return(list(
        completed=completed,
        failed=failed
    ))
}

# GTF only!
addCustomAnnotation <- function(gtfFile,metadata,db=getDbPath(),rewrite=TRUE) {
    # Check metadata
    if (is.null(metadata$organism))
        stop("An organism name must be provided with metadata!")
    if (is.null(metadata$source)) {
        warning("A source should be provided with metadata! Using 'inhouse'...",
            immediate.=TRUE)
        metadata$source <- "inhouse"
    }
    if (is.null(metadata$version)) {
        warning("A version should be provided with metadata! Using today...",
            immediate.=TRUE)
        metadata$version <- format(Sys.Date(),"%Y%m%d")
    }
    if (is.null(metadata$chromInfo)) {
        warning("Chromosomal lengths should be provided with metadata! ",
            "Only chromosome names will be available... ",immediate.=TRUE)
        metadata$chromInfo <- NULL
    }
    else {
        str <- metadata$chromInfo
        if (is.character(str)) {
            out <- tryCatch(open(Rsamtools::BamFile(str)),error=function(e) e)
            if (inherits(out,"error")) # Not a BAM file, try to read.delim
                metadata$chromInfo <- read.delim(str,row.names=1)
            else
                metadata$chromInfo <- .chromInfoFromBAM(str)
        }
        if (!is.data.frame(metadata$chromInfo))
            stop("metadata$chromInfo must be a data frame!")
    }
    
    # Check database path
    if (!dir.exists(dirname(db)))
        dir.create(dirname(db),recursive=TRUE,mode="0755")
    
    # Initialize or open the annotation SQLite datatabase
    message("Opening sitadela SQLite database ",db)
    con <- initDatabase(db)
    
    parsed <- parseCustomGtf(gtfFile)
    
    s <- tolower(metadata$source)
    o <- tolower(metadata$organism)
    v <- metadata$version
    
    # Retrieve gene annotations
    if (.annotationExists(con,o,s,v,"gene") && !rewrite)
        message("Gene annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving gene annotation for ",o," from ",s," version ",v,
            " from ",gtfFile)
        ann <- annotationFromCustomGtf(parsed,type="gene",asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"gene")
        nr <- .insertContent(con,o,s,v,"gene",0,1)
        nid <- .annotationExists(con,o,s,v,"gene",FALSE,out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfGene <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfGene$content_id <- rep(nid,nrow(sfGene))
        dbWriteTable(con,"gene",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfGene,row.names=FALSE,append=TRUE)
    }
    
    # Retrieve transcript annotations
    if (.annotationExists(con,o,s,v,"transcript") && !rewrite)
        message("Transcript annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving transcript annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,type="transcript",asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"transcript")
        nr <- .insertContent(con,o,s,v,"transcript",0,1)
        nid <- .annotationExists(con,o,s,v,"transcript",FALSE,out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfTranscript <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfTranscript$content_id <- rep(nid,nrow(sfTranscript))
        dbWriteTable(con,"transcript",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfTranscript,row.names=FALSE,append=TRUE)
    }
    
    # Then summarize the transcripts and write again with type sum_transcript
    if (.annotationExists(con,o,s,v,"summarized_transcript") && !rewrite)
        message("Summarized transcript annotation for ",o," from ",s,
            " version ",v," has already been created and will be skipped.\nIf ",
            "you wish to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized transcript annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,type="transcript",summarized=TRUE,
            asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"summarized_transcript")
        nr <- .insertContent(con,o,s,v,"summarized_transcript",0,1)
        nid <- .annotationExists(con,o,s,v,"summarized_transcript",FALSE,
            out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfSumTranscript <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfSumTranscript$content_id <- rep(nid,nrow(sfSumTranscript))
        dbWriteTable(con,"summarized_transcript",ann,row.names=FALSE,
            append=TRUE)
        dbWriteTable(con,"seqinfo",sfSumTranscript,row.names=FALSE,append=TRUE)
    }
    
    # Retrieve 3' UTR annotations
    if (.annotationExists(con,o,s,v,"utr") && !rewrite)
        message("3' UTR annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving 3' UTR annotation for ",o," from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,type="utr",asdf=TRUE)
        if (nrow(ann) > 0) {
            nr <- .dropAnnotation(con,o,s,v,"utr")
            nr <- .insertContent(con,o,s,v,"utr",0,1)
            nid <- .annotationExists(con,o,s,v,"utr",FALSE,out="id")
            ann$content_id <- rep(nid,nrow(ann))
            sfUtr <- .chromInfoToSeqInfoDf(metadata$chromInfo)
            sfUtr$content_id <- rep(nid,nrow(sfUtr))
            dbWriteTable(con,"utr",ann,row.names=FALSE,append=TRUE)
            dbWriteTable(con,"seqinfo",sfUtr,row.names=FALSE,append=TRUE)
        }
        else
            message("3' UTR annotation for ",o," from ",s," version ",v," is ",
                "not available in the provided GTF file.")
    }
    
    # Then summarize the 3'UTRs and write again with type sum_transcript
    if (.annotationExists(con,o,s,v,"summarized_3utr") && !rewrite)
        message("Summarized 3' UTR annotation for ",o," from ",s," version ",v,
            " has already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized 3' UTR annotation per gene for ",o,
            " from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,type="utr",summarized=TRUE,
            asdf=TRUE)
        
        if (nrow(ann) > 0) {
            activeLength <- attr(ann,"activeLength")
            
            nr <- .dropAnnotation(con,o,s,v,"summarized_3utr")
            nr <- .insertContent(con,o,s,v,"summarized_3utr",0,1)
            nid <- .annotationExists(con,o,s,v,"summarized_3utr",FALSE,out="id")
            ann$content_id <- rep(nid,nrow(ann))
            sfSumUtr <- .chromInfoToSeqInfoDf(metadata$chromInfo)
            sfSumUtr$content_id <- rep(nid,nrow(sfSumUtr))
            dbWriteTable(con,"summarized_3utr",ann,row.names=FALSE,append=TRUE)
            dbWriteTable(con,"seqinfo",sfSumUtr,row.names=FALSE,append=TRUE)
            
            nr <- .dropAnnotation(con,o,s,v,"active_utr_length")
            nr <- .insertContent(con,o,s,v,"active_utr_length",0,1)
            nid <- .annotationExists(con,o,s,v,"active_utr_length",FALSE,
                out="id")
            active <- data.frame(
                name=names(activeLength),
                length=activeLength,
                content_id=rep(nid,length(activeLength))
            )
            dbWriteTable(con,"active_utr_length",active,row.names=FALSE,
                append=TRUE)
        }
        else
            message("3' UTR annotation for ",o," from ",s," version ",v," is ",
                "not available in the provided GTF file.")
    }
    
    # Then summarize the 3'utrs per transcript and write again with 
    # type sum_utr
    if (.annotationExists(con,o,s,v,"summarized_3utr_transcript") && !rewrite)
        message("Summarized 3' UTR annotation per transcript for ",o," from ",s,
            " version ",v," has already been created and will be skipped.\nIf ",
            "you wish to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized 3' UTR annotation per transcript for ",o,
            " from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,type="transutr",summarized=TRUE,
            asdf=TRUE)

        if (nrow(ann) > 0) {
            activeLength <- attr(ann,"activeLength")
            nr <- .dropAnnotation(con,o,s,v,"summarized_3utr_transcript")
            nr <- .insertContent(con,o,s,v,"summarized_3utr_transcript",0,1)
            nid <- .annotationExists(con,o,s,v,"summarized_3utr_transcript",
                FALSE,out="id")
            ann$content_id <- rep(nid,nrow(ann))
            sfSumUtrTranscript <- .chromInfoToSeqInfoDf(metadata$chromInfo)
            sfSumUtrTranscript$content_id <- rep(nid,nrow(sfSumUtrTranscript))
            dbWriteTable(con,"summarized_3utr_transcript",ann,row.names=FALSE,
                append=TRUE)
            dbWriteTable(con,"seqinfo",sfSumUtrTranscript,row.names=FALSE,
                append=TRUE)
            
            nr <- .dropAnnotation(con,o,s,v,"active_trans_utr_length")
            nr <- .insertContent(con,o,s,v,"active_trans_utr_length",0,1)
            nid <- .annotationExists(con,o,s,v,"active_trans_utr_length",
                FALSE,out="id")
            active <- data.frame(
                name=names(activeLength),
                length=activeLength,
                content_id=rep(nid,length(activeLength))
            )
            dbWriteTable(con,"active_trans_utr_length",active,row.names=FALSE,
                append=TRUE)
        }
        else
            message("3' UTR annotation for ",o," from ",s," version ",v," is ",
                "not available in the provided GTF file.")
    }
    
    # Retrieve exon annotations
    if (.annotationExists(con,o,s,v,"exon") && !rewrite)
        message("Exon annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving exon annotation for ",o," from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,type="exon",asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"exon")
        nr <- .insertContent(con,o,s,v,"exon",0,1)
        nid <- .annotationExists(con,o,s,v,"exon",FALSE,out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfExon <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfExon$content_id <- rep(nid,nrow(sfExon))
        dbWriteTable(con,"exon",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfExon,row.names=FALSE,append=TRUE)
    }
    
    # Then summarize the exons and write again with type sum_exon
    if (.annotationExists(con,o,s,v,"summarized_exon") && !rewrite)
        message("Summarized exon annotation for ",o," from ",s," version ",v,
            " has already been created and will be skipped.\nIf you wish to ",
            "to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized exon annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,type="exon",summarized=TRUE,
            asdf=TRUE)
        activeLength <- attr(ann,"activeLength")
        
        nr <- .dropAnnotation(con,o,s,v,"summarized_exon")
        nr <- .insertContent(con,o,s,v,"summarized_exon",0,1)
        nid <- .annotationExists(con,o,s,v,"summarized_exon",FALSE,out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfSumExon <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfSumExon$content_id <- rep(nid,nrow(sfSumExon))
        dbWriteTable(con,"summarized_exon",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfSumExon,row.names=FALSE,append=TRUE)

        nr <- .dropAnnotation(con,o,s,v,"active_length")
        nr <- .insertContent(con,o,s,v,"active_length",0,1)
        nid <- .annotationExists(con,o,s,v,"active_length",FALSE,out="id")
        active <- data.frame(
            name=names(activeLength),
            length=activeLength,
            content_id=rep(nid,length(activeLength))
        )
        dbWriteTable(con,"active_length",active,row.names=FALSE,append=TRUE)
    }
    
    if (.annotationExists(con,o,s,v,"transexon") && !rewrite)
        message("Extended exon annotation for ",o," from ",s," version ",v,
            " has already been created and will be skipped.\nIf you wish to ",
            "to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving extended exon annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,type="transexon",asdf=TRUE)
        
        nr <- .dropAnnotation(con,o,s,v,"transexon")
        nr <- .insertContent(con,o,s,v,"transexon",0,1)
        nid <- .annotationExists(con,o,s,v,"transexon",FALSE,out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfTransExon <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfTransExon$content_id <- rep(nid,nrow(sfExon))
        dbWriteTable(con,"transexon",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfTransExon,row.names=FALSE,append=TRUE)
    }
    
    # Then summarize the exons per transcript and write again with type 
    # summarized_transcript_exon
    if (.annotationExists(con,o,s,v,"summarized_transcript_exon")
        && !rewrite)
        message("Summarized exon annotation per transcript for ",o," from ",
            s," version ",v," has already been created and will be skipped.\n",
            "If you wish to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized transcript exon annotation for ",o,
            " from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,type="transexon",summarized=TRUE,
            asdf=TRUE)
        activeLength <- attr(ann,"activeLength")
        
        nr <- .dropAnnotation(con,o,s,v,"summarized_transcript_exon")
        nr <- .insertContent(con,o,s,v,"summarized_transcript_exon",0,1)
        nid <- .annotationExists(con,o,s,v,"summarized_transcript_exon",
            FALSE,out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfSumTransExon <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfSumTransExon$content_id <- rep(nid,nrow(sfSumTransExon))
        dbWriteTable(con,"summarized_transcript_exon",ann,row.names=FALSE,
            append=TRUE)
        dbWriteTable(con,"seqinfo",sfSumTransExon,row.names=FALSE,append=TRUE)

        nr <- .dropAnnotation(con,o,s,v,"active_trans_exon_length")
        nr <- .insertContent(con,o,s,v,"active_trans_exon_length",0,1)
        nid <- .annotationExists(con,o,s,v,"active_trans_exon_length",
            FALSE,out="id")
        active <- data.frame(
            name=names(activeLength),
            length=activeLength,
            content_id=rep(nid,length(activeLength))
        )
        dbWriteTable(con,"active_trans_exon_length",active,row.names=FALSE,
            append=TRUE)    
    }
}

# level and type must be re-organized to one single argument, as it is not
# dependent anymore to metaseqR terminology. Suggested:
# gene       : GRanges with genes from start to end
# transcript : GRanges with (summarized) transcripts from start to end
# utr        : GRanges with (summarized) 3' UTRs grouped per gene
# transexon  : GRanges with (summarized) exons grouped per transcript
# transutr   : GRanges with (summarized) 3' UTRs grouped per transcript
# exon       : GRanges with (summarized) exons
loadAnnotation <- function(genome,refdb,type=c("gene","transcript","utr",
    "transutr","transexon","exon"),version="auto",wtv=FALSE,
    db=getDbPath(),summarized=FALSE,asdf=FALSE,rc=NULL) {
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to load annotation from database!")
    
    genome <- tolower(genome[1])
    refdb <- tolower(refdb[1])
    type <- type[1]
    .checkTextArgs("type",type,c("gene","transcript","utr","transexon",
        "transutr","exon"),multiarg=FALSE)
    if (version != "auto")
        .checkNumArgs("version",version,"numeric")
    
    # Check if local storage has been set
    onTheFly <- FALSE
    if (file.exists(db)) {
        # Open the connection
        drv <- dbDriver("SQLite")
        con <- dbConnect(drv,dbname=db)
        
        # Is the general resource (organism, source) installed?
        if (!.annotationExists(con,genome,refdb)) {
            warning("The requested annotation does not seem to exist in the ",
                "database! It will be loaded on the fly.\nConsider importing ",
                "it by using addAnnotation.",immediate.=TRUE)
            onTheFly <- TRUE
        }
        
        # Have we asked for transcript versions? Does the database contain 
        # any versioned transcript annotation?
        if (wtv && !.containsVersionedGT(con,genome,refdb)) {
            warning("A versioned gene/transcript annotation was requested, ",
                "however it does not seem to exist in the database!\n",
                "It will be loaded on the fly. Consider importing it by ",
                "using addAnnotation.",immediate.=TRUE)
            onTheFly = TRUE
        }
        
        # Have we asked for non versioned transcripts in a database that 
        # contains ONLY versioned transcripts?
        if (!wtv && .containsVersionedGT(con,genome,refdb,TRUE)) {
            warning("Non versioned gene/transcript annotation was requested, ",
                "however it does not seem to exist in the database!\n",
                "It will be loaded on the fly. Consider importing it by ",
                "using addAnnotation.",immediate.=TRUE)
            onTheFly = TRUE
        }
        
        # If main source exists, decide on version
        if (!onTheFly) {
            if (version != "auto") {
                if (!.annotationExists(con,genome,refdb,version,
                    .annotationTypeFromInputArgs(type))) {
                    warning("The requested annotation version does not seem ",
                        "to exist! Have you run addAnnotation or possibly ",
                        "mispelled? Will use newest existing version.",
                        immediate.=TRUE)
                    version <- "auto"
                }
            }
            if (version == "auto") {
                # Check if annotation exists, has been performed before
                vers <- .installedVersions(con,genome,refdb)
                vers <- sort(vers,decreasing=TRUE)
                version <- vers[1]
            }
            ann <- .loadPrebuiltAnnotation(con,genome,refdb,version,type,wtv,
                summarized)
            dbDisconnect(con)
            
            if (asdf) {
                a <- attr(ann,"activeLength")
                ann <- as.data.frame(unname(ann))
                ann <- ann[,c(1,2,3,6,7,5,8,9)]
                names(ann)[1] <- "chromosome"
                if (!is.null(a))
                    attr(ann,"activeLength") <- a
                return(ann)
            }
            else
                return(ann)
        }
    }
    else
        onTheFly <- TRUE
        
    if (onTheFly) {
        if (genome %in% getSupportedOrganisms()
            && refdb %in% getSupportedRefDbs()) {
            message("Getting latest annotation on the fly for ",genome," from ",
                refdb)
            ann <- .loadAnnotationOnTheFly(genome,refdb,type,wtv,rc)
            if (asdf) {
                a <- attr(ann,"activeLength")
                ann <- as.data.frame(unname(ann))
                ann <- ann[,c(1,2,3,6,7,5,8,9)]
                names(ann)[1] <- "chromosome"
                if (!is.null(a))
                    attr(ann,"activeLength") <- a
                return(ann)
            }
            else
                return(ann)
        }
        else {
            stop("genome and refdb not in supported automatically download ",
                "annotation options. Please use importCustomAnnotation.")
        }
    }
}

importCustomAnnotation <- function(gtfFile,metadata,
    type=c("gene","transcript","utr","transexon","transutr","exon")) {
    type <- tolower(type[1])
    # Check metadata
    if (is.null(metadata$organism)) {
        tmpOrg <- paste("species",format(Sys.Date(),"%Y%m%d"),sep="_")
        warning("An organism name must be provided with metadata for ",
            "reporting purposes! Using ",tmpOrg,immediate.=TRUE)
        metadata$organism <- tmpOrg
    }

    if (is.null(metadata$source)) {
        tmpSource <- paste("source",format(Sys.Date(),"%Y%m%d"),sep="_")
        warning("A source should be provided with metadata for reporting ",
            "purposes ! Using ",tmpSource,immediate.=TRUE)
        metadata$source <- tmpSource
    }
    if (is.null(metadata$version)) {
        tmpVer <- paste("version",format(Sys.Date(),"%Y%m%d"),sep="_")
        warning("A version should be provided with metadata for reporting ",
            "purposes! Using ",tmpVer,immediate.=TRUE)
        metadata$version <- tmpVer
    }
    if (is.null(metadata$chromInfo)) {
        warning("Chromosomal lengths should be provided with metadata! ",
            "Only chromosome names will be available... ",immediate.=TRUE)
        metadata$chromInfo <- NULL
    }
    else {
        str <- metadata$chromInfo
        if (is.character(str) && file.exists(str)) {
            out <- tryCatch(open(Rsamtools::BamFile(str)),error=function(e) e)
            if (inherits(out,"error")) # Not a BAM file, try to read.delim
                metadata$chromInfo <- read.delim(str,row.names=1)
            else
                metadata$chromInfo <- .chromInfoFromBAM(str)
        }
        if (!is.data.frame(metadata$chromInfo))
            stop("metadata$chromInfo must be a data frame!")
    }
    
    # For display meta information
    s <- metadata$source
    o <- metadata$organism
    v <- metadata$version
    
    # Parse the GTF file... If something wrong, it will crash here
    parsed <- parseCustomGtf(gtfFile)
    
    switch(type,
        gene = {
            message("Retrieving gene annotation for ",o," from ",s," version ",
                v," from ",gtfFile)
            annGr <- annotationFromCustomGtf(parsed,type="gene")
        },
        transcript = {
            message("Retrieving transcript annotation for ",o," from ",s,
                " version ",v," from ",gtfFile)
            annGr <- annotationFromCustomGtf(parsed,type="transcript")
        },
        utr = {
            message("Retrieving summarized 3' UTR annotation per gene ","for ",
                o," from ",s," version ",v," from ",gtfFile)
            annGr <- annotationFromCustomGtf(parsed,type="utr",summarized=TRUE)
        },
        transutr = {
            message("Retrieving summarized 3' UTR annotation per ",
                "transcript for ",o," from ",s," version ",v," from ",
                gtfFile)
            annGr <- annotationFromCustomGtf(parsed,type="transutr",
                summarized=TRUE)
        },
        transexon = {
            message("Retrieving summarized exon annotation per transcript ",
                "for ",o," from ",s," version ",v," from ",gtfFile)
            annGr <- annotationFromCustomGtf(parsed,type="transexon",
                summarized=TRUE)
        },
        exon = {
            message("Retrieving exon annotation for ",o," from ",s," version ",
                v," from ",gtfFile)
            annGr <- annotationFromCustomGtf(parsed,type="exon",summarized=TRUE)
        }
    )
    
    return(annGr)
}

removeAnnotation <- function(org,refdb,ver=NULL,db=NULL) {
    if (missing(db) || is.null(db))
        con <- initDatabase(getDbPath())
    else {
        if (!is.character(db) || !file.exists(db))
            stop("The path to the sitadela database must be a valid path ",
                "to an existing file! Assuming default...")
    }
    con <- initDatabase(db)
    content <- getInstalledAnnotations(con)
    orgs <- unique(as.character(content$organism))
    sources <- unique(as.character(content$source))
    versions <- unique(content$version)
    
    if (!(org %in% orgs)) {
        message(org," organism was not found in the database! Doing nothing...")
        return()
    }
    if (!(refdb %in% sources)) {
        message(refdb," source was not found in the database! Doing nothing...")
        return()
    }
    if (!is.null(ver) && !(ver %in% versions)) {
        message(ver," version was not found in the database! Doing nothing...")
        return()
    }
    
    # Reopen as getInstalledAnnotations closes the connection
    con <- initDatabase(db)
    nr <- .dropAnnotation(con,org,refdb,ver)
    dbDisconnect(con)
    return(nr)
}

getInstalledAnnotations <- function(obj=NULL) {
    con <- .validateDbCon(obj)
    content <- .browseContent(con)
    dbDisconnect(con)
    return(content[,-1])
}

getUserAnnotations <- function(obj=NULL) {
    con <- .validateDbCon(obj)
    userContent <- .browseUserContent(con)
    dbDisconnect(con)
    return(userContent[,-1])
}

getAnnotation <- function(org,type,refdb="ensembl",ver=NULL,tv=FALSE,rc=NULL) {
    org <- tolower(org)
    switch(refdb,
        ensembl = { return(getEnsemblAnnotation(org,type,ver,tv)) },
        ucsc = { return(getUcscAnnotation(org,type,refdb,tv,rc=rc)) },
        refseq = { return(getUcscAnnotation(org,type,refdb,tv,rc=rc)) },
        ncbi = { return(getUcscAnnotation(org,type,refdb,tv,rc=rc)) }
    )
}

getEnsemblAnnotation <- function(org,type,ver=NULL,tv=FALSE) {
    if (org=="tair10")
        dat <- "plants_mart"
    else
        #dat <- "ENSEMBL_MART_ENSEMBL"
        dat <- "genes"
    host <- .getHost(org,ver)
    message("Using Ensembl host ",host)
    #mart <- useMart(biomart=dat,host=host,dataset=.getDataset(org))
    mart <- useEnsembl(biomart=dat,host=host,dataset=.getDataset(org))

    chrsExp <- paste("^",.getValidChrs(org),"$",sep="",collapse="|")
    bm <- tryCatch(
        getBM(attributes=.getAttributes(org,type,ver,tv),mart=mart),
        error=function(e) {
            message("Caught error: ",e)
            message("This error is most probably related to biomaRt request ",
                "timeouts! Using fallback solution...")
            #https://github.com/grimbough/biomaRt/issues/22
            .bypassTimeoutByFilters(org,type,ver,tv,mart)
            #.myGetBM(attributes=.getAttributes(org,type,ver,tv),mart=mart)
        },
        finally=""
    )
    
    if (type=="gene") {
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$start_position,
            end=bm$end_position,
            gene_id=bm$ensembl_gene_id,
            gc_content=if (org %in% 
                c("hg18","hg19","mm9","rn5","dm3","danrer7")) 
                bm$percentage_gc_content else bm$percentage_gene_gc_content,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) bm$external_gene_id 
                else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        rownames(ann) <- ann$gene_id
    }
    else if (type=="transcript") {
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$transcript_start,
            end=bm$transcript_end,
            transcript_id=bm$ensembl_transcript_id,
            gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        rownames(ann) <- as.character(ann$transcript_id)
    }
    else if (type=="utr") {
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$`3_utr_start`,
            end=bm$`3_utr_end`,
            tstart=bm$transcript_start,
            tend=bm$transcript_end,
            transcript_id=bm$ensembl_transcript_id,
            gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9","tair10")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        ann <- correctTranscripts(ann)
        ann <- ann[,c("chromosome","start","end","transcript_id","gene_id",
            "strand","gene_name","biotype")]
    }
    else if (type=="exon") {
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$exon_chrom_start,
            end=bm$exon_chrom_end,
            exon_id=bm$ensembl_exon_id,
            gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        # At some point we should number the exons returned here, but it is not
        # easy as there is heavy overlap... We are doing this at the 
        # summarization level...
        rownames(ann) <- ann$exon_id
    }
    else if (type=="transexon") {
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$exon_chrom_start,
            end=bm$exon_chrom_end,
            exon_id=bm$ensembl_exon_id,
            transcript_id=bm$ensembl_transcript_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
    }
    ann <- ann[order(ann$chromosome,ann$start),]
    ann <- ann[grep(chrsExp,ann$chromosome),]
    ann$chromosome <- as.character(ann$chromosome)
    
    return(ann)
}

getUcscAnnotation <- function(org,type,refdb="ucsc",versioned=FALSE,
    chunkSize=500,rc=NULL) {
    if (!requireNamespace("RMySQL")) {
        rmysqlPresent <- FALSE
        warning("R package RMySQL is not present! Annotation will be ",
            "retrieved by downloading temporary files from UCSC and the usage
            of a temporary SQLite database...",immediate.=TRUE)
    }
    else
        rmysqlPresent <- TRUE
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to use annotation from UCSC!")
    
    if (org=="tair10") {
        warning("Arabidopsis thaliana genome is not supported by UCSC Genome ",
            "Browser database! Switching to Ensembl...",immediate.=TRUE)
        return(getEnsemblAnnotation("tair10",type))
    }
    
    validChrs <- .getValidChrs(org)
    chrsExp <- paste("^",paste(validChrs,collapse="$|^"),"$",sep="")

    dbOrg <- getUcscOrganism(org)
    if (rmysqlPresent) {
        # The UTR case is handled later
        if (type != "utr") {
            dbCreds <- .getUcscCredentials()
            drv <- dbDriver("MySQL")
            con <- dbConnect(drv,user=dbCreds[2],password=NULL,dbname=dbOrg,
                host=dbCreds[1])
            type2 <- ifelse(type=="transexon","exon",type)
            query <- .getUcscQuery(org,type2,refdb,versioned)
            rawAnn <- dbGetQuery(con,query)
            dbDisconnect(con)
        }
    }
    else {
        tmpSqlite <- getUcscDbl(org,refdb,versioned)
        
        if (type != "utr") { # Otherwise direct download is used
            drv <- dbDriver("SQLite")
            con <- dbConnect(drv,dbname=tmpSqlite)
            query <- .getUcscQuery(org,type,refdb,versioned)
            rawAnn <- dbGetQuery(con,query)
            dbDisconnect(con)
        }
    }
    if (type=="gene") {
        ann <- rawAnn
        ann <- ann[grep(chrsExp,ann$chromosome,perl=TRUE),]
        ann$chromosome <- as.character(ann$chromosome)
        rownames(ann) <- ann$gene_id
        gcContent <- getGcContent(ann,org)
        ann$gc_content <- gcContent
        rownames(ann) <- ann$gene_id
    }
    else if (type=="transcript") {
        ann <- rawAnn
        ann <- ann[grep(chrsExp,ann$chromosome,perl=TRUE),]
        d <- which(duplicated(ann))
        if (length(d) > 0)
            ann <- ann[-d,]
        ann$chromosome <- as.character(ann$chromosome)
        ann$transcript_id <- as.character(ann$transcript_id)
        ann$gene_id <- as.character(ann$transcript_id)
        # There are still duplicated transcript ids
        d <- which(duplicated(ann$transcript_id))
        iter <- 1
        while(length(d) > 0) {
            ann$transcript_id[d] <- paste(ann$transcript_id[d],iter,sep="_")
            iter <- iter + 1
            d <- which(duplicated(ann$transcript_id))
        }
        rownames(ann) <- ann$transcript_id
        ann <- ann[,c(1,2,3,4,8,5:7)]
    }
    else if (type=="exon" || type=="transexon") {
        rawAnn <- rawAnn[grep(chrsExp,rawAnn$chromosome,perl=TRUE),]
        exList <- cmclapply(as.list(seq_len(nrow(rawAnn))),function(x,d) {
            r <- d[x,,drop=FALSE]
            starts <- as.numeric(strsplit(r[,"start"],",")[[1]])
            ends <- as.numeric(strsplit(r[,"end"],",")[[1]])
            nexons <- length(starts)
            exonNumbering <- seq_len(nexons)
            if (r[,"strand"] == "-")
                exonNumbering <- rev(exonNumbering)
            ret <- data.frame(
                rep(r[,"chromosome"],nexons),
                starts,ends,
                paste(r[,"exon_id"],"_e",exonNumbering,sep=""),
                rep(r[,"strand"],nexons),
                rep(r[,"gene_id"],nexons),
                rep(r[,"gene_name"],nexons),
                rep(r[,"biotype"],nexons)
            )
            names(ret) <- names(r)
            rownames(ret) <- ret$exon_id
            return(ret)
        },rawAnn,rc=rc)
        
        # For some reason rbind takes ages for large datasets... We have to 
        # split in chunks of 1000
        N <- length(exList)
        mo <- N%%chunkSize
        if (mo == 0) {
            fl <- N/chunkSize
            fac <- factor(rep(seq_len(fl),each=chunkSize))
        }
        else {
            fl <- (N-mo)/chunkSize
            fac <- factor(c(rep(seq_len(fl),each=chunkSize),rep(fl,mo)))
        }
        exChunkedList <- split(exList,fac)
        # Merge the chunks
        tmp <- cmclapply(names(exChunkedList),function(n,X) {
            message("Binding chunk ",n,"...")
            return(do.call("rbind",X[[n]]))
        },exChunkedList,rc=rc)
        # Final merge
        message("Binding all chunks...")
        tmpAnn <- do.call("rbind",tmp)
        
        ann <- data.frame(
            chromosome=as.character(tmpAnn$chromosome),
            start=tmpAnn$start,
            end=tmpAnn$end,
            exon_id=as.character(tmpAnn$exon_id),
            gene_id=as.character(tmpAnn$gene_id),
            strand=as.character(tmpAnn$strand),
            gene_name=as.character(tmpAnn$gene_name),
            biotype=as.character(tmpAnn$biotype)
        )
        rownames(ann) <- ann$exon_id
        
        if (type == "transexon")
            names(ann)[5] <- "transcript_id"
    }
    else if (type=="utr") {
        # We are already supposed to have the necessary elements to construct
        # the required data frame. Should be something like
        # 1. Read the GTF as TxDb
        # 2. Import the GTF with rtracklayer
        # 3. Call the 3UTR function of TxDb on (1)
        # 4. Construct a map to add gene_name and other things (essentially
        #    the biotype if we make it, if not we have to connect it from the
        #    respective gene annotation during the pipeline execution)
        # 5. Add the mapped elements to the GRanges/data.frame
        # 6. Return a data.frame
        #
        # All the process is streamlined in getUcscUtr function
        
        preAnn <- getUcscUtr(org,refdb,versioned,.rmysql=rmysqlPresent)
        preAnn <- as.data.frame(preAnn)
        preAnn <- preAnn[,c(1,2,3,6,7,5,8,9)]
        preAnn <- preAnn[grep(chrsExp,preAnn$seqnames,perl=TRUE),]
        names(preAnn) <- c("chromosome","start","end","transcript_id","gene_id",
            "strand","gene_name","biotype")
        # preAnn now has exon names as rownames... OK...
        #rownames(preAnn) <- paste("T",seq_len(nrow(preAnn)),sep="_")
        ann <- preAnn
    }
    
    ann <- ann[order(ann$chromosome,ann$start),]
    return(ann)
}

getUcscUtr <- function(org,refdb="ucsc",versioned=FALSE,.rmysql=FALSE) {
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    .checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    .checkTextArgs("refdb",refdb,getSupportedUcscDbs())
    
    if (!requireNamespace("GenomicFeatures"))
        stop("Bioconductor package GenomicFeatures is required!")
        
    # If RefSeq and versioned, the we must either:
    # 1) Run an SQL query, fetch the results and write the genePred file so it
    #    can be parsed by genePredToGtf
    # 2) Download the refGene and gbCdnaInfo tables locally, create a local
    #    SQLite, run a concat query, create the genePred and parse...
    # The function should create the table and write .gz in session tmpdir
    if (refdb=="refseq" && versioned)
        tableUtr <- .makeUcscRefseqUtrTable(org,.rmysql)
    else {
        httpBase <- paste("http://hgdownload.soe.ucsc.edu/goldenPath/",
        getUcscOrganism(org),"/database/",sep="")
        tableUtr <- getUcscTableNameUtr(org,refdb) # Need one table
        message("  Retrieving table ",tableUtr," for 3'UTR annotation ",
            "generation from ",refdb," for ",org)
        download.file(paste(httpBase,paste(tableUtr,".txt.gz",sep=""),sep=""),
            file.path(tempdir(),paste(tableUtr,".txt.gz",sep="")),quiet=TRUE)
    }
    
    # Do the conversion stuff. AS there is no easy way to check if genePredToGtf
    # exists in the system, we should download it on the fly (once for the 
    # session). If no Linux machine, then problem.
    genePredToGtf <- file.path(tempdir(),"genePredToGtf")
    if (!file.exists(file.path(tempdir(),"genePredToGtf"))) {
        message("  Retrieving genePredToGtf tool")
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
            genePredToGtf,quiet=TRUE
        )
        system(paste("chmod 775",genePredToGtf))
    }
    
    # Then do the conversion... No need for windows case as Kent tools do not
    # work in Windows anyway
    gtfFile <- file.path(tempdir(),paste(tableUtr,".gtf",sep=""))
    message("  Converting ",tableUtr," to GTF")
    tmpName <- file.path(tempdir(),paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgtf",sep="."))
    commandUcsc <- paste(
        "zcat ",file.path(tempdir(),paste(tableUtr,".txt.gz",sep=""))," | ",
        "cut -f1-10 - | ",genePredToGtf," file stdin ",tmpName," -source=",
        tableUtr," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",gtfFile,
        sep=""
    )
    commandRefSeq <- paste(
        "zcat ",file.path(tempdir(),paste(tableUtr,".txt.gz",sep=""))," | ",
        "cut -f2- | ",genePredToGtf," file stdin ",tmpName," -source=",
        tableUtr," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",gtfFile,
        sep=""
    )
    command <- commandRefSeq
    # There is an exception for organisms that do not exist in UCSC databases
    # so we must use RefSeq
    ucscUnsup <- c("rn5","rn6","dm3","dm6","danrer7","danrer10","danrer11",
        "pantro4","pantro5","susscr3","susscr11","equcab2")
    if (refdb == "ucsc" && !(org %in% ucscUnsup))
        command <- commandUcsc
    
    # Run the command and process the data
    message("Executing: ",command)
    system(command)
    parsed <- parseCustomGtf(gtfFile)
    return(.makeGeneUtrFromTxDb(parsed$txdb,parsed$map,asdf=FALSE))
}

getGcContent <- function(ann,org) {
    if (missing(ann))
        stop("A valid annotation data frame must be provided in order to ",
            "retrieve GC-content.")
    org <- tolower(org[1])
    .checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    # Convert annotation to GRanges
    message("Converting annotation to GenomicRanges object...")
    annGr <- tryCatch(GRanges(ann),error=function(e) {
        if (packageVersion("GenomicRanges")<1.14)
            return(GRanges(
                seqnames=Rle(ann[,1]),
                ranges=IRanges(start=ann[,2],end=ann[,3]),
                strand=Rle(ann[,6]),
                name=as.character(ann[,4])
            ))
        else
            return(makeGRangesFromDataFrame(
                df=ann,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            ))
    },finaly="")
    
    bsg <- loadBsGenome(org)
    if (is(bsg,"BSgenome")) {
        message("Getting DNA sequences...")
        seqs <- tryCatch(getSeq(bsg,names=annGr),error=function(e) {
            warning("Caught error ",e,immediate.=TRUE)
            message("Cannot get ",org," sequences! Returning NA GC content...")
            return(NA)
        },finally="")
        if (!is.na(seqs[1])) {
            message("Getting GC content...")
            freqMatrix <- alphabetFrequency(seqs,as.prob=TRUE,baseOnly=TRUE)
            gcContent <- apply(freqMatrix,1,function(x) round(100*sum(x[2:3]),
                digits=2))
        }
        else
            gcContent <- rep(NA,nrow(ann))
    }
    else
        gcContent <- rep(NA,nrow(ann))
    names(gcContent) <- as.character(ann[,4])
    return(gcContent)
}

getSeqInfo <- function(org,asSeqinfo=FALSE) {
    sf <- tryCatch(.chromInfoWrapperGID(getUcscOrganism(org)),
        error=function(e) {
            message("GenomeInfoDb chrom info fetch mechanism failed with the ",
                "following error: ")
            message(e)
            message("")
            message("Trying a direct download...")
            getChromInfo(getUcscOrganism(org))
        },finally="")
    rownames(sf) <- as.character(sf[,1])
    sf <- sf[.getValidChrs(org),]
    if (asSeqinfo)
        return(Seqinfo(seqnames=sf[,1],seqlengths=sf[,2],
            isCircular=sf[,"circular"],genome=getUcscOrganism(org)))
    else
        return(data.frame(chromosome=sf[,1],length=as.integer(sf[,2])))
}

getUcscOrganism <- function(org) {
    switch(org,
        hg18 = { return("hg18") },
        hg19 = { return("hg19") },
        hg38 = { return("hg38") },
        mm9 = { return("mm9") },
        mm10 = { return("mm10") },
        rn5 = { return("rn5") },
        rn6 = { return("rn6") },
        dm3 = { return("dm3") },
        dm6 = { return("dm6") },
        danrer7 = { return("danRer7") },
        danrer10 = { return("danRer10") },
        danrer11 = { return("danRer11") },
        pantro4 = { return("panTro4") },
        pantro5 = { return("panTro5") },
        susscr3 = { return("susScr3") },
        susscr11 = { return("susScr11") },
        equcab2 = { return("equCab2") },
        equcab3 = { return("equCab3") },
        tair10 = { return("TAIR10") }
    )
}

getBsOrganism <- function(org,.warn=TRUE) {
    switch(org,
        hg18 = {
            return("BSgenome.Hsapiens.UCSC.hg18")
        },
        hg19 = {
            return("BSgenome.Hsapiens.UCSC.hg19")
        },
        hg38 = {
            return("BSgenome.Hsapiens.UCSC.hg38")
        },
        mm9 = {
            return("BSgenome.Mmusculus.UCSC.mm9")
        },
        mm10 = {
            return("BSgenome.Mmusculus.UCSC.mm10")
        },
        rn5 = {
            return("BSgenome.Rnorvegicus.UCSC.rn5")
        },
        rn6 = {
            return("BSgenome.Rnorvegicus.UCSC.rn6")
        },
        dm3 = {
            return("BSgenome.Dmelanogaster.UCSC.dm3")
        },
        dm6 = {
            return("BSgenome.Dmelanogaster.UCSC.dm6")
        },
        danrer7 = {
            return("BSgenome.Drerio.UCSC.danRer7")
        },
        danrer10 = {
            return("BSgenome.Drerio.UCSC.danRer10")
        },
        danrer11 = {
            if (.warn)
                warning("danRer11 is not supported by BSgenome! Please use ",
                "Ensembl as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
            #return("BSgenome.Drerio.UCSC.danRer11")
        },
        pantro4 = {
            if (.warn)
                warning("panTro4 is not supported by BSgenome! Please use ",
                    "Ensembl as annotation source if GC content is important.",
                    immediate.=TRUE)
            return(NA)
        },
        pantro5 = {
            return("BSgenome.Ptroglodytes.UCSC.panTro5")
        },
        susscr3 = {
            return("BSgenome.Sscrofa.UCSC.susScr3")
        },
        susscr11 = {
            if (.warn)
                warning("susScr11 is not supported by BSgenome! Please use ",
                    "Ensembl as annotation source if GC content is important.",
                    immediate.=TRUE)
            return(NA)
        },
        equcab2 = {
            if (.warn)
                warning("equCab2 is not supported by BSgenome! Please use ",
                    "Ensembl as annotation source if GC content is important.",
                    immediate.=TRUE)
            return(NA)
        },
        equcab3 = {
            if (.warn)
                warning("equCab3 is not supported by BSgenome! Please use ",
                    "Ensembl as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
        },
        tair10 = {
            if (.warn)
                warning("TAIR10 is not supported by BSgenome! Please use ",
                    "Ensembl as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
        }
    )
}

loadBsGenome <- function(org) {
    if (!requireNamespace("BSgenome"))
        stop("The Bioconductor package BSgenome is required to proceed!")
    bsOrg <- getBsOrganism(org)
    if (!is.na(bsOrg)) {
        if (bsOrg %in% BSgenome::installed.genomes())
            bsObj <- BSgenome::getBSgenome(getUcscOrganism(org))
        else {
            warning(bsOrg," BSgenome package is not installed! Please install ",
                "it\nin order to have gene\nGC content available... Will ",
                "continue with gc_content = NA for ",org)
            bsObj <- NA
        }
        return(bsObj)
    }
    else
        return(NA)
}

getChromInfo <- function(org,
    goldenPath="http://hgdownload.cse.ucsc.edu/goldenPath/") {
    download.file(paste(goldenPath,org,"/database/chromInfo.txt.gz",sep=""),
        file.path(tempdir(),"chromInfo.txt.gz"),quiet=TRUE)
    chromInfo <- read.delim(file.path(tempdir(),"chromInfo.txt.gz"),
        header=FALSE)
    chromInfo <- chromInfo[,1,2]
    chromInfo[,1] <- as.character(chromInfo[,1])
    chromInfo$V3 <- rep(FALSE,nrow(chromInfo))
    m <- grep("M",chromInfo[,1])
    if (length(m) > 0)
        chromInfo$V3[m] <- TRUE
    return(chromInfo)
}

getSupportedRefDbs <- function() {
    return(c("ensembl","ucsc","refseq","ncbi"))
}

getSupportedOrganisms <- function() {
    return(c("hg18","hg19","hg38","mm9","mm10","rn5","rn6","dm3","dm6",
        "danrer7","danrer10","danrer11","pantro4","pantro5",#"pantro6",
        "susscr3","susscr11","equcab2","equcab3","tair10"))
}

getSupportedUcscDbs <- function() {
    return(c("ucsc","refseq","ncbi"))
}

setDbPath <- function(db=NULL) {
    if (is.null(db))
        db <- .defaultDbPath()
    else {
        if (!is.character(db)) {
            warning("The path to the sitadela database must be a valid path ",
                "to a file which\nwill be created if not existing! Assuming ",
                "default...",immediate.=TRUE)
            db <- .defaultDbPath()
        }
    }
    options("sitadela_dbpath"=db)
}

getDbPath <- function() {
    db <- getOption("sitadela_dbpath")
    if (!is.null(db) && is.character(db)) {
        if (file.exists(db))
            return(db)
        else
            return(.defaultDbPath())
    }
    else
        return(.defaultDbPath())
}

importCustomGtf <- function(gtfFile,type=c("gene","transcript","utr",
    "transexon","transutr","exon")) {
    if (!requireNamespace("GenomicFeatures"))
        stop("Bioconductor package GenomicFeatures is required to build ",
            "custom annotation!")
        
    # Some argument checking
    type <- type[1]
    .checkTextArgs("type",type,c("gene","transcript","utr","transexon",
        "transutr","exon"),multiarg=FALSE)
    
    # Import the GTF with rtracklayer to create a map of available metadata
    message("  Importing GTF ",gtfFile," as GTF to make id map")
    desiredColumns <- c("type","gene_id","transcript_id","exon_id",
        "gene_name","gene_biotype")
    gr <- import(gtfFile,format="gtf",colnames=desiredColumns,
        feature.type=.GFF_FEATURE_TYPES)
    grdf <- as.data.frame(gr)
    grdf <- grdf[grdf$type=="exon",]

    # Need to map gene ids to names and biotypes. We need a collapsed 
    # structure exon_id - transcript_id - gene_id - gane_name - biotype
    message("  Making id map")
    map <- .makeIdMap(grdf)
    
    message("  Importing GTF ",gtfFile," as TxDb")
    txdb <- suppressWarnings(makeTxDbFromGFF(gtfFile))
    
    switch(type,
        gene = {
            return(.makeGeneGeneFromTxDb(txdb,map))
        },
        transcript = {
            return(.makeTranscriptGeneFromTxDb(txdb,map))
        },
        utr = {
            return(.makeGeneUtrFromTxDb(txdb,map))
        },
        transexon = {
            # Stub
        },
        transutr = {
            return(.makeTranscriptUtrFromTxDb(txdb,map))
        },
        exon = {
            return(.makeExonExonFromTxDb(txdb,map))
        }
    )
}

# parsed <- parseCustomGtf(gffFile)
parseCustomGtf <- function(gtfFile) {
    if (!requireNamespace("GenomicFeatures"))
        stop("Bioconductor package GenomicFeatures is required to parse ",
            "custom GTF file!")
    
    # Import the GTF with rtracklayer to create a map of available metadata
    message("  Importing GTF ",gtfFile," as GTF to make id map")
    desiredColumns <- c("type","gene_id","transcript_id","exon_id",
        "gene_name","gene_biotype")
    # Let it recognize automatically GFF/GTF
    gr <- import(gtfFile,colnames=desiredColumns,
        feature.type=.GFF_FEATURE_TYPES)
    grdf <- as.data.frame(gr)
    grdf <- grdf[grdf$type=="exon",]

    # Need to map gene ids to names and biotypes. We need a collapsed 
    # structure exon_id - transcript_id - gene_id - gane_name - biotype
    message("  Making id map")
    map <- .makeIdMap(grdf)
    
    message("  Importing GTF ",gtfFile," as TxDb")
    txdb <- suppressWarnings(makeTxDbFromGFF(gtfFile))
    
    return(list(txdb=txdb,map=map))
}

annotationFromCustomGtf <- function(parsed,type=c("gene","transcript","utr",
    "transexon","transutr","exon"),summarized=FALSE,asdf=FALSE) {
    # Some argument checking
    if (!is.logical(summarized))
        stop("summarized must be TRUE or FALSE")
    type <- type[1]
    .checkTextArgs("type",type,c("gene","transcript","utr","transexon",
        "transutr","exon"),multiarg=FALSE)
    
    txdb <- parsed$txdb
    map <- parsed$map
    
    switch(type,
        gene = {
            return(.makeGeneGeneFromTxDb(txdb,map,asdf))
        },
        transcript = {
            if (summarized)
                return(.makeSumTranscriptGeneFromTxDb(txdb,map,asdf))
            else
                return(.makeTranscriptGeneFromTxDb(txdb,map,asdf))
        },
        utr = {
            if (summarized)
                return(.makeSumGeneUtrFromTxDb(txdb,map,asdf))
            else
                return(.makeGeneUtrFromTxDb(txdb,map,asdf))
        },
        transexon = {
            if (summarized)
                return(.makeSumTranscriptExonFromTxDb(txdb,map,asdf))
            else
                return(.makeTranscriptExonFromTxDb(txdb,map,asdf))
        },
        transutr = {
            if (summarized)
                return(.makeSumTranscriptUtrFromTxDb(txdb,map,asdf))
            else
                return(.makeTranscriptUtrFromTxDb(txdb,map,asdf))
        },
        exon = {
            if (summarized)
                return(.makeSumGeneExonFromTxDb(txdb,map,asdf))
            else
                return(.makeExonExonFromTxDb(txdb,map,asdf))
        }
    )
}

cmclapply <- function(...,rc) {
    if (suppressWarnings(!requireNamespace("parallel")) 
        || .Platform$OS.type!="unix")
        m <- FALSE
    else {
        m <- TRUE
        ncores <- parallel::detectCores()
        if (ncores==1) 
            m <- FALSE
        else {
            if (!missing(rc) && !is.null(rc))
                ncores <- ceiling(rc*ncores)
            else 
                m <- FALSE
        }
    }
    if (m)
        return(mclapply(...,mc.cores=ncores,mc.set.seed=FALSE))
    else
        return(lapply(...))
}

.loadPrebuiltAnnotation <- function(con,genome,refdb,version,type,tv,
    summarized=FALSE) {
    metaType <- .annotationTypeFromInputArgs(type,summarized)
    cid <- .annotationExists(con,genome,refdb,version,metaType,tv,out="id")
    if (metaType == "summarized_exon")
        tName <- "active_length"
    else if (metaType == "summarized_3utr")
        tName <- "active_utr_length"
    else if (metaType == "summarized_3utr_transcript")
        tName <- "active_trans_utr_length"
    else if (metaType == "summarized_transcript_exon")
        tName <- "active_trans_exon_length"
    cida <- .annotationExists(con,genome,refdb,version,"active_length",tv,
        out="id")
        
    querySet <- .makeAnnotationQuerySet(metaType,cid,cida)
    
    preAnn <- dbGetQuery(con,querySet$main)
    preAnn$`_id` <- NULL
    preAnn$content_id <- NULL
    ann <- GRanges(preAnn)
    seqlevels(ann) <- unique(preAnn$chromosome)
    
    preSf <- dbGetQuery(con,querySet$seqinfo)
    preSf$`_id` <- NULL
    preSf$content_id <- NULL
    rownames(preSf) <- as.character(preSf[,1])
    if (genome %in% getSupportedOrganisms())
        sfg <- getUcscOrganism(genome)
    else
        sfg <- genome
    sf <- Seqinfo(seqnames=preSf[,1],seqlengths=preSf[,2],
        isCircular=rep(FALSE,nrow(preSf)),genome=sfg)
    
    if (length(sf) > 0) {
        if (length(seqlevels(ann)) != length(seqlevels(sf)))
            # If a subset, this is enough
            seqinfo(ann) <- sf[intersect(seqlevels(ann),seqlevels(sf))]
        else if (!all(seqlevels(ann) == seqlevels(sf))) {
            # Must also be sorted in the same way
            seqlevels(ann) <- seqlevels(sf)
            seqinfo(ann) <- sf
        }
        else
            seqinfo(ann) <- sf
    }
    
    ann <- .nameAnnotationFromMetaType(ann,metaType)
    
    preActive <- NULL
    if (!is.null(querySet$active)) {
        preActive <- dbGetQuery(con,querySet$active)
        preActive$`_id` <- NULL
        preActive$content_id <- NULL
        active <- as.integer(preActive$length)
        names(active) <- as.character(preActive$name)
        # FIXME: Will have to take care later of this
        #active <- active[names(ann)]
        attr(ann,"activeLength") <- active
    }
    
    return(ann)
}

.loadAnnotationOnTheFly <- function(genome,refdb,type,versioned=FALSE,rc=NULL) {
    message("Retrieving genome information for ",genome," from ",refdb)
    sf <- getSeqInfo(genome,asSeqinfo=TRUE)
    switch(type,
        gene = {
            message("Retrieving latest gene annotation for ",genome," from ",
                refdb)
            ann <- getAnnotation(genome,"gene",refdb=refdb,tv=versioned,rc=rc)
            annGr <- makeGRangesFromDataFrame(
                df=ann,
                seqinfo=sf,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            )
            names(annGr) <- as.character(annGr$gene_id)
        },
        transcript = {
            message("Retrieving latest transcript annotation for ",genome,
                " from ",refdb)
            ann <- getAnnotation(genome,"transcript",refdb=refdb,tv=versioned,
                rc=rc)
            annGr <- makeGRangesFromDataFrame(
                df=ann,
                seqinfo=sf,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            )
            names(annGr) <- as.character(annGr$transcript_id)
        },
        utr = {
            message("Retrieving latest 3' UTR annotation for ",genome," from ",
                refdb)
            ann <- getAnnotation(genome,"utr",refdb=refdb,tv=versioned,rc=rc)
            tmpGr <- makeGRangesFromDataFrame(
                df=ann,
                seqinfo=sf,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            )
            message("Merging 3' UTRs for ",genome," from ",refdb)
            annList <- reduceExons(tmpGr)
            annGr <- annList$model
            names(annGr) <- as.character(annGr$transcript_id)
            activeLength <- annList$length
            names(activeLength) <- unique(annGr$gene_id)
            attr(annGr,"activeLength") <- activeLength
        },
        transexon = {
            message("Retrieving latest transcrpit exon annotation for ",
                genome," from ",refdb)
            ann <- getAnnotation(genome,"transexon",refdb=refdb,tv=versioned,
                rc=rc)
            annGr <- makeGRangesFromDataFrame(
                df=ann,
                seqinfo=sf,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            )
            names(annGr) <- as.character(annGr$transcript_id)
        },
        transutr = {
            message("Retrieving latest 3' UTR annotation for ",genome," from ",
                refdb)
            ann <- getAnnotation(genome,"utr",refdb=refdb,tv=versioned,rc=rc)
            annGr <- makeGRangesFromDataFrame(
                df=ann,
                seqinfo=sf,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            )
            message("Merging 3' UTRs for ",genome," from ",refdb)
            annList <- reduceTranscriptsUtr(tmpGr)
            annGr <- annList$model
            names(annGr) <- as.character(annGr$transcript_id)
            activeLength <- annList$length
            names(activeLength) <- unique(annGr$transcript_id)
            attr(annGr,"activeLength") <- activeLength
        },
        exon = {
            message("Retrieving latest exon annotation for ",genome," from ",
                refdb)
            ann <- getAnnotation(genome,"exon",refdb=refdb,tv=versioned,rc=rc)
            annGr <- makeGRangesFromDataFrame(
                df=ann,
                seqinfo=sf,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            )
            names(annGr) <- as.character(annGr$exon_id)
        }
    )

    return(annGr)
}

.nameAnnotationFromMetaType <- function(ann,type) {
    switch(type,
        gene = {
            names(ann) <- as.character(ann$gene_id)
        },
        summarized_exon = {
            names(ann) <- as.character(ann$exon_id)
        },
        exon = {
            names(ann) <- as.character(ann$exon_id)
        },
        summarized_3utr = {
            names(ann) <- as.character(ann$transcript_id)
        },
        utr = {
            names(ann) <- as.character(ann$transcript_id)
        },
        summarized_transcript = {
            names(ann) <- as.character(ann$transcript_id)
        },
        transcript = {
            names(ann) <- as.character(ann$transcript_id)
        },
        summarized_3utr_transcript = {
            names(ann) <- as.character(ann$transcript_id)
        }
    )
    return(ann)
}

.annotationTypeFromInputArgs <- function(type,summarized=FALSE) {
    switch(type,
        gene = {
            return("gene")
        },
        transcript = {
            if (summarized)
                return("summarized_transcript")
            else
                return("transcript")
        },
        utr = {
            if (summarized)
                return("summarized_3utr")
            else
                return("utr")
        },
        transexon = {
            if (summarized)
                return("summarized_transcript_exon")
            else
                return("transexon")
        },
        transutr = {
            if (summarized)
                return("summarized_3utr_transcript")
            else
                return("utr")
        },
        exon = {
            if (summarized)
                return("summarized_exon")
            else
                return("exon")
        }
    )
}

.chromInfoFromBAM <- function(bam) {
    # Danger of including non-canonical chromosomes
    b <- scanBamHeader(bam)
    ci <- as.data.frame(b[[bam]]$targets)
    names(ci) <- "length"
    return(ci)
}

.chromInfoFromSeqinfo <- function(sf) {
    sf <- as.data.frame(sf)
    sf <- sf[,1,drop=FALSE]
    names(sf) <- "length"
    return(sf)
}

.chromInfoToSeqInfoDf <- function(ci,o="custom",circ=FALSE,asSeqinfo=FALSE) {
    if (asSeqinfo)
        return(Seqinfo(seqnames=rownames(ci),seqlengths=ci[,1],
            isCircular=rep(circ,nrow(ci)),genome=o))
    else
        return(data.frame(chromosome=rownames(ci),length=as.integer(ci[,1])))
}

.chromInfoWrapperGID <- function(o) {
    if (o == "TAIR10") {
        g <- GenomeInfoDb::getChromInfoFromEnsembl(o,division="plants")
        g$name = paste("chr",g$name,sep="")
        return(g)
    }
    else
        return(GenomeInfoDb::getChromInfoFromUCSC(o))
}

.validateDbCon <- function(obj) {
    # obj can be an already opened connection or the db file or missing. In the
    # latter case, the function looks at the package filesystem location
    if (missing(obj) || is.null(obj)) {
        db <-  getDbPath()
        drv <- dbDriver("SQLite")
        con <- tryCatch(dbConnect(drv,dbname=db),error=function(e) {
            message("Caught error: ",e)
            stop("Have you constructed the metaseqR annotation database?")
        },finally="")
    }
    if (is.character(obj) && file.exists(obj)) {
        drv <- dbDriver("SQLite")
        con <- tryCatch(dbConnect(drv,dbname=obj),error=function(e) {
            message("Caught error: ",e)
            stop("Is obj an SQLite database?")
        },finally="")
    }
    else if (is(obj,"SQLiteConnection"))
        con <- obj
    return(con)
}

.getValidChrs <- function(org) {
    switch(org,
        hg18 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        hg19 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        hg38 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        mm9 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY"
            ))
        },
        mm10 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY"
            ))
        },
        rn5 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX"
            ))
        },
        rn6 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX"
            ))
        },
        dm3 = {
            return(c(
                "chr2L","chr2LHet","chr2R","chr2RHet","chr3L","chr3LHet",
                "chr3R","chr3RHet","chr4","chrU","chrUextra","chrX","chrXHet",
                "chrYHet"
            ))
        },
        dm6 = {
            return(c(
                "chr2L","chr2R","chr3L","chr3R","chr4","chrX"
            ))
        },
        danrer7 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        danrer10 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        danrer11 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        pantro4 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        pantro5 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        pantro6 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        susscr3 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr2","chr3","chr4","chr5","chr6","chr7",
                "chr8","chr9","chrX","chrY"
            ))
        },
        susscr11 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr2","chr3","chr4","chr5","chr6","chr7",
                "chr8","chr9","chrX","chrY"
            ))
        },
        equcab2 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr26","chr27","chr28","chr29","chr3","chr30",
                "chr31","chr4","chr5","chr6","chr7","chr8","chr9","chrX"#,"chrY"
            ))
        },
        equcab3 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr26","chr27","chr28","chr29","chr3","chr30",
                "chr31","chr4","chr5","chr6","chr7","chr8","chr9","chrX"#,"chrY"
            ))
        },
        tair10 = {
            return(c(
                "chr1","chr2","chr3","chr4","chr5"
            ))
        }
    )
}

.getValidChrsWithMit <- function(org) {
    if (org %in% c("hg18","hg19","hg38","mm9","mm10","rn5","rn6","pantro4",
        "pantro5","pantro6","susscr3","susscr11","equcab2","equcab3"))
        return(c(.getValidChrs(org),"chrM"))
    else
        return(.getValidChrs(org))
}

.defaultDbPath <- function() {
    # We end up here if option for sitadela database path does not exist or has
    # not been provided by the user
    defPath <- .getDefaultDbPath()
    
    # User conformation only if database path not provided explicitly by the
    # user. If provided explicitly, user knows a database is going to be created
    # at the user-specified path.
    if (!interactive())
        .checkDbPath(defPath)
    else {
        if (!dir.exists(dirname(defPath))) {
            confirm <- menu(c("Yes","No"),
                title=sprintf(paste0("The sitadela database is going to be ",
                    "created at %s. Do you agree?"),defPath))
            if (confirm == 1)
                .checkDbPath()
            else
                stop(paste0("The sitadela database cannot be created without user ",
                    "agreement.\nPlease agree or use the 'db' argument in the ",
                    "'addAnnotation' function."))
        }        
    }
    
    return(defPath)
}

.checkDbPath <- function(db=NULL) {
    if (missing(db) || is.null(db))
        db <- .getDefaultDbPath()
    dbd <- dirname(db)
    if (!dir.exists(dbd)) {
        message("sitadela database created at ",dbd," directory")
        dir.create(dbd,recursive=TRUE,mode="0755",showWarnings=FALSE)
    }
    else
        message("sitadela database found at ",dbd," directory")
}

.getDefaultDbPath <- function() {
    return(file.path(tools::R_user_dir("sitadela","data"),"annotation.sqlite"))
}

.collapseFailures <- function(f) {
    df <- data.frame(
        org=vapply(f,function(x) return(x$org),character(1)),
        refdb=vapply(f,function(x) return(x$refdb),character(1)),
        ver=vapply(f,function(x) return(x$ver),numeric(1)),
        tv=vapply(f,function(x) return(x$tv),logical(1))
    )
    
    uorgs <- unique(df$org)
    organisms <- vector("list",length(uorgs))
    names(organisms) <- uorgs
    for (o in uorgs) {
        organisms[[o]] <- 
            df$ver[which(df$org==o & df$refdb=="ensembl")]
        if (length(organisms[[o]]) == 0) {
            # Not an Ensembl failure, but there must be a version
            # which will be ignored during the build unless
            # forceDownload = TRUE
            vs <- .getUcscToEnsembl(o)
            organisms[[o]] <- vs[length(vs)]
        }
    }
    
    return(list(organisms=organisms,sources=unique(df$refdb)))
}

.checkIfBSgenomesInstalled <- function(orgs,stopIfNotBS=FALSE) {
    miss <- bsi <- rep(NA,length(orgs))
    inst <- BSgenome::installed.genomes()
    ii <- 1
    for (org in orgs) {
        bsOrg <- getBsOrganism(org,.warn=FALSE)
        if (!(bsOrg %in% inst) && !is.na(bsOrg)) {
            miss[ii] <- org
            bsi[ii] <- bsOrg
            ii <- ii + 1
        }
    }
    if (!all(is.na(miss))) {
        miss <- miss[!is.na(miss)]
        bsi <- bsi[!is.na(bsi)]
        
        message("\nThe following BSgenome packages are missing:")
        for (i in seq_along(miss))
            message(bsi[i]," for ",miss[i])
        message("")
        
        if (!stopIfNotBS) {
            warning("Some requested organisms and sources require BSgenome ",
                "packages which are not installed!\nPlease install them or ",
                "the GC content in gene annotations will not be available...",
                immediate.=TRUE)    
        }
        else
            stop("Some requested organisms and sources require BSgenome ",
                "packages which are not installed!\nPlease install them and ",
                "restart building or choose stopIfNotBs = FALSE but\nthe GC ",
                "content in gene annotations will not be available...")
    }
}

.getUcscTabledefUcsc <- function(org,what="queries") {
    switch(org,
        hg18 = {
            return(list(
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownGene=.getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                refFlat=.getUcscTblTpl("refFlat",what)
            ))
        },
        hg19 = {
            return(list(
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownGene=.getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what),
                refFlat=.getUcscTblTpl("refFlat",what)
            ))
        },
        hg38 = {
            return(list(
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownGene=.getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                refFlat=.getUcscTblTpl("refFlat",what)
            ))
        },
        mm9 = {
            return(list(
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownGene=.getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what),
                refFlat=.getUcscTblTpl("refFlat",what)
            ))
        },
        mm10 = {
            return(list(
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownGene=.getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what),
                refFlat=.getUcscTblTpl("refFlat",what)
            ))
        },
        rn5 = {
            return(list(
                mgcGenes=.getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        rn6 = {
            return(list(
                mgcGenes=.getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        dm3 = {
            return(list(
                flyBaseCanonical=
                    .getUcscTblTpl("flyBaseCanonical",what),
                flyBaseGene=
                    .getUcscTblTpl("flyBaseGene",what),
                flyBaseToRefSeq=
                    .getUcscTblTpl("flyBaseToRefSeq",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what),
                refFlat=.getUcscTblTpl("refFlat",what)
            ))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer7 = {
            return(list(
                mgcGenes=.getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer10 = {
            return(list(
                mgcGenes=.getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! ",
                "Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        #pantro6 = {
        #    warning("No UCSC Genome annotation for Pan ",
        #        "troglodytes! Will use RefSeq instead...",
        #        immediate.=TRUE)
        #    return(list(
        #        refFlat=.getUcscTblTpl("refFlat",what),
        #        ensemblToGeneName=
        #            .getUcscTblTpl("ensemblToGeneName",what),
        #        ensemblSource=
        #            .getUcscTblTpl("ensemblSource",what)
        #    ))
        #},
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        equcab2 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v2! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        }
    )
}

.getUcscQueryUcscGene <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should have been like hg19 but it's like hg18
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
            #   "`chromStart` AS `start`,",
            #   "`chromEnd` AS `end`,",
            #   "`transcript` AS `gene_id`,",
            #   "0 AS `gc_content`,",
            #   "knownGene.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownCanonical` INNER JOIN `knownGene`", 
            #   "ON knownCanonical.transcript=knownGene.name",
            #   "INNER JOIN `knownToRefSeq`", 
            #   "ON knownCanonical.transcript=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat`",
            #   "ON knownToRefSeq.value=refFlat.name",
            #   "GROUP BY `gene_id`",
            #   "ORDER BY `chromosome`,`start`"))
            ## No Ensembl source...
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm3 = {
            return(paste("SELECT flyBaseCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "flyBaseGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `flyBaseCanonical` INNER JOIN `flyBaseGene`", 
                "ON flyBaseCanonical.transcript=flyBaseGene.name",
                "INNER JOIN `flyBaseToRefSeq`",
                "ON flyBaseCanonical.transcript=flyBaseToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON flyBaseToRefSeq.value=refFlat.name",
                "INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`", 
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! Will use ",
                "RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan troglodytes v4! Will ",
                "use RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v5! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab2 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v2! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscQueryUcscTranscript <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownGene.name=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat` ON", 
                "knownToRefSeq.value=refFlat.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is like hg18
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownGene.name=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat` ON",
                "knownToRefSeq.value=refFlat.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT knownGene.chrom AS `chromosome`,",
            #   "knownGene.txStart AS `start`,",
            #   "knownGene.txEnd AS `end`,",
            #   "knownGene.name AS `transcript_id`,",
            #   "knownGene.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
            #   "ON knownGene.name=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownGene.name=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat` ON",
            #   "knownToRefSeq.value=refFlat.name",
            #   "GROUP BY `transcript_id`",
            #   "ORDER BY `chromosome`,`start`"))
            ## No ensemblSource...
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'transcript' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            return(paste("SELECT flyBaseGene.chrom AS `chromosome`,",
                "flyBaseGene.txStart AS `start`,",
                "flyBaseGene.txEnd AS `end`,",
                "flyBaseGene.name AS `transcript_id`,",
                "flyBaseGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `flyBaseGene` INNER JOIN `flyBaseToRefSeq`",
                "ON flyBaseGene.name=flyBaseToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON flyBaseToRefSeq.value=refFlat.name",
                "INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer7 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'transcript' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer10 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'transcript' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! Will use ",
                "RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v4! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v5! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab2 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v2! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v2! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        }
    )
}

.getUcscQueryUcscExon <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`", 
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`",
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT knownGene.chrom AS `chromosome`,",
            #   "knownGene.exonStarts AS `start`,",
            #   "knownGene.exonEnds AS `end`,",
            #   "knownGene.name AS `exon_id`,",
            #   "knownGene.strand AS `strand`,",
            #   "`transcript` AS `gene_id`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownGene` INNER JOIN `knownCanonical`",
            #   "ON knownGene.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToRefSeq`",
            #   "ON knownCanonical.transcript=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat`",
            #   "ON knownToRefSeq.value=refFlat.name",
            #   "GROUP BY knownGene.name",
            #   "ORDER BY `chromosome`,`start`"))
            ## No Ensembl source...
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`", 
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
            #    "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`" ,
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`", 
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            return(paste("SELECT flyBaseCanonical.chrom AS `chromosome`,",
                "flyBaseGene.exonStarts AS `start`,",
                "flyBaseGene.exonEnds AS `end`,",
                "`transcript` AS `exon_id`,",
                "flyBaseGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `flyBaseCanonical` INNER JOIN `flyBaseGene` ON",
                "flyBaseCanonical.transcript=flyBaseGene.name",
                "INNER JOIN `flyBaseToRefSeq`",
                "ON flyBaseCanonical.transcript=flyBaseToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON flyBaseToRefSeq.value=refFlat.name",
                "INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
            #    "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
            #    "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #   "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! Will use ",
                "RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan troglodytes v4! Will ",
                "use RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v5! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab2 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v2! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

test_sitadela_known <- function() {
    testResult <- testKnownBuild("dm6","ensembl",102)
    checkTrue(all(testResult))
}

test_sitadela_gtf <- function() {
    gtf <- file.path(system.file(package="sitadela"),"dummy.gtf.gz")
    chromInfo <- data.frame(length=c(1000L,2000L,1500L),
        row.names=c("A","B","C"))
    metadata=list(
        organism="dummy",
        source="dummy_db",
        version=1,
        chromInfo=chromInfo
    )
    testResult <- testCustomBuild(gtf,metadata)
    checkTrue(all(testResult))
}

test_sitadela_ensembl_often <- function() {
    f1 <- testEnsemblSimple("mm10","gene")
    checkTrue(is.null(f1))
    
    f2 <- testEnsemblSimple("mm10","transcript")
    checkTrue(is.null(f2))
}

test_sitadela_ensembl_rare <- function() {
    f1 <- testEnsemblSimple("danrer11","gene")
    checkTrue(is.null(f1))
    
    f2 <- testEnsemblSimple("danrer11","transcript")
    checkTrue(is.null(f2))
}

test_sitadela_ucsc_often <- function() {
    if (require(RMySQL)) {
        f1 <- testUcsc("hg19","refseq","transcript",FALSE)
        f2 <- testUcsc("mm10","refseq","gene",FALSE)
        f3 <- testUcsc("hg19","ucsc","transcript",FALSE)
    }
    else
        f1 <- f2 <- f3 <- NULL
    checkTrue(is.null(f1))
    checkTrue(is.null(f2))
    checkTrue(is.null(f3))
}

test_sitadela_ucsc_rare <- function() {
    if (require(RMySQL)) {
        f1 <- testUcsc("dm3","ucsc","gene",FALSE)
        f2 <- testUcsc("equcab3","refseq","gene",FALSE)
        f3 <- testUcsc("danrer11","refseq","transcript",FALSE)
    }
    else
        f1 <- f2 <- f3 <- NULL
    checkTrue(is.null(f1))
    checkTrue(is.null(f2))
    checkTrue(is.null(f3))
}

test_sitadela_ucsc_versioned <- function() {
    if (require(RMySQL)) {
        f1 <- testUcsc("hg19","refseq","gene",TRUE)
        f2 <- testUcsc("mm10","refseq","gene",TRUE)
        f3 <- testUcsc("danrer11","refseq","transcript",TRUE)
    }
    else
        f1 <- f2 <- f3 <- NULL
    checkTrue(is.null(f1))
    checkTrue(is.null(f2))
    checkTrue(is.null(f3))
}

test_sitadela_argcheck <- function() {
    f1 <- tryCatch({
        sitadela:::.checkTextArgs("test","test","test",FALSE)
        sitadela:::.checkTextArgs("test","test1",c("test1","test2"),TRUE)
        TRUE
    },error=function(e) {
        return(FALSE)
    },finally="")
    
    f2 <- tryCatch({
        # Should crash before TRUE
        sitadela:::.checkTextArgs("test","tes","test",FALSE)
        TRUE
    },error=function(e) {
        return(FALSE)
    },finally="")
    
    f3 <- tryCatch({
        sitadela:::.checkNumArgs("test",4L,"integer",3L,"gt")
        sitadela:::.checkNumArgs("test",4,"numeric",5,"lt")
        sitadela:::.checkNumArgs("test",4,"numeric",c(3,4 ),"both")
        TRUE
    },error=function(e) {
        return(FALSE)
    },finally="")
    
    f4 <- tryCatch({
        sitadela:::.checkNumArgs("test",3L,"integer",3L,"gt")
        TRUE
    },error=function(e) {
        return(FALSE)
    },finally="")
    
    checkTrue(f1)
    checkTrue(!f2)
    checkTrue(f3)
    checkTrue(!f4)
}

correctTranscripts <- function(ann) {
    rownames(ann) <- paste("T",seq_len(nrow(ann)),sep="_")
    len <- ann[,3] - ann[,2]
    len <- len[-which(is.na(len))]
    len[len==0] <- 1
    defUtrLen <- round(2^mean(log2(len)))
    nas <- which(is.na(ann$start))
    annNa <- ann[nas,]
    annNa$start <- annNa$tstart
    annNa$end <- annNa$tend
    tmp <- makeGRangesFromDataFrame(df=annNa)
    tmp <- flank(resize(tmp,width=1,fix="end"),start=FALSE,width=defUtrLen)
    ann[names(tmp),"start"] <- start(tmp)
    ann[names(tmp),"end"] <- end(tmp)
    return(ann)
}

reduceTranscripts <- function(gr) {
    # Get the gene ids to use as split factor
    gene <- unique(as.character(gr$gene_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
    
    # There will be a transcript_id
    ir <- which(names(meta) == "transcript_id")
    map <- meta[,-ir]
    
    # Remove duplicates and name the map
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,]
    # In some cases, transcripts are recorded with multiple biotypes...
    if (length(map$gene_id) != length(unique(map$gene_id))) {
        mapp <- map
        ir <- which(names(mapp) == "biotype")
        mapp <- mapp[,-ir]
        d <- which(duplicated(mapp))
        if (length(d) > 0)
            map <- map[-d,]
    }
    rownames(map) <- map$gene_id
    map <- map[gene,]
    
    # Gene names and biotypes for later reconstruction
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$gene_id)
    # Re-order for common reference with the map
    grList <- grList[gene]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    grNew <- unname(unlist(grList))
    gene_id <- rep(gene,lens)
    transcript_id <- paste(gene_id,"MET",inds,sep="_")
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        transcript_id=transcript_id,
        gene_id=gene_id,
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$gene_id)
    tmp <- tmp[gene]
    len <- vapply(width(tmp),sum,integer(1))
    
    #return(grNew)
    return(list(model=grNew,length=len))
}

reduceTranscriptsUtr <- function(gr) {
    # Get the transcript ids to use ordering element
    trans <- unique(as.character(gr$transcript_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
    
    # Make the map and remove duplicates
    map <- meta
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,,drop=FALSE]
    rownames(map) <- map$transcript_id
    map <- map[trans,]
    
    # Gene ids, names and biotypes for later reconstruction
    gi <- as.character(map$gene_id)
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$transcript_id)
    # Re-order for common reference with the map
    grList <- grList[trans]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    grNew <- unname(unlist(grList))
    transcript_id <- paste(rep(trans,lens),"MEU",inds,sep="_")
    gene_id <- rep(gi,lens)
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        transcript_id=transcript_id,
        #gene_id=gene_id,
        gene_id=rep(trans,lens),
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$gene_id)
    tmp <- tmp[trans]
    len <- vapply(width(tmp),sum,integer(1))
    
    #return(grNew)
    return(list(model=grNew,length=len))
}

reduceExons <- function(gr) {
    # Get the gene ids to use as split factor
    gene <- unique(as.character(gr$gene_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
    
    # There will be an exon_id
    ir <- which(names(meta) == "exon_id")
    map <- meta[,-ir]
    
    # Remove duplicates and name the map
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,]
    rownames(map) <- map$gene_id
    map <- map[gene,]
    
    # Gene names and biotypes for later reconstruction
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$gene_id)
    # Re-order for common reference with the map
    grList <- grList[gene]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    strs <- as.character(runValue(strand(grList)))
    #inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    inds <- unlist(lapply(seq_along(lens),function(j,L,S) {
        if (S[j] == "+")
            return(seq_len(L[j]))
        else
            return(rev(seq_len(L[j])))
    },lens,strs),use.names=FALSE)
    grNew <- unname(unlist(grList))
    gene_id <- rep(gene,lens)
    exon_id <- paste(gene_id,"MEX",inds,sep="_")
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        exon_id=exon_id,
        gene_id=gene_id,
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$gene_id)
    tmp <- tmp[gene]
    len <- vapply(width(tmp),sum,integer(1))
    
    return(list(model=grNew,length=len))
}

reduceTranscriptsExons <- function(gr) {
    # Get the transcript ids to use ordering element
    trans <- unique(as.character(gr$transcript_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
        
    # There will be an exon_id
    ir <- which(names(meta) == "exon_id")
    map <- meta[,-ir]
    
    # Make the map and remove duplicates
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,,drop=FALSE]
    rownames(map) <- map$transcript_id
    map <- map[trans,]
    
    # Gene ids, names and biotypes for later reconstruction
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$transcript_id)
    # Re-order for common reference with the map
    grList <- grList[trans]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    strs <- as.character(runValue(strand(grList)))
    #inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    inds <- unlist(lapply(seq_along(lens),function(j,L,S) {
        if (S[j] == "+")
            return(seq_len(L[j]))
        else
            return(rev(seq_len(L[j])))
    },lens,strs),use.names=FALSE)
    grNew <- unname(unlist(grList))
    exon_id <- paste(rep(trans,lens),"MTE",inds,sep="_")
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        exon_id=exon_id,
        transcript_id=rep(trans,lens),
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$transcript_id)
    tmp <- tmp[trans]
    len <- vapply(width(tmp),sum,integer(1))
    
    return(list(model=grNew,length=len))
}

initDatabase <- function(db) {
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to build the annotation database!")
    if (missing(db))
        stop("A database file must be provided!")
    drv <- dbDriver("SQLite")
    if (file.exists(db))
        # The database has been created at least with the tables defined
        con <- dbConnect(drv,dbname=db)
    else {
        # Create database file and define tables
        con <- dbConnect(drv,dbname=db)
        rs <- .initTables(con)
    }
    return(con)
}

getUcscTableNameUtr <- function(org,refdb) {
    switch(refdb,
        ucsc = {
            switch(org,
                hg18 = {
                    return("knownGene")
                },
                hg19 = {
                    return("knownGene")
                },
                hg38 = {
                    return("knownGene")
                },
                mm9 = {
                    return("knownGene")
                },
                mm10 = {
                    return("knownGene")
                },
                rn5 = {
                    return("mgcGenes")
                },
                rn6 = {
                    return("mgcGenes")
                },
                dm3 = {
                    return("flyBaseGene")
                },
                dm6 = {
                    warning("No UCSC Genome annotation for Drosophila ",
                        "melanogaster v6! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                danrer7 = {
                    return("mgcGenes")
                },
                danrer10 = {
                    return("mgcGenes")
                    
                },
                danrer11 = {
                    warning("No UCSC Genome annotation for Danio rerio v11! ",
                        "Will use RefSeq instead...",immediate.=TRUE)
                    return("refGene")
                },
                pantro4 = {
                    warning("No UCSC Genome annotation for Pan ",
                        "troglodytes! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                pantro5 = {
                    warning("No UCSC Genome annotation for Pan ",
                        "troglodytes! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                susscr3 = {
                    warning("No UCSC Genome annotation for Sus ",
                        "scrofa v3! Will use RefSeq instead...",
                        immediate.=TRUE)
                        return("refGene")
                },
                susscr11 = {
                    warning("No UCSC Genome annotation for Sus ",
                        "scrofa v11! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                equcab2 = {
                    warning("No UCSC Genome annotation for Equus ",
                        "caballus v2! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                equcab3 = {
                    warning("No UCSC Genome annotation for Equus ",
                        "caballus v3! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                }
            )
        },
        refseq = {
            return("refGene")
        },
        ncbi = {
            switch(org,
                hg18 = {
                    warning("No NCBI RefSeq Genome annotation for Homo ",
                        "sapiens hg18! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                hg19 = {
                    return("ncbiRefSeq")
                },
                hg38 = {
                    return("ncbiRefSeq")
                },
                mm9 = {
                    warning("No NCBI RefSeq Genome annotation for Mus ",
                        "musculus mm9! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                mm10 = {
                    return("ncbiRefSeq")
                },
                rn5 = {
                    warning("No NCBI RefSeq Genome annotation for Rattus ",
                        "norvegicus rn5! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                rn6 = {
                    return("ncbiRefSeq")
                },
                dm3 = {
                    warning("No NCBI RefSeq Genome annotation for Drosophila ",
                        "melanogaster dm3! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                dm6 = {
                    return("ncbiRefSeq")
                },
                danrer7 = {
                    warning("No NCBI RefSeq Genome annotation for Danio ",
                        "rerio danrer7! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                danrer10 = {
                    return("ncbiRefSeq")
                    
                },
                danrer11 = {
                    return("ncbiRefSeq")
                },
                pantro4 = {
                    warning("No NCBI RefSeq Genome annotation for Pan ",
                        " troglodytes pantro4! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                pantro5 = {
                    warning("No NCBI RefSeq Genome annotation for Pan ",
                        " troglodytes pantro5! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                susscr3 = {
                    warning("No NCBI RefSeq Genome annotation for Sus scrofa ",
                        " v3! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                susscr11 = {
                    warning("No NCBI RefSeq Genome annotation for Sus scrofa ",
                        " v11! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                equcab2 = {
                    warning("No NCBI RefSeq Genome annotation for Equus ",
                        " cabalus v2! Will use UCSC RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                equcab3 = {
                    return("ncbiRefSeq")
                }
            )
        }
    )
}

getUcscDbl <- function(org,refdb="ucsc",versioned=FALSE) {
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    .checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    .checkTextArgs("refdb",refdb,getSupportedUcscDbs())
    
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to use annotation from UCSC!")

    tableDefs <- .getUcscTabledef(org,refdb,"fields",versioned)
    fileList <- vector("list",length(tableDefs))
    names(fileList) <- names(tableDefs)
    
    # If versioning required, fetch gbCdnaInfo only once
    if (versioned && refdb=="refseq") {
        verInfo <- paste0("http://hgdownload.soe.ucsc.edu/goldenPath/hgFixed/",
            "database/gbCdnaInfo.txt.gz")
        if (!file.exists(file.path(tempdir(),"gbCdnaInfo.txt.gz")))
            download.file(verInfo,file.path(tempdir(),"gbCdnaInfo.txt.gz"),
                quiet=TRUE)
        # gbCdnaInfo should be the last element
        fileList <- fileList[-length(fileList)]
    }
    
    httpBase <- paste("http://hgdownload.soe.ucsc.edu/goldenPath/",
        getUcscOrganism(org),"/database/",sep="")
    for (n in names(fileList))
        fileList[[n]] <- paste(httpBase,n,".txt.gz",sep="")
        
    # Fill the fields for each table
    drv <- dbDriver("SQLite")
    dbTmp <- tempfile()
    con <- dbConnect(drv,dbname=dbTmp)
    message("  Retrieving tables for temporary SQLite ",refdb," ",org,
        " subset database")
    for (n in names(fileList)) {
        message("    Retrieving table ",n)
        download.file(fileList[[n]],file.path(tempdir(),
            paste(n,".txt.gz",sep="")),quiet=TRUE)
        sqlDf <- read.delim(file.path(tempdir(),paste(n,".txt.gz",sep="")),
            row.names=NULL,header=FALSE,strip.white=TRUE)
        names(sqlDf) <- tableDefs[[n]]
        dbWriteTable(con,n,sqlDf,row.names=FALSE)
    }
    
    # And now read also gbCdnaInfo in
    if (versioned && refdb=="refseq") {
        sqlDf <- read.delim(file.path(tempdir(),"gbCdnaInfo.txt.gz"),
            row.names=NULL,header=FALSE,strip.white=TRUE)
        names(sqlDf) <- tableDefs[["gbCdnaInfo"]]
        dbWriteTable(con,"gbCdnaInfo",sqlDf,row.names=FALSE)
    }
    
    dbDisconnect(con)
    return(dbTmp)
}

.initTables <- function(con) {
    queries <- .localTblDef()
    rs <- dbSendQuery(con,queries[[1]])
    if (dbHasCompleted(rs))
        dbClearResult(rs)
    for (n in names(queries)) {
        rs <- dbSendStatement(con,queries[[n]])
        if (dbHasCompleted(rs))
            dbClearResult(rs)
    }
}

.localTblDef <- function() {
    return(list(
        enable_fkey="PRAGMA foreign_keys=1;",
        #content=paste(
        #    "CREATE TABLE IF NOT EXISTS content (",
        #    "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
        #    "source TEXT,",
        #    "organism TEXT,",
        #    "version INTEGER,",
        #    "type TEXT,",
        #    "user INTEGER",
        #    ");"
        #),
        content=paste(
            "CREATE TABLE IF NOT EXISTS content (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "source TEXT,",
            "organism TEXT,",
            "version INTEGER,",
            "type TEXT,",
            "has_tv INTEGER,",
            "user INTEGER",
            ");"
        ),
        seqinfo=paste(
            "CREATE TABLE IF NOT EXISTS seqinfo (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "length INTEGER,",
            #"source TEXT,",
            #"organism TEXT,",
            #"version INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        gene=paste(
            "CREATE TABLE IF NOT EXISTS gene (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "gene_id TEXT,",
            "gc_content REAL,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        transcript=paste(
            "CREATE TABLE IF NOT EXISTS transcript (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        exon=paste(
            "CREATE TABLE IF NOT EXISTS exon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        transexon=paste(
            "CREATE TABLE IF NOT EXISTS transexon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "transcript_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        utr=paste(
            "CREATE TABLE IF NOT EXISTS utr (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_transcript=paste(
            "CREATE TABLE IF NOT EXISTS summarized_transcript (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_exon=paste(
            "CREATE TABLE IF NOT EXISTS summarized_exon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_transcript_exon=paste(
            "CREATE TABLE IF NOT EXISTS summarized_transcript_exon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "transcript_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_3utr=paste(
            "CREATE TABLE IF NOT EXISTS summarized_3utr (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_3utr_transcript=paste(
            "CREATE TABLE IF NOT EXISTS summarized_3utr_transcript (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_length=paste(
            "CREATE TABLE IF NOT EXISTS active_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_utr_length=paste(
            "CREATE TABLE IF NOT EXISTS active_utr_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_trans_utr_length=paste(
            "CREATE TABLE IF NOT EXISTS active_trans_utr_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_trans_exon_length=paste(
            "CREATE TABLE IF NOT EXISTS active_trans_exon_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        )
    ))
}

.makeAnnotationQuerySet <- function(t,i,j=NULL) {
    mainQuery <- paste("SELECT * FROM ",t," WHERE content_id=",i,sep="")
    seqInfoQuery <- paste("SELECT * FROM seqinfo WHERE content_id=",i,sep="")
    activeQuery <- NULL
    if (t == "summarized_exon" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_length WHERE content_id=",
            j,sep="")
    else if (t == "summarized_3utr" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_utr_length WHERE ",
            "content_id=",j,sep="")
    else if (t == "summarized_3utr_transcript" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_trans_utr_length WHERE ",
            "content_id=",j,sep="")
    else if (t == "summarized_transcript_exon" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_trans_exon_length WHERE ",
            "content_id=",j,sep="")
    return(list(
        main=mainQuery,
        seqinfo=seqInfoQuery,
        active=activeQuery
    ))
}

.insertContent <- function(con,o,s,v,t,h=0,u=0) {
    query <- paste(
        "INSERT INTO content (source, organism, version, type, has_tv, user) ",
        "VALUES (",paste("'",s,"', ","'",o,"', ",v,", '",t,"', ",h,", ",u,
        sep=""),")",sep=""
    )
    nr <- dbExecute(con,query)
    return(nr)
}

.browseContent <- function(con) {
    return(dbGetQuery(con,"SELECT * FROM content"))
}

.browseUserContent <- function(con) {
    return(dbGetQuery(con,"SELECT * FROM content WHERE user=1"))
}

.annotationExists <- function(con,o,s,v=NULL,t=NULL,h=NULL,
    out=c("tf","nr","id")) {
    out <- out[1]
    if (!is.null(h) && is.logical(h))
        h <- ifelse(h,1,0)
    else
        h <- NULL
    query <- paste("SELECT _id FROM content WHERE source='",s,
        "' AND organism='",o,"'",sep="")
    if (!is.null(v))
        query <- paste(query," AND version=",v,sep="")
    if (!is.null(t))
        query <- paste(query," AND type='",t,"'",sep="")
    if (!is.null(h))
        query <- paste(query," AND has_tv=",h,sep="")
    res <- dbGetQuery(con,query)
    if (out == "tf")
        return(nrow(res) > 0)
    else if (out=="nr")
        return(nrow(res))
    else if (out == "id") {
        if (nrow(res) > 0)
            return(res[1,1])
    }
}

.containsVersionedGT <- function(con,o,s,only=FALSE) {
    query <- paste("SELECT has_tv FROM content WHERE source='",s,
        "' AND organism='",o,"'",sep="")
    res <- dbGetQuery(con,query)
    out <- as.integer(res[,1])
    if (!only)
        return(any(out==1))
    else
        return(all(out==1))
}

.dropAnnotation <- function(con,o,s,v,t,h) {
    # At least organism must exist
    if (missing(o))
        stop("At least an organism name must be provided for deletion!")
    # A basic deletion query based on organism. Since the content table is
    # connected with the rest through foreing keys, deletion from there should
    # be enough.
    query <- paste("DELETE FROM content WHERE organism='",o,"'",sep="")
    # Augment according to given arguments.
    if (!missing(s) && !is.null(s))
        query <- paste(query," AND source='",s,"'",sep="")
    if (!missing(v) && !is.null(v))
        query <- paste(query," AND version=",v,sep="")
    if (!missing(t) && !is.null(t))
        query <- paste(query," AND type='",t,"'",sep="")
    if (!missing(h) && !is.null(h))
        query <- paste(query," AND has_tv=",h,sep="")
    # Execute
    nr <- dbExecute(con,query)
    return(nr)
}

.installedVersions <- function(con,o,s) {
    query <- paste(
        "SELECT version FROM content WHERE source='",s,"' AND organism='",o,"'",
        sep=""
    )
    res <- dbGetQuery(con,query)
    if (nrow(res) > 0)
        return(as.numeric(res[,1]))
    else
        return(NA)
}

.getUcscTabledef <- function(org,refdb="ucsc",what="queries",versioned=FALSE) {
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    what <- tolower(what[1])
    .checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    .checkTextArgs("refdb",refdb,getSupportedUcscDbs())
    .checkTextArgs("what",what,c("queries","fields"))
    switch(refdb,
        ucsc = {
            return(.getUcscTabledefUcsc(org,what))
        },
        refseq = {
            return(.getUcscTabledefRefseq(org,what,versioned))
        },
        ncbi = {
            return(.getUcscTabledefNcbi(org,what))
        }
    )
}

.getUcscTblTpl <- function(tab,what="queries") {
    if (what=="queries") {
        switch(tab,
            knownCanonical = {
                return(paste(
                    "CREATE TABLE",
                    "`knownCanonical` (",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`chromStart` INTEGER NOT NULL DEFAULT '0',",
                    "`chromEnd` INTEGER NOT NULL DEFAULT '0',",
                    "`clusterId` INTEGER NOT NULL DEFAULT '0',",
                    "`transcript` TEXT NOT NULL DEFAULT '',",
                    "`protein` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            knownGene = {
                return(paste(
                    "CREATE TABLE",
                    "`knownGene` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`strand` TEXT NOT NULL DEFAULT '',",
                    "`txStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`txEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonCount` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL,",
                    "`proteinID` TEXT NOT NULL DEFAULT '',",
                    "`alignID` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            knownToRefSeq = {
                return(paste(
                    "CREATE TABLE",
                    "`knownToRefSeq` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`value` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            refFlat = {
                return(paste("CREATE TABLE",
                    "`refFlat` (",
                    "`geneName` TEXT NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            refGene = {
                return(paste("CREATE TABLE",
                    "`refGene` (",
                    "`bin` UNSIGNED INTEGER NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    "`score` INTEGER NOT NULL,",
                    "`name2` TEXT NOT NULL,",
                    "`cdsStartStat` TEXT NOT NULL,",
                    "`cdsEndStat` TEXT NOT NULL,",
                    "`exonFrames` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            knownToEnsembl = {
                return(paste(
                    "CREATE TABLE",
                    "`knownToEnsembl` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`value` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            ensemblSource = {
                return(paste(
                    "CREATE TABLE",
                    "`ensemblSource` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`source` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            mgcGenes = {
                return(paste(
                    "CREATE TABLE `mgcGenes` (",
                    "`bin` UNSIGNED INTEGER NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL,",
                    "`score` INTEGER DEFAULT NULL,",
                    "`name2` TEXT NOT NULL,",
                    "`cdsStartStat` TEXT NOT NULL,",
                    "`cdsEndStat` TEXT NOT NULL,",
                    "`exonFrames` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            ensemblToGeneName = {
                return(paste(
                    "CREATE TABLE",
                    "`knownToGeneName` (",
                    "`name` TEXT NOT NULL,",
                    "`value` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            flyBaseCanonical = {
                return(paste(
                    "CREATE TABLE",
                    "`flyBaseCanonical` (",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`chromStart` INTEGER NOT NULL DEFAULT '0',",
                    "`chromEnd` INTEGER NOT NULL DEFAULT '0',",
                    "`clusterId` INTEGER unsigned NOT NULL DEFAULT '0',",
                    "`transcript` TEXT NOT NULL DEFAULT '',",
                    "`protein` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            flyBaseGene = {
                return(paste(
                    "CREATE TABLE",
                    "`flyBaseGene` (",
                    "`bin` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`strand` TEXT NOT NULL DEFAULT '',",
                    "`txStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`txEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonCount` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            flyBaseToRefSeq = {
                return(paste(
                    "CREATE TABLE",
                    "`flyBaseToRefSeq` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`value` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            gbCdnaInfo = {
                return(paste(
                    "CREATE TABLE",
                    "`gbCdnaInfo` (",
                    "`bin` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`acc` TEXT NOT NULL DEFAULT '',",
                    "`version` UNSIGNED INTEGER NOT NULL DEFAULT '1',",
                    "`type` TEXT NOT NULL DEFAULT '',",
                    "`direction` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`source` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`organism` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`library` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`mrnaClone` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`sex` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`tissue` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`development` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cell` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cds` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`keyword` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`description` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`geneName` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`productName` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`author` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`gi` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`mol` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            ncbiRefSeq = {
                return(paste("CREATE TABLE",
                    "`ncbiRefSeq` (",
                    "`bin` UNSIGNED INTEGER NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    "`score` INTEGER NOT NULL,",
                    "`name2` TEXT NOT NULL,",
                    "`cdsStartStat` TEXT NOT NULL,",
                    "`cdsEndStat` TEXT NOT NULL,",
                    "`exonFrames` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
        )
    }
    else if (what=="fields") {
        switch(tab,
            knownCanonical = {
                return(c("chrom","chromStart","chromEnd","clusterId",
                "transcript","protein"))
            },
            knownGene = {
                return(c("name","chrom","strand","txStart","txEnd","cdsStart",
                    "cdsEnd","exonCount","exonStarts","exonEnds","proteinID",
                    "alignID"))
            },
            knownToRefSeq = {
                return(c("name","value"))
            },
            refFlat = {
                return(c("geneName","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds"))
            },
            refGene = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds","score",
                "name2","cdsStartStat","cdsEndStat","exonFrames"))
            },
            knownToEnsembl = {
                return(c("name","value"))
            },
            ensemblSource = {
                return(c("name","source"))
            },
            mgcGenes = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds",
                    "score","name2","cdsStartStat","cdsEndStat","exonFrames"
                ))
            },
            ensemblToGeneName = {
                return(c("name","value"))
            },
            flyBaseCanonical = {
                return(c("chrom","chromStart","chromEnd","clusterId",
                    "transcript","protein"))
            },
            flyBaseGene = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds"))
            },
            flyBaseToRefSeq = {
                return(c("name","value"))
            },
            gbCdnaInfo = {
                return(c("bin","acc","version","moddate","type","direction",
                    "source","organism","library","mrnaClone","sex","tissue",
                    "development","cell","cds","keyword","description",
                    "geneName","productName","author","gi","mol"))
            },
            ncbiRefSeq = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds",
                    "score","name2","cdsStartStat","cdsEndStat","exonFrames"))
            }
        )
    }
}

.getUcscQuery <- function(org,type,refdb="ucsc",versioned=FALSE) {
    type <- tolower(type[1])
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    .checkTextArgs("type",type,c("gene","exon","transcript"))
    .checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    .checkTextArgs("refdb",refdb,c("ucsc","refseq","ncbi"))
    switch(type,
        gene = {
            switch(refdb,
                ucsc = {
                    return(.getUcscQueryUcscGene(org))
                },
                refseq = {
                    return(.getUcscQueryRefseqGene(org,versioned=versioned))
                },
                ncbi = {
                    return(.getUcscQueryNcbiGene(org))
                }
            )
        },
        exon = {
            switch(refdb,
                ucsc = {
                    return(.getUcscQueryUcscExon(org))
                },
                refseq = {
                    return(.getUcscQueryRefseqExon(org,versioned=versioned))
                },
                ncbi = {
                    return(.getUcscQueryNcbiExon(org))
                }
            )
        },
        transcript = {
            switch(refdb,
                ucsc = {
                    return(.getUcscQueryUcscTranscript(org))
                },
                refseq = {
                    return(.getUcscQueryRefseqTranscript(org,
                        versioned=versioned))
                },
                ncbi = {
                    return(.getUcscQueryNcbiTranscript(org))
                }
            )
        }
    )
}

.makeUcscRefseqUtrTable <- function(org,.rmysql=FALSE) {
    vq <- .getUcscRefseqVersionedUtrQuery()
    if (.rmysql) {
        dbCreds <- .getUcscCredentials()
        drv <- dbDriver("MySQL")
        con <- dbConnect(drv,user=dbCreds[2],password=NULL,
            dbname=getUcscOrganism(org),host=dbCreds[1])
        rawAnn <- dbGetQuery(con,vq)
        dbDisconnect(con)
    }
    else {
        # Download tables, build sqlite and run vq
        tmpSqlite <- getUcscDbl(org=org,refdb="ucsc",versioned=TRUE)
        drv <- dbDriver("SQLite")
        con <- dbConnect(drv,dbname=tmpSqlite)
        rawAnn <- dbGetQuery(con,vq)
        dbDisconnect(con)
    }
    utrTmp <- tempfile()
    options(scipen=32) # genePredToGtf has problem with scientific notation
    write.table(rawAnn,file=gzfile(paste0(utrTmp,".txt.gz")),sep="\t",
        col.names=FALSE,row.names=FALSE,quote=FALSE)
    options(scipen=0)
    return(basename(utrTmp))
}

.getUcscCredentials <- function() {
    return(c(
        #host="genome-mysql.cse.ucsc.edu",
        host="genome-mysql.soe.ucsc.edu",
        user="genome",
        password=""
    ))
}

#..getAllUcsc <- function(to) {
#    if (missing(to))
#        stop("Please provide a path to put the sqlite databases!")
#    if (!dir.exists(to))
#        dir.create(to,recursive=TRUE)
#        
#    for (refdb in getSupportedUcscDbs())
#        for (org in getSupportedOrganisms())
#            ..getUcscSqlite(org,refdb,to)
#}
#
#..getUcscSqlite <- function(org,refdb,to) {
#    if (missing(to))
#        stop("Please provide a path to put the sqlite databases!")
#    if (!dir.exists(to))
#        dir.create(to,recursive=TRUE)
#    
#    message("RETRIEVING ",org," FROM ",refdb)
#    dbTmp <- getUcscDbl(org=org,refdb=refdb)
#    name <- paste(org,"_",refdb,".sqlite",sep="")
#    file.copy(dbTmp,file.path(to,name),recursive=TRUE)
#    return()
#}

.getUcscTabledefNcbi <- function(org,what="queries") {
    switch(org,
        hg18 = {
            warning("No NCBI RefSeq Genome annotation for Homo ",
                "sapiens hg18! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what)
            ))
        },
        hg19 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        hg38 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        mm9 = {
            warning("No NCBI RefSeq Genome annotation for Mus ",
                "musculus mm9! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        mm10 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        rn5 = {
            warning("No NCBI RefSeq Genome annotation for Rattus ",
                "norvegicus rn5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        rn6 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        dm3 = {
            warning("No NCBI RefSeq Genome annotation for Drosophila ",
                "melanogaster dm3! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        dm6 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        danrer7 = {
            warning("No NCBI RefSeq Genome annotation for Danio ",
                "rerio danrer7! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer10 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        danrer11 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        pantro4 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro4! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro5 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro6 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        },
        susscr3 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v3! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        susscr11 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v11! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        equcab2 = {
            warning("No NCBI RefSeq Genome annotation for Equus cabalus v2! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            ))
        },
        equcab3 = {
            return(list(
                ncbiRefSeq=.getUcscTblTpl("ncbiRefSeq",what)
            ))
        }
    )
}


.getUcscQueryNcbiGene <- function(org) {
    switch(org,
        hg18 = {
            warning("No NCBI RefSeq Genome annotation for Homo ",
                "sapiens hg18! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        hg38 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        mm9 = {
            warning("No NCBI RefSeq Genome annotation for Mus ",
                "musculus mm9! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        rn5 = {
            warning("No NCBI RefSeq Genome annotation for Rattus ",
                "norvegicus rn5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        dm3 = {
            warning("No NCBI RefSeq Genome annotation for Drosophila ",
                "melanogaster dm3! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm6 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        danrer7 = {
            warning("No NCBI RefSeq Genome annotation for Danio ",
                "rerio danrer7! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        danrer11 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        pantro4 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro4! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro6 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        susscr3 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v3! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v11! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab2 = {
            warning("No NCBI RefSeq Genome annotation for Equus cabalus v2! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `gene_id`,",
                "0 AS `gc_content`,",
                "`strand` AS `strand`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        }
    )
}

.getUcscQueryNcbiTranscript <- function(org) {
    switch(org,
        hg18 = {
            warning("No NCBI RefSeq Genome annotation for Homo ",
                "sapiens hg18! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "ORDER BY `chromosome`, `start`"))
        },
        hg19 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            warning("No NCBI RefSeq Genome annotation for Mus ",
                "musculus mm9! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            warning("No NCBI RefSeq Genome annotation for Rattus ",
                "norvegicus rn5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            warning("No NCBI RefSeq Genome annotation for Drosophila ",
                "melanogaster dm3! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer7 = {
            warning("No NCBI RefSeq Genome annotation for Danio ",
                "rerio danrer7! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer10 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer11 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro4 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro4! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro5 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro6 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr3 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v3! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr11 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v11! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab2 = {
            warning("No NCBI RefSeq Genome annotation for Equus cabalus v2! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab3 = {
            return(paste("SELECT `chrom` AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "`name` AS `transcript_id`,",
                "`strand` AS `strand`,",
                "name2 AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        }
    )
}

.getUcscQueryNcbiExon <- function(org) {
    switch(org,
        hg18 = {
            warning("No NCBI RefSeq Genome annotation for Homo ",
                "sapiens hg18! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        hg38 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        mm9 = {
            warning("No NCBI RefSeq Genome annotation for Mus ",
                "musculus mm9! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        rn5 = {
            warning("No NCBI RefSeq Genome annotation for Rattus ",
                "norvegicus rn5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        dm3 = {
            warning("No NCBI RefSeq Genome annotation for Drosophila ",
                "melanogaster dm3! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm6 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        danrer7 = {
            warning("No NCBI RefSeq Genome annotation for Danio ",
                "rerio danrer7! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        danrer11 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        pantro4 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro4! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            warning("No NCBI RefSeq Genome annotation for Pan troglodytes ",
                "pantro5! Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro6 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        },
        susscr3 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v3! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            warning("No NCBI RefSeq Genome annotation for Sus scrofa v11! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab2 = {
            warning("No NCBI RefSeq Genome annotation for Equus cabalus v2! ",
                "Will use UCSC RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            return(paste(
                "SELECT `chrom` AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "`name` AS `exon_id`,",
                "`strand` AS `strand`,",
                "`name` AS `gene_id`,",
                "`name2` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `ncbiRefSeq`", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`"
            ))
        }
    )
}

testKnownBuild <- function(org,refdb,ver=NULL,tv=FALSE) {
    if (missing(org)) {
        org <- "mm9"
        if (missing(refdb)) {
            refdb <- "ensembl"
            ver <- 67
        }
    }
    
    if (refdb=="ensembl") {
        if (!is.null(ver)) {
            organisms <- list(ver)
            names(organisms) <- org
        }
        else {
            organisms <- org
            names(organisms) <- org
        }
        sources <- refdb
    }
    else {
        organisms <- org
        names(organisms) <- org
        sources <- refdb
    }
    
    tmpdb <- tempfile()
    testResult <- logical(2)
    
    message("Scheduling 2 tests")
    message("==================================================\n")
    message("Running test 1 of 2 scheduled")
    tryCatch({
        addAnnotation(organisms,sources,db=tmpdb,versioned=tv,
            forceDownload=TRUE,rc=NULL)
        genes <- loadAnnotation(genome=names(organisms)[1],refdb=sources[1],
            type="gene",db=tmpdb)
        if (is(genes,"GRanges")) {
            message("Test 1 successful!")
            testResult[1] <- TRUE
        }
    },error=function(e) {
        message("Test 1 failed with error:")
        message(e$message)
    },finally="")
    
    message("\nRunning test 2 of 2 scheduled")
    tryCatch({
        org <- names(organisms)[1]
        refdb <- sources[1]
        n <- removeAnnotation(org,refdb,ver,db=tmpdb)
        if (n > 0) {
            message("Test 2 successful!")
            testResult[2] <- TRUE
        }
    },error=function(e) {
        message("Test 2 failed with error:")
        message(e$message)
    },finally="")
    
    message("Deleting temporary database ",tmpdb)
    unlink(tmpdb)
    
    return(testResult)
}

testCustomBuild <- function(gtf,metadata) {
    if (missing(gtf) || missing(metadata)) {
        gtf <- file.path(system.file(package="sitadela"),"dummy.gtf.gz")
        chromInfo <- data.frame(length=c(1000L,2000L,1500L),
            row.names=c("A","B","C"))
        metadata=list(
            organism="dummy",
            source="dummy_db",
            version=1,
            chromInfo=chromInfo
        )
    }
    
    tmpdb <- tempfile()
    testResult <- logical(2)
    
    message("Scheduling 2 tests")
    message("==================================================\n")
    message("Running test 1 of 2 scheduled")
    tryCatch({        
        addCustomAnnotation(gtfFile=gtf,metadata=metadata,db=tmpdb)
        genes <- loadAnnotation(genome=metadata$organism,refdb=metadata$source,
            type="gene",db=tmpdb)
        if (is(genes,"GRanges")) {
            message("Test 1 successful!")
            testResult[1] <- TRUE
        }
    },error=function(e) {
        message("Test 1 failed with error:")
        message(e$message)
    },finally="")
    
    message("\nRunning test 2 of 2 scheduled")
    tryCatch({
        n <- removeAnnotation(metadata$organism,metadata$source,db=tmpdb)
        if (n > 0) {
            message("Test 2 successful!")
            testResult[2] <- TRUE
        }
    },error=function(e) {
        message("Test 2 failed with error:")
        message(e$message)
    },finally="")
    
    message("Deleting temporary database ",tmpdb)
    unlink(tmpdb)
    
    return(testResult)
}

testUcscAll  <- function() {
    orgs <- getSupportedOrganisms()
    refdbs <- getSupportedUcscDbs()
    types <- c("gene","transcript","exon")
    
    # Remove tair10
    orgs <- orgs[-which(orgs=="tair10")]
    
    failNoVer <- testUcsc(orgs,refdbs,types)
    failVer <- testUcsc(orgs,refdbs,types,versioned=TRUE)
    return(list(
        failNoVersion=failNoVer,
        failVersion=failVer
    ))
}

testUcscUtrAll <- function() {
    orgs <- getSupportedOrganisms()
    refdbs <- getSupportedUcscDbs()
    
    # Remove tair10
    orgs <- orgs[-which(orgs=="tair10")]
    
    failNoVer <- testUcscUtr(orgs,refdbs)
    failVer <- testUcscUtr(orgs,refdbs,versioned=TRUE)
    return(list(
        failNoVersion=failNoVer,
        failVersion=failVer
    ))
}

testEnsembl <- function(level=c("normal","long","short"),versioned=FALSE) {
    level <- level[1]
    
    type <- c("gene","transcript","utr","exon","transexon")
    if (level == "normal") {
        org <- list(
            hg18=67,
            hg19=75,
            hg38=101:102,
            mm9=67,
            mm10=101:102,
            rn5=78:79,
            rn6=101:102,
            dm3=77:78,
            dm6=101:102,
            danrer7=79,
            danrer10=90:91,
            danrer11=101:102,
            pantro4=89:90,
            pantro5=101:102,
            susscr3=88:89,
            susscr11=101:102,
            equcab2=93:94,
            equcab3=101:102
        )
    }
    else if (level == "long") {
        org <- .ucscToEnsembl()
    }
    else if (level == "short") {
        org <- as.list(getSupportedOrganisms())
        names(org) <- unlist(org)
        org[seq_along(org)] <- "auto"
    }
    
    nTests <- length(type)*sum(lengths(org))
    message("Scheduling ",nTests," tests")
    message("==================================================\n")
    
    succ <- fail <- 0
    failReasons <- rep(NA,nTests)
    
    currTest <- 0
    for (o in names(org)) {
        for (v in org[[o]]) {
            if (v == "auto") v <- NULL
            for (z in type) {
                currTest <- currTest + 1
                message("\nRunning test ",currTest," of ",nTests," scheduled")
                message("Testing level ",z," for version ",v," from ",o)
                
                fr <- tryCatch({
                    ann <- getEnsemblAnnotation(o,z,v,versioned)
                    if (is(ann,"data.frame")) {
                        message("Test ",currTest," successful!")
                        message("Created ",nrow(ann)," features")
                        message("Sample data:")
                        print(head(ann))
                        succ <- succ + 1
                    }
                },error=function(e) {
                    message("Test ",currTest," failed with error:")
                    message(e$message)
                    return(paste("Test ",currTest,":",e$message,sep=""))
                },finally="")
                
                if (is.character(fr)) {
                    fail <- fail + 1
                    failReasons[currTest] <- fr
                }
            }
        }
    }
    
    message("Testing finished!\n")
    message("Summary")
    message("==================================================\n")
    message("Succesful tests: ",succ," out of ",nTests)
    message("Failed tests   : ",fail," out of ",nTests)
    message(" ")
    
    d <- which(is.na(failReasons))
    if (length(d) == nTests) # All successful
        return(NULL)
    else {
        message("Check the output for failure details")
        if (length(d) > 0) # If not, all failed!
            failReasons <- failReasons[-d]
        return(failReasons)
    }
}

testEnsemblSimple <- function(orgs,types,versioned=FALSE) {
    nTests <- length(orgs)*length(types)
    message("Scheduling ",nTests," tests")
    message("==================================================\n")
    
    succ <- fail <- 0
    failReasons <- rep(NA,nTests)
    v <- NULL
    
    currTest <- 0
    for (o in orgs) {
        for (z in types) {
            currTest <- currTest + 1
            message("\nRunning test ",currTest," of ",nTests," scheduled")
            message("Testing level ",z," from ",o," from latest version")
            
            fr <- tryCatch({
                ann <- getEnsemblAnnotation(o,z,v,versioned)
                if (is(ann,"data.frame")) {
                    message("Test ",currTest," successful!")
                    message("Created ",nrow(ann)," features")
                    message("Sample data:")
                    print(head(ann))
                    succ <- succ + 1
                }
            },error=function(e) {
                message("Test ",currTest," failed with error:")
                message(e$message)
                return(paste("Test ",currTest,": ",e$message,sep=""))
            },finally="")
            
            if (is.character(fr)) {
                fail <- fail + 1
                failReasons[currTest] <- fr
            }
        }
    }
    
    message("Testing finished!\n")
    message("Summary")
    message("==================================================\n")
    message("Succesful tests: ",succ," out of ",nTests)
    message("Failed tests   : ",fail," out of ",nTests)
    message(" ")
    
    d <- which(is.na(failReasons))
    if (length(d) == nTests) # All successful
        return(NULL)
    else {
        message("Check the output for failure details")
        if (length(d) > 0) # If not, all failed!
            failReasons <- failReasons[-d]
        return(failReasons)
    }
}

testUcsc <- function(orgs,refdbs,types,versioned=FALSE) {
    if (.Platform$OS.type != "unix") {
        message("Cannot run 3' UTR tests from UCSC on non Unix/Linux machines!")
        return(NULL)
    }
    
    nTests <- length(orgs)*length(refdbs)*length(types)
    message("Scheduling ",nTests," tests")
    message("==================================================\n")
    
    succ <- fail <- 0
    failReasons <- rep(NA,nTests)
    
    if ("tair10" %in% orgs)
        orgs <- orgs[-which(orgs=="tair10")]
    
    dbCreds <- .getUcscCredentials()
    drv <- dbDriver("MySQL")
    currTest <- 0
    
    for (o in orgs) {
        message("==========> Now testing ",getUcscOrganism(o))
        message("Connecting to UCSC database ",getUcscOrganism(o),"...\n")
        con <- dbConnect(drv,user=dbCreds[2],password=NULL,
            dbname=getUcscOrganism(o),host=dbCreds[1])
        for (r in refdbs) {
            for (t in types) {
                currTest <- currTest + 1
                message("---------------------------------------------------")
                message("Running test ",currTest," of ",nTests," scheduled")
                message("Source: ", r)
                message("Type  : ", t)
                query <- .getUcscQuery(org=o,type=t,refdb=r,versioned=versioned)
                message("Query is: ",query)
                fr <- tryCatch({
                    dat <- dbGetQuery(con,query)
                    if (is.data.frame(dat)) {
                        message("Test ",currTest," successful!")
                        message("Fetched ",nrow(dat)," features")
                        message("Sample data:")
                        print(head(dat))
                        succ <- succ + 1
                    }                    
                },error=function(e) {
                    message("Test ",currTest," failed with error:")
                    message(e$message)
                    return(paste("Test ",currTest,":",e$message,sep=""))
                },finally="")
                message("---------------------------------------------------\n")
                if (is.character(fr)) {
                    fail <- fail + 1
                    failReasons[currTest] <- fr
                }
            }
        }
        message("Disconnecting from UCSC database...")
        dbDisconnect(con)
    }
    
    message("Testing finished!\n")
    message("Summary")
    message("==================================================\n")
    message("Succesful tests: ",succ," out of ",nTests)
    message("Failed tests   : ",fail," out of ",nTests)
    message(" ")
    
    d <- which(is.na(failReasons))
    if (length(d) == nTests) # All successful
        return(NULL)
    else {
        message("Check the output for failure details")
        if (length(d) > 0) # If not, all failed!
            failReasons <- failReasons[-d]
        return(failReasons)
    }
}

testUcscUtr <- function(orgs,refdbs,versioned=FALSE) {
    if (.Platform$OS.type != "unix") {
        message("Cannot run 3' UTR tests from UCSC on non Unix/Linux machines!")
        return(NULL)
    }
    
    nTests <- length(orgs)*length(refdbs)
    message("Scheduling ",nTests," tests")
    message("==================================================\n")
    
    succ <- fail <- 0
    failReasons <- rep(NA,nTests)
    
    if ("tair10" %in% orgs)
        orgs <- orgs[-which(orgs=="tair10")]
    
    currTest <- 0
    for (o in orgs) {
        message("==========> Now testing ",getUcscOrganism(o))
        for (r in refdbs) {
            currTest <- currTest + 1
            message("---------------------------------------------------")
            message("Running test ",currTest," of ",nTests," scheduled")
            message("Source: ", r)
            message("Type  : utr")
            # For logging only
            if (r=="refseq" && versioned) {
                query <- .getUcscRefseqVersionedUtrQuery()
                message("Query is: ",query)
            }
            fr <- tryCatch({
                dat <- getUcscUtr(o,r,versioned,.rmysql=TRUE)
                if (is(dat,"GRanges")) {
                    message("Test ",currTest," successful!")
                    message("Fetched ",length(dat)," features")
                    message("Sample data:")
                    print(head(dat))
                    succ <- succ + 1
                }                    
            },error=function(e) {
                message("Test ",currTest," failed with error:")
                message(e$message)
                return(paste("Test ",currTest,":",e$message,sep=""))
            },finally="")
            message("---------------------------------------------------\n")
            if (is.character(fr)) {
                fail <- fail + 1
                failReasons[currTest] <- fr
            }
        }
    }
    
    message("Testing finished!\n")
    message("Summary")
    message("==================================================\n")
    message("Succesful tests: ",succ," out of ",nTests)
    message("Failed tests   : ",fail," out of ",nTests)
    
    d <- which(is.na(failReasons))
    if (length(d) == nTests) # All successful
        return(NULL)
    else {
        message("Check the output for failure details")
        if (length(d) > 0) # If not, all failed!
            failReasons <- failReasons[-d]
        return(failReasons)
    }
}


testCustomGtf <- function(gtf) {
    message("Custom GTF file is ",gtf)
    message("==================================================")
    
    parsed <- parseCustomGtf(gtf)
    txdb <- parsed$txdb
    map <- parsed$map
    
    nTests <- 12
    
    message("\nScheduling ",nTests," tests")
    message("==================================================")
    
    succ <- fail <- 0
    failReasons <- rep(NA,nTests)
    
    currTest <- 0
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing gene gene from GTF")
    ann <- tryCatch({
        .makeGeneGeneFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing gene exon from GTF")
    ann <- tryCatch({
        .makeGeneExonFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing summarized gene exon from GTF")
    ann <- tryCatch({
        .makeSumGeneExonFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing gene utr from GTF")
    ann <- tryCatch({
        .makeGeneUtrFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing summarized gene utr from GTF")
    ann <- tryCatch({
        .makeSumGeneUtrFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing transcript gene from GTF")
    ann <- tryCatch({
        .makeTranscriptGeneFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing summarized transcript gene from GTF")
    ann <- tryCatch({
        .makeSumTranscriptGeneFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing transcript exon from GTF")
    ann <- tryCatch({
        .makeTranscriptExonFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing summarized transcript exon from GTF")
    ann <- tryCatch({
        .makeSumTranscriptExonFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing transcript utr from GTF")
    ann <- tryCatch({
        .makeTranscriptUtrFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing summarized transcript utr from GTF")
    ann <- tryCatch({
        .makeSumTranscriptUtrFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    currTest <- currTest + 1
    message("\nRunning test ",currTest," of ",nTests," scheduled")
    message("Testing exon exon from GTF")
    ann <- tryCatch({
        .makeExonExonFromTxDb(txdb,map,FALSE)
    },error=function(e) {
        return(paste("Test ",currTest,":",e$message,sep=""))
    },finally="")
    if (is(ann,"GRanges")) {
        message("Test ",currTest," successful!")
        message("Created ",length(ann)," features")
        message("Sample data:")
        print(head(ann))
        succ <- succ + 1
    }
    else if (is.character(ann)) {
        message("Test ",currTest," failed with error:")
        message(ann)
        fail <- fail + 1
        failReasons[currTest] <- ann
    }
    
    message("Testing finished!\n")
    message("Summary")
    message("==================================================\n")
    message("Succesful tests: ",succ," out of ",nTests)
    message("Failed tests   : ",fail," out of ",nTests)
    message(" ")
    
    d <- which(is.na(failReasons))
    if (length(d) == nTests) # All successful
        return(NULL)
    else {
        message("Check the output for failure details")
        if (length(d) > 0) # If not, all failed!
            failReasons <- failReasons[-d]
        return(failReasons)
    }
}

#
#testQuery <- function(query,org) {
#    dbCreds <- .getUcscCredentials()
#    drv <- dbDriver("MySQL")
#    con <- dbConnect(drv,user=dbCreds[2],password=NULL,dbname=org,
#        host=dbCreds[1])
#    rawAnn <- tryCatch(dbGetQuery(con,query),
#        error=function(e) print(e),
#        finally=dbDisconnect(con))
#    return(rawAnn)
#}
#

## ----install-0, eval=FALSE, echo=TRUE-----------------------------------------
# if(!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# BiocManager::install("sitadela")

## ----load-0, eval=TRUE, echo=FALSE, tidy=FALSE, message=FALSE, warning=FALSE----
library(sitadela)

## ----example-1, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
library(sitadela)

buildDir <- file.path(tempdir(),"test_anndb")
dir.create(buildDir)

# The location of the custom database
myDb <- file.path(buildDir,"testann.sqlite")

# Since we are using Ensembl, we can also ask for a version
organisms <- list(mm10=100)
sources <- ifelse(.Platform$OS.type=="unix",c("ensembl","refseq"),"ensembl")

# If the example is not running in a multicore system, rc is ignored
addAnnotation(organisms,sources,forceDownload=FALSE,db=myDb,rc=0.5)

## Alternatively
# setDbPath(myDb)
# addAnnotation(organisms,sources,forceDownload=FALSE,rc=0.5)

## ----example-2, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# Load standard annotation based on gene body coordinates
genes <- loadAnnotation(genome="mm10",refdb="ensembl",type="gene",db=myDb)
genes

# Load standard annotation based on 3' UTR coordinates
utrs <- loadAnnotation(genome="mm10",refdb="ensembl",type="utr",db=myDb)
utrs

# Load summarized exon annotation based used with RNA-Seq analysis
sumEx <- loadAnnotation(genome="mm10",refdb="ensembl",type="exon",
    summarized=TRUE,db=myDb)
sumEx

## Load standard annotation based on gene body coordinates from RefSeq
#if (.Platform$OS.type=="unix") {
#    refGenes <- loadAnnotation(genome="mm10",refdb="refseq",type="gene",
#        db=myDb)
#    refGenes
#}

## ----example-3, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# Load standard annotation based on gene body coordinates
genes <- loadAnnotation(genome="mm10",refdb="ensembl",type="gene",db=myDb,
    asdf=TRUE)
head(genes)

## ----example-4, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
gtf <- system.file(package="sitadela","extdata",
    "gadMor1_HE567025.gtf.gz")
chrom <- system.file(package="sitadela","extdata",
    "gadMor1_HE567025.txt.gz")
chromInfo <- read.delim(chrom,header=FALSE,row.names=1)
names(chromInfo) <- "length"
metadata <- list(
    organism="gadMor1_HE567025",
    source="sitadela_package",
    chromInfo=chromInfo
)
tmpdb <- tempfile()

addCustomAnnotation(gtfFile=gtf,metadata=metadata,db=tmpdb)

# Try to retrieve some data
g <- loadAnnotation(genome="gadMor1_HE567025",refdb="sitadela_package",
    type="gene",db=tmpdb)
g

# Delete the temporary database
unlink(tmpdb)

## ----example-5, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
gtf <- system.file(package="sitadela","extdata",
    "eboVir3_KM034562v1.gtf.gz")
chrom <- system.file(package="sitadela","extdata",
    "eboVir3_KM034562v1.txt.gz")
chromInfo <- read.delim(chrom,header=FALSE,row.names=1)
names(chromInfo) <- "length"
metadata <- list(
    organism="gadMor1_HE567025",
    source="sitadela_package",
    chromInfo=chromInfo
)
tmpdb <- tempfile()

addCustomAnnotation(gtfFile=gtf,metadata=metadata,db=tmpdb)

# Try to retrieve some data
g <- loadAnnotation(genome="gadMor1_HE567025",refdb="sitadela_package",
    type="gene",db=tmpdb)
g

# Delete the temporary database
unlink(tmpdb)

## ----example-5-1, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# # Setup a temporary directory to download files etc.
# customDir <- file.path(tempdir(),"test_custom")
# dir.create(customDir)
# 
# # Convert from GenePred to GTF - Unix/Linux only!
# if (.Platform$OS.type == "unix" && !grepl("^darwin",R.version$os)) {
#     # Download data from UCSC
#     goldenPath="http://hgdownload.cse.ucsc.edu/goldenPath/"
#     # Gene annotation dump
#     download.file(paste0(goldenPath,"eboVir3/database/ncbiGene.txt.gz"),
#         file.path(customDir,"eboVir3_ncbiGene.txt.gz"))
#     # Chromosome information
#     download.file(paste0(goldenPath,"eboVir3/database/chromInfo.txt.gz"),
#         file.path(customDir,"eboVir3_chromInfo.txt.gz"))
# 
#     # Prepare the build
#     chromInfo <- read.delim(file.path(customDir,"eboVir3_chromInfo.txt.gz"),
#         header=FALSE)
#     chromInfo <- chromInfo[,1:2]
#     rownames(chromInfo) <- as.character(chromInfo[,1])
#     chromInfo <- chromInfo[,2,drop=FALSE]
# 
#     # Coversion from genePred to GTF
#     genePredToGtf <- file.path(customDir,"genePredToGtf")
#     if (!file.exists(genePredToGtf)) {
#         download.file(
#         "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
#             genePredToGtf
#         )
#         system(paste("chmod 775",genePredToGtf))
#     }
#     gtfFile <- file.path(customDir,"eboVir3.gtf")
#     tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
#         "tgtf",sep="."))
#     command <- paste0(
#         "zcat ",file.path(customDir,"eboVir3_ncbiGene.txt.gz"),
#         " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
#         " -source=eboVir3"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
#         gtfFile
#     )
#     system(command)
# 
#     # Build with the metadata list filled (you can also provide a version)
#     addCustomAnnotation(
#         gtfFile=gtfFile,
#         metadata=list(
#             organism="eboVir3_test",
#             source="ucsc_test",
#             chromInfo=chromInfo
#         ),
#         db=myDb
#     )
# 
#     # Try to retrieve some data
#     eboGenes <- loadAnnotation(genome="eboVir3_test",refdb="ucsc_test",
#         type="gene",db=myDb)
#     eboGenes
# }

## ----example-6, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
gtfFile <- system.file(package="sitadela","extdata",
    "gadMor1_HE567025.gtf.gz")
chromInfo <- read.delim(system.file(package="sitadela","extdata",
    "gadMor1_HE567025.txt.gz"),header=FALSE)

# Build with the metadata list filled (you can also provide a version)
addCustomAnnotation(
    gtfFile=gtfFile,
    metadata=list(
        organism="gadMor1_test",
        source="ucsc_test",
        chromInfo=chromInfo
    ),
    db=myDb
)

# Try to retrieve some data
gadGenes <- loadAnnotation(genome="gadMor1_test",refdb="ucsc_test",
    type="gene",db=myDb)
gadGenes

## ----example-7, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
gtfFile <- system.file(package="sitadela","extdata",
    "dasNov3_JH569334.gtf.gz")
chromInfo <- read.delim(system.file(package="sitadela",
    "extdata","dasNov3_JH569334.txt.gz"),header=FALSE)

# Build with the metadata list filled (you can also provide a version)
addCustomAnnotation(
    gtfFile=gtfFile,
    metadata=list(
        organism="dasNov3_test",
        source="ensembl_test",
        chromInfo=chromInfo
    ),
    db=myDb
)

# Try to retrieve some data
dasGenes <- loadAnnotation(genome="dasNov3_test",refdb="ensembl_test",
    type="gene",db=myDb)
dasGenes

## ----example-8, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# organisms <- list(
#     hg18=54,
#     hg19=75,
#     hg38=110:111,
#     mm9=54,
#     mm10=110:111,
#     rn5=77,
#     rn6=110:111,
#     dm3=77,
#     dm6=110:111,
#     danrer7=77,
#     danrer10=80,
#     danrer11=110:111,
#     pantro4=80,
#     pantro5=110:111,
#     susscr3=80,
#     susscr11=110:111,
#     equcab2=110:111
# )
# 
# sources <- c("ensembl","ucsc","refseq","ncbi")
# 
# addAnnotation(organisms,sources,forceDownload=FALSE,rc=0.5)

## ----pseudo-1, eval=TRUE, echo=TRUE, message=TRUE, warning=FALSE--------------
metadata <- list(
    organism="ORGANISM_NAME",
    source="SOURCE_NAME",
    chromInfo="CHROM_INFO"
)

## ----si-1, eval=TRUE, echo=TRUE-----------------------------------------------
sessionInfo()

.getUcscTabledefRefseq <- function(org,what="queries",versioned=FALSE) {
    switch(org,
        hg18 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what)
            )
        },
        hg19 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        hg38 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what)
            )
        },
        mm9 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        mm10 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    .getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    .getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    .getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        rn5 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        rn6 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        dm3 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        dm6 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        danrer7 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        danrer10 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        pantro4 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        pantro5 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        #pantro6 = {
        #    def <- list(
        #        refFlat=.getUcscTblTpl("refFlat",what),
        #        ensemblToGeneName=
        #            .getUcscTblTpl("ensemblToGeneName",what),
        #        ensemblSource=
        #            .getUcscTblTpl("ensemblSource",what)
        #    )
        #},
        susscr3 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        susscr11 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        equcab2 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        },
        equcab3 = {
            def <- list(
                refFlat=.getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    .getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    .getUcscTblTpl("ensemblSource",what)
            )
        }
    )
    
    if (versioned)
        def <- c(def,gbCdnaInfo=.getUcscTblTpl("gbCdnaInfo",what))
    
    return(def)
}

.getUcscQueryRefseqGene <- function(org,versioned=FALSE) {
    switch(org,
        hg18 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab2 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscQueryRefseqTranscript <- function(org,versioned=FALSE) {
    switch(org,
        hg18 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "ORDER BY `chromosome`, `start`"))
        },
        hg19 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "ORDER BY `chromosome`, `start`"))
        },
        mm9 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "ORDER BY `chromosome`, `start`"))
        },
        rn5 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer7 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer10 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer11 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro4 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro5 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr3 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr11 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab2 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab3 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `transcript_id`,")
                else
                    "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        }
    )
}

.getUcscQueryRefseqExon <- function(org,versioned=FALSE) {
    switch(org,
        hg18 = {
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab2 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `exon_id`,")
                else
                    "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                if (versioned)
                    paste0("CONCAT(refFlat.name,'.',",
                        "hgFixed.gbCdnaInfo.version) AS `gene_id`,")
                else
                    "refFlat.name AS `gene_id`,",
                "refFlat.geneName AS `gene_name`,",
                "ensemblSource.source AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                if (versioned)
                    paste("INNER JOIN hgFixed.gbCdnaInfo",
                        "ON refFlat.name=hgFixed.gbCdnaInfo.acc")
                else "",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscRefseqVersionedUtrQuery <- function() {
    return(paste(
        "SELECT refGene.bin AS `bin`,",
        paste0("CONCAT(refGene.name,'.',",
            "hgFixed.gbCdnaInfo.version) AS `name`,"),
        "refGene.chrom AS `chrom`,",
        "refGene.strand AS `strand`,",
        "refGene.txStart AS `txStart`,",
        "refGene.txEnd AS `txEnd`,",
        "refGene.cdsStart AS `cdsStart`,",
        "refGene.cdsEnd AS `cdsEnd`,",
        "refGene.exonCount AS `exonCount`,",
        "refGene.exonStarts AS `exonStarts`,",
        "refGene.exonEnds AS `exonEnds`,",
        "refGene.score AS `score`,",
        "refGene.name2 AS `name2`,",
        "refGene.cdsStartStat AS `cdsStartStat`,",
        "refGene.cdsEndStat AS `cdsEndStat`,",
        "refGene.exonFrames AS `exonFrames`",
        "FROM `refGene` INNER JOIN hgFixed.gbCdnaInfo",
        "ON refGene.name=hgFixed.gbCdnaInfo.acc",
        "ORDER BY `chrom`,`txStart`"
    ))
}

.checkMainArgs <- function(mainArgs) {
    inArgs <- names(mainArgs)[-1] # 1st member name of calling function
    validArgs <- .getValidArgs()
    invalid <- setdiff(inArgs,validArgs)
    if (length(invalid) > 0) {
        for (i in seq_len(length(invalid)))
            warning("Unknown input argument to sitadela: ",invalid[i],
                " ...Ignoring...",immediate.=TRUE)
    }
}

.checkTextArgs <- function(argName,argValue,argList,multiarg=FALSE) {
    if (multiarg) {
        argValue <- tolower(argValue)
        if (!all(argValue %in% argList))
            stop("\"",argName,"\""," parameter must be one or more of ",
                paste(paste("\"",argList,sep=""),collapse="\", "),"\"!")
    }
    else {
        argSave <- argValue[1]
        argValue <- tolower(argValue[1])    
        if (!(argValue %in% argList))
            stop("\"",argName,"\""," parameter must be one of ",
                paste(paste("\"",argList,sep=""),collapse="\", "),"\"!")
    }
}

.checkNumArgs <- function(argName,argValue,argType,argBounds,direction) {
    # First generic check so not to continue if fail
    if (!is(argValue,argType))
        stop("\"",argName,"\" parameter must be a(n) ",argType," value!")
    
    # Then, proceed with a lookup table to avoid repetition (suggested by
    # Marcel Ramos during package review)
    lookup <- list(
        both=list(
            fail=function(x) x<argBounds[1] || x>argBounds[2],
            cls=class,
            msg=function(x) paste0("larger than or equal to ",
                argBounds[1]," and smaller than or equal to ",
                argBounds[2])
        ),
        botheq=list(
            fail=function(x) x<=argBounds[1] || x>=argBounds[2],
            cls=class,
            msg=function(x) paste0("larger than ",argBounds[1],
                " and smaller than ",argBounds[2])
        ),
        gt=list(
            fail=function(x) x<=argBounds[1], 
            cls=class,
            msg=function(x) paste0("greater than ",argBounds[1])
        ),
        lt=list(
            fail=function(x) x>=argBounds[1], 
            cls=class,
            msg=function(x) paste0("lower than ",argBounds[1])
        ),
        gte=list(
            fail=function(x) x<argBounds[1], 
            cls=class,
            msg=function(x) paste0("greater than or equal to ",
                argBounds[1])
        ),
        lte=list(
            fail=function(x) x>argBounds[1], 
            cls=class,
            msg=function(x) paste0("lower than or equal to ",
                argBounds[1])
        )
    )
    
    check <- lapply(lookup[[direction]],function(f) f(argValue))
    if (check$fail || check$cls != argType)
        stop("\"",argName,"\""," parameter must be a(n) ",argType,
            " value ",check$msg,"!")
}

.getValidArgs <- function() {
    return(c("organisms","sources","db","versioned","forceDownload","rc"))
}

.bypassTimeoutByFilters <- function(org,type,ver,tv,mart) {
    filter <- .getFilter(org,type,ver,tv)
    filterValues <- getBM(attributes=filter,mart=mart)
    return(getBM(attributes=.getAttributes(org,type,ver,tv),filters=filter,
        values=filterValues,mart=mart))
}

.getFilter <- function(org,type,ver=NULL,tv=FALSE) {
    ver <- .checkEnsVer(ver,org)
    if (type == "gene") {
        if (tv)
            return(.getVersionedFilter(org,"gene",ver))
        else
            return("ensembl_gene_id")
    }
    else if (type == "exon")
        return("ensembl_exon_id")
    else if (type %in% c("transcript","utr","transexon")) {
        if (tv)
            return(.getVersionedFilter(org,"transcript",ver))
        else
            return("ensembl_transcript_id")
    }
}

.getVersionedFilter <- function(org,type,ver) {
    if (org %in% .orgsWithNoVersion())
        return(paste0("ensembl_",type,"_id"))
    else if (org %in% .orgsWithVersionAfter90()) {
        if (ver < 90)
            return(paste0("ensembl_",type,"_id"))
        else
            return(paste0("ensembl_",type,"_id_version"))
    }
    else if (org %in% .orgsWithVersion())
        return(paste0("ensembl_",type,"_id_version"))
}

.getAttributes <- function(org,type,ver=NULL,tv=FALSE) {
    ver <- .checkEnsVer(ver,org)
    switch(type,
        gene = {
            if (tv)
                return(.getVersionedGeneAttributes(org,ver))
            else
                return(.getGeneAttributes(org))
        },
        transcript = {
            if (tv)
                return(.getVersionedTranscriptAttributes(org,ver))
            else
                return(.getTranscriptAttributes(org))
        },
        exon = {
            if (tv)
                return(.getVersionedExonAttributes(org,ver))
            else
                return(.getExonAttributes(org))
        },
        utr = {
            if (tv)
                return(.getVersionedTranscriptUtrAttributes(org,ver))
            else
                return(.getTranscriptUtrAttributes(org))
        },
        transexon = {
            if (tv)
                return(.getVersionedTranscriptExonAttributes(org,ver))
            else
                return(.getTranscriptExonAttributes(org))
        }
    )
}

.getGeneAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "start_position",
            "end_position",
            "ensembl_gene_id",
            "percentage_gc_content",
            "strand",
            "external_gene_id",
            "gene_biotype"
        ))
    else if (org %in% c("rn5","danrer7","dm3"))
        return(c(
            "chromosome_name",
            "start_position",
            "end_position",
            "ensembl_gene_id",
            "percentage_gc_content",
            "strand",
            "external_gene_name",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "start_position",
            "end_position",
            "ensembl_gene_id",
            "percentage_gene_gc_content",
            "strand",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getVersionedGeneAttributes <- function(org,ver) {
    if (org %in% .orgsWithNoVersion())
        return(.getGeneAttributes(org))
    else if (org %in% .orgsWithVersionAfter90()) {
        if (ver < 90)
            return(.getGeneAttributes(org))
        else
            return(c(
                "chromosome_name",
                "start_position",
                "end_position",
                "ensembl_gene_id_version",
                "percentage_gene_gc_content",
                "strand",
                "external_gene_name",
                "gene_biotype"
            ))
    }
    else if (org %in% .orgsWithVersion())
        return(c(
                "chromosome_name",
                "start_position",
                "end_position",
                "ensembl_gene_id_version",
                "percentage_gene_gc_content",
                "strand",
                "external_gene_name",
                "gene_biotype"
            ))
}

.getTranscriptAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getVersionedTranscriptAttributes <- function(org,ver) {
    if (org %in% .orgsWithNoVersion())
        return(.getTranscriptAttributes(org))
    else if (org %in% .orgsWithVersionAfter90()) {
        if (ver < 90)
            return(.getTranscriptAttributes(org))
        else
            return(c(
                "chromosome_name",
                "transcript_start",
                "transcript_end",
                "ensembl_transcript_id_version",
                "strand",
                "ensembl_gene_id_version",
                "external_gene_name",
                "gene_biotype"
            ))
    }
    else if (org %in% .orgsWithVersion())
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "ensembl_transcript_id_version",
            "strand",
            "ensembl_gene_id_version",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getTranscriptUtrAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "3_utr_start",
            "3_utr_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "3_utr_start",
            "3_utr_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getVersionedTranscriptUtrAttributes <- function(org,ver) {
    if (org %in% .orgsWithNoVersion())
        return(.getTranscriptUtrAttributes(org))
    else if (org %in% .orgsWithVersionAfter90()) {
        if (ver < 90)
            return(.getTranscriptUtrAttributes(org))
        else
            return(c(
                "chromosome_name",
                "transcript_start",
                "transcript_end",
                "3_utr_start",
                "3_utr_end",
                "ensembl_transcript_id_version",
                "strand",
                "ensembl_gene_id_version",
                "external_gene_name",
                "gene_biotype"
            ))
    }
    else if (org %in% .orgsWithVersion())
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "3_utr_start",
            "3_utr_end",
            "ensembl_transcript_id_version",
            "strand",
            "ensembl_gene_id_version",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getExonAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getVersionedExonAttributes <- function(org,ver) {
    if (org %in% .orgsWithNoVersion())
        return(.getExonAttributes(org))
    else if (org %in% .orgsWithVersionAfter90()) {
        if (ver < 90)
            return(.getExonAttributes(org))
        else
            return(c(
                "chromosome_name",
                "exon_chrom_start",
                "exon_chrom_end",
                "ensembl_exon_id",
                "strand",
                "ensembl_gene_id_version",
                "external_gene_name",
                "gene_biotype"
            ))
    }
    else if (org %in% .orgsWithVersion())
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_gene_id_version",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getTranscriptExonAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_transcript_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_transcript_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getVersionedTranscriptExonAttributes <- function(org,ver) {
    if (org %in% .orgsWithNoVersion())
        return(.getTranscriptExonAttributes(org))
    else if (org %in% .orgsWithVersionAfter90()) {
        if (ver < 90)
            return(.getTranscriptExonAttributes(org))
        else
            return(c(
                "chromosome_name",
                "exon_chrom_start",
                "exon_chrom_end",
                "ensembl_exon_id",
                "strand",
                "ensembl_transcript_id_version",
                "external_gene_name",
                "gene_biotype"
            ))
    }
    else if (org %in% .orgsWithVersion())
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_transcript_id_version",
            "external_gene_name",
            "gene_biotype"
        ))
}

.ucscToEnsembl <- function() {
    return(list(
        hg18=54,
        hg19=75,
        hg38=95:111,
        mm9=54,
        mm10=95:111,
        rn5=77,
        rn6=c(80,95:111),
        dm3=c(75,77),
        dm6=c(80,95:111),
        danrer7=c(75,77),
        danrer10=80,
        danrer11=95:111,
        pantro4=c(75,77,80),
        pantro5=95:111,
        #pantro6=,
        susscr3=c(75,77,80),
        susscr11=95:111,
        equcab2=c(75,77),
        equcab3=95:111
    ))
}

.getHost <- function(org,ver=NULL) {
    if (!requireNamespace("biomaRt"))
        stop("The Bioconductor package biomaRt is required to proceed!")
    
    org <- tolower(org[1])
    .checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    if (!is.null(ver) 
        && (!is.numeric(ver) || is.na(suppressWarnings(as.numeric(ver)))))
        stop("ver must be numeric or coercible to numeric if not NULL!")
        
    if (org == "tair10")
        return("plants.ensembl.org")
    
    aver <- .getUcscToEnsembl(org)
    if (!is.null(ver) && !(ver %in% aver)) {
        warning("Version ",ver," not available/existing for ",org,"! Will ",
            "use the latest available version...",immediate.=TRUE)
        ver <- NULL
    }
    
    if (is.null(ver)) {
        u2e <- .ucscToEnsembl()
        vers <- u2e[[org]]
        ver <- vers[length(vers)]
    }
    
    ea <- biomaRt::listEnsemblArchives()
    i <- grep(as.character(ver),ea[,"version"])
    if (length(i) > 0) {
        if (ea[i,"current_release"] == "*")
            return("https://www.ensembl.org")
        else
            return(ea[i,"url"])
    }
    else {
        warning("Version ",ver," not found in biomaRt archives for ",org,"! ",
            "Will use the latest available version for ",org,"...",
            immediate.=TRUE)
        u2e <- .ucscToEnsembl()
        vers <- u2e[[org]]
        for (v in rev(vers)) {
            newi <- grep(as.character(v),ea[,"version"])
            if (length(newi) > 0)
                return(ea[newi,"url"])
        }
    }
    # If everything fails...
    return(NULL)
}

.getDataset <- function(org) {
    switch(org,
        hg18 = { return("hsapiens_gene_ensembl") },
        hg19 = { return("hsapiens_gene_ensembl") },
        hg38 = { return("hsapiens_gene_ensembl") },
        mm9 = { return("mmusculus_gene_ensembl") },
        mm10 = { return("mmusculus_gene_ensembl") },
        rn5 = { return("rnorvegicus_gene_ensembl") },
        rn6 = { return("rnorvegicus_gene_ensembl") },
        dm3 = { return("dmelanogaster_gene_ensembl") },
        dm6 = { return("dmelanogaster_gene_ensembl") },
        danrer7 = { return("drerio_gene_ensembl") },
        danrer10 = { return("drerio_gene_ensembl") },
        danrer11 = { return("drerio_gene_ensembl") },
        pantro4 = { return("ptroglodytes_gene_ensembl") },
        pantro5 = { return("ptroglodytes_gene_ensembl") },
        #pantro6 = { return("ptroglodytes_gene_ensembl") },
        susscr3 = { return("sscrofa_gene_ensembl") },
        susscr11 = { return("sscrofa_gene_ensembl") },
        equcab2 = { return("ecaballus_gene_ensembl") },
        equcab3 = { return("ecaballus_gene_ensembl") },
        tair10 = { return("athaliana_eg_gene") }
    )
}

.validateEnsemblVersions <- function(o,v) {
    ea <- biomaRt::listEnsemblArchives()
    found <- as.character(v) %in% ea[,"version"]
    if (any(!found)) {
        nf <- which(!found)
        warning("Version(s) ",paste0(v[nf],collapse=", ")," not found in ",
            "biomaRt archives for current organism!\nWill use the latest ",
            "available version(s)...",immediate.=TRUE)
        newv <- v[-nf]
        if (length(newv) == 0) {
            warning("No Ensembl versions left after validation! Will return ",
                "only the latest available...",immediate.=TRUE)
            u2e <- .ucscToEnsembl()
            vers <- u2e[[o]]
            for (newv in rev(vers)) {
                newi <- grep(as.character(newv),ea[,"version"])
                if (length(newi) > 0)
                    return(newv)
            }
        }
        return(newv)
    }
    return(v)
}

.getUcscToEnsembl <- function(org) {
    u2e <- .ucscToEnsembl()
    return(u2e[[org]])
}

.checkUcscToEnsembl <- function(org,ver) {
    u2e <- .getUcscToEnsembl()
    return(ver %in% u2e[[org]])
}

.orgsWithNoVersion <- function() {
    return(c("hg18","hg19","mm9","rn5","dm3","dm6","pantro4","danrer7",
        "danrer10","susscr3","equcab2","tair10"))
}

.orgsWithVersionAfter90 <- function() {
    return(c("hg38","mm10","rn6","pantro5","susscr11"))
}

.orgsWithVersion <- function() {
    return(c("danrer11","equcab3"))
}

.checkEnsVer <- function(ver,org) {
    if (is.null(ver)) {
        u2e <- .ucscToEnsembl()
        vers <- u2e[[org]]
        ver <- vers[length(vers)]
    }
    return(ver)
}

.getBiotypes <- function(org) {
    if (!(org %in% getSupportedOrganisms()))
        return(NULL)
    switch(org,
        hg18 = {
            return(c("unprocessed_pseudogene","pseudogene","miRNA",
                "retrotransposed","protein_coding","processed_pseudogene",
                "snRNA","snRNA_pseudogene","Mt_tRNA_pseudogene",
                "miRNA_pseudogene","misc_RNA","tRNA_pseudogene","snoRNA",
                "scRNA_pseudogene","rRNA_pseudogene","snoRNA_pseudogene","rRNA",
                "misc_RNA_pseudogene","IG_V_gene","IG_D_gene","IG_J_gene",
                "IG_C_gene","IG_pseudogene","scRNA"))
        },
        hg19 = {
            return(c("pseudogene","lincRNA","protein_coding","antisense",
                "processed_transcript","snRNA","sense_intronic","miRNA",
                "misc_RNA","snoRNA","rRNA","polymorphic_pseudogene",
                "sense_overlapping","3prime_overlapping_ncrna","TR_V_gene",
                "TR_V_pseudogene","TR_D_gene","TR_J_gene","TR_C_gene",
                "TR_J_pseudogene","IG_C_gene","IG_C_pseudogene","IG_J_gene",
                "IG_J_pseudogene","IG_D_gene","IG_V_gene","IG_V_pseudogene"))
        },
        hg38 = {
            return(c("protein_coding","polymorphic_pseudogene","lincRNA",
                "unprocessed_pseudogene","processed_pseudogene","antisense",
                "processed_transcript","transcribed_unprocessed_pseudogene",
                "sense_intronic","unitary_pseudogene","IG_V_gene",
                "IG_V_pseudogene","TR_V_gene","sense_overlapping",
                "transcribed_processed_pseudogene","miRNA","snRNA","misc_RNA",
                "rRNA","snoRNA","IG_J_pseudogene","IG_J_gene","IG_D_gene",
                "3prime_overlapping_ncrna","IG_C_gene","IG_C_pseudogene",
                "pseudogene","TR_V_pseudogene","Mt_tRNA","Mt_rRNA",
                "translated_processed_pseudogene","TR_J_gene","TR_C_gene",
                "TR_D_gene","TR_J_pseudogene","LRG_gene"))
        },
        mm9 = {
            return(c("pseudogene","snRNA","protein_coding","antisense","miRNA",
                "lincRNA","snoRNA","processed_transcript","misc_RNA","rRNA",
                "sense_overlapping","sense_intronic","polymorphic_pseudogene",
                "non_coding","3prime_overlapping_ncrna","IG_C_gene",
                "IG_J_gene","IG_D_gene","IG_V_gene","ncrna_host"))
        },
        mm10 = {
            return(c("pseudogene","snRNA","protein_coding","antisense","miRNA",
                "snoRNA","lincRNA","processed_transcript","misc_RNA","rRNA",
                "sense_intronic","sense_overlapping","polymorphic_pseudogene",
                "IG_C_gene","IG_J_gene","IG_D_gene","IG_LV_gene","IG_V_gene",
                "IG_V_pseudogene","TR_V_gene","TR_V_pseudogene",
                "3prime_overlapping_ncrna"))
        },
        dm3 = {
            return(c("protein_coding","ncRNA","snoRNA","pre_miRNA","pseudogene",
                "snRNA","tRNA","rRNA"))
        },
        dm6 = {
            return(c("protein_coding","ncRNA","snoRNA","pre_miRNA","pseudogene",
                "snRNA","tRNA","rRNA"))
        },
        rn5 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","misc_RNA"))
        },
        rn6 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","misc_RNA"))
        },
        danrer7 = {
            return(c("antisense","protein_coding","miRNA","snoRNA","rRNA",
                "lincRNA","processed_transcript","snRNA","pseudogene",
                "sense_intronic","misc_RNA","polymorphic_pseudogene",
                "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                "non_coding","sense_overlapping"
            ))
        },
        danrer10 = {
            return(c("antisense","protein_coding","miRNA","snoRNA","rRNA",
                "lincRNA","processed_transcript","snRNA","pseudogene",
                "sense_intronic","misc_RNA","polymorphic_pseudogene",
                "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                "non_coding","sense_overlapping"
            ))
        },
        danrer11 = {
            return(c("antisense","protein_coding","miRNA","snoRNA","rRNA",
                "lincRNA","processed_transcript","snRNA","pseudogene",
                "sense_intronic","misc_RNA","polymorphic_pseudogene",
                "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                "non_coding","sense_overlapping"
            ))
        },
        pantro4 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","snRNA","snoRNA","misc_RNA"))
        },
        pantro5 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","snRNA","snoRNA","misc_RNA"))
        },
        pantro6 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","snRNA","snoRNA","misc_RNA"))
        },
        susscr3 = {
            return(c("antisense","protein_coding","lincRNA","pseudogene",
                "processed_transcript","miRNA","rRNA","snRNA","snoRNA",
                "misc_RNA","non_coding","IG_C_gene","IG_J_gene",
                "IG_V_gene","IG_V_pseudogene"))
        },
        equcab2 = {
            return(c("miRNA","misc_RNA","protein_coding","pseudogene","rRNA",
                "processed_pseudogene","snoRNA","snRNA"))
        },
        equcab3 = {
            return(c("miRNA","misc_RNA","protein_coding","pseudogene","rRNA",
                "processed_pseudogene","snoRNA","snRNA"))
        },
        tair10 = {
            return(c("miRNA","ncRNA","protein_coding","pseudogene","rRNA",
                "snoRNA","snRNA","transposable_element","tRNA"))
        }
    )
}

\name{getsetDbPath}
\alias{setDbPath}
\alias{getDbPath}
\title{Get and set sitadela default database path}
\usage{
    setDbPath(db = NULL)
    getDbPath()
}
\arguments{
    \item{db}{path to a valid SQLite database file.}
}
\value{
    This function does not have a return value.
}
\description{
    The \code{setDbPath} and \code{getDbPath} functions are 
    used to set and get the path to a sitadela annotation
    database. If not explicitly provided, it defaults to
    \code{file.path(system.file(package="sitadela"),"annotation.sqlite")}.
    Essentially, the setter function adds an option to the R
    environment pointing to the desired path.
}
\examples{
myPath <- "/home/me/test.sqlite"
setDbPath(myPath)
getDbPath()
}
\author{
    Panagiotis Moulos
}

\name{loadAnnotation}
\alias{loadAnnotation}
\title{Load a sitadela simple annotation element}
\usage{
    loadAnnotation(genome, refdb, 
        type = c("gene", "transcript", "utr",
            "transutr", "transexon", "exon"), 
        version="auto", wtv = FALSE,
        db = getDbPath(), summarized = FALSE,
        asdf = FALSE, rc = NULL)
}
\arguments{
    \item{genome}{a sitadela supported organism or a custom 
    organism name imported by the user.}

    \item{refdb}{a sitadela supported annotation source or a 
    custom name imported by the user.}
    
    \item{type}{the transcriptional unit annotation level 
    to load. It can be one of \code{"gene"} (default),
    \code{"transcript"}, \code{"utr"}, \code{"transexon"},
    \code{"transutr"}, \code{"exon"}. See Details for 
    further explanation of each option.}
    
    \item{version}{the version of the annotation to use. 
    See Details.}
    
    \item{wtv}{load annotations with versioned genes and
    transcripts when/where available.}
    
    \item{db}{same as the \code{db} in 
    \code{\link{addAnnotation}}.}
    
    \item{summarized}{if \code{TRUE}, retrieve summarized,
    non-overlaping elements where appropriate (e.g. exons).}
    
    \item{asdf}{return the result as a \code{\link{data.frame}}
    (default \code{FALSE}).}
    
    \item{rc}{same as the \code{rc} in 
    \code{\link{addAnnotation}}.}
}
\value{
    The function returns a \code{GenomicRanges} object or a
    \code{\link{data.frame}} with the requested annotation.
}
\description{
    This function loads an annotation element from a local 
    sitadela annotation database. If the annotation is not
    found and the organism is supported, the annotation is 
    fetched and created on the fly but not imported in the 
    local database. Use \code{addAnnotation} for this
    purpose (build/update/add annotations).
}
\details{
    Regarding \code{org}, it can be, for human genomes 
    \code{"hg18"}, \code{"hg19"} or \code{"hg38"}, for mouse 
    genomes \code{"mm9"}, \code{"mm10"}, for rat genomes 
    \code{"rn5"} or \code{"rn6"}, for drosophila genome
    \code{"dm3"} or \code{"dm6"}, for zebrafish genome 
    \code{"danrer7"}, \code{"danrer10"} or \code{"danrer11"}, 
    for chimpanzee genome \code{"pantro4"}, \code{"pantro5"}, 
    for pig genome \code{"susscr3"}, \code{"susscr11"}, for 
    Arabidopsis thaliana genome \code{"tair10"} and for 
    Equus caballus genome \code{"equcab2"} and \code{"equcab3"}. 
    Finally, it can be \code{"USER_NAMED_ORG"} with a custom 
    organism which has been imported to the annotation database 
    by the user using a GTF/GFF file. For example 
    \code{org="mm10_p1"}.
    
    Regarding \code{type}, it defines the level of 
    transcriptional unit (gene, transcript, 3' UTR, exon)
    coordinates to be loaded or fetched if not present. The
    following types are supported:
    
    \itemize{
        \item \code{"gene"}: canonical gene coordinates are
        retrieved from the chosen database.
        \item \code{"transcript"}: all transcript 
        coordinates are retrieved from the chosen database.
        \item \code{"utr"}: all 3' UTR coordinates are
        retrieved from the chosen database, grouped per 
        gene.
        \item \code{"transutr"}: all 3' UTR coordinates are
        retrieved from the chosen database, grouped per \
        transcript.
        \item \code{"transexon"}: all exon coordinates are
        retrieved from the chosen database, grouped per 
        transcript.
        \item \code{"exon"}: all exon coordinates are
        retrieved from the chosen database.
    }
    
    Regarding \code{version}, this is an integer denoting the
    version of the annotation to use from the local annotation
    database or fetch on the fly. For Ensembl, it corresponds
    to Ensembl releases, while for UCSC/RefSeq, it is the
    date of creation (locally).
}
\examples{
db <- file.path(system.file(package="sitadela"),
    "annotation.sqlite")
if (file.exists(db))
    gr <- loadAnnotation(genome="hg19",refdb="ensembl",
        type="gene",db=db)
}
\author{
    Panagiotis Moulos
}

\name{getSeqInfo}
\alias{getSeqInfo}
\title{Retrieve sequence length and other information}
\usage{
    getSeqInfo(org, asSeqinfo = FALSE)
}
\arguments{
    \item{org}{a supported organism to retrieve sequence
    (aka chromosome) information for. See also 
    \code{\link{addAnnotation}} about supported organisms.}

    \item{asSeqinfo}{return a \code{Seqinfo} object or a
    \code{data.frame}.}
}
\value{
    The function returns a \code{Seqinfo} or a
    \code{data.frame} with the a subset of a 
    \code{Seqinfo} information. See also 
    \code{\link{Seqinfo}}.
}
\description{
    This function retrieves sequence (chromosome) length
    and other information for a set of reference sequences
    for a sitadela supported organism. If the organism is
    supported by the \code{\link[GenomeInfoDb]{getChromInfoFromUCSC}}
    function of the \pkg{GenomeInfoDb} package, then this function
    is used, otherwise, a directo download from the UCSC golden
    path takes place to retrieve the required data.
}
\examples{
require(GenomeInfoDb)
s <- getSeqInfo("mm10")
}
\author{
    Panagiotis Moulos
}

\name{getAnnotation}
\alias{getAnnotation}
\title{Annotation downloader}
\usage{
    getAnnotation(org, type, refdb = "ensembl", ver = NULL,
        tv = FALSE, rc = NULL)
}
\arguments{
    \item{org}{the organism for which to download
    annotation (one of the supported ones, see Details).}

    \item{type}{the transcriptional unit annotation level 
    to load. It can be one of \code{"gene"} (default),
    \code{"transcript"}, \code{"utr"}, \code{"transexon"},
    \code{"transutr"}, \code{"exon"}. See Details for 
    further explanation of each option.}

    \item{refdb}{the online source to use to fetch 
    annotation. It can be \code{"ensembl"} (default), 
    \code{"ucsc"}, \code{"refseq"} or \code{"ncbi"}. In 
    the later three cases, an SQL connection is opened 
    with the UCSC public databases.}
    
    \item{ver}{the version of the annotation to use.}
    
    \item{tv}{attach or not gene/transcript version to 
    gene/transcript name. Defaults to \code{FALSE}.}

    \item{rc}{Fraction of cores to use. Same as the 
    \code{rc} in \code{\link{addAnnotation}}.}
}
\value{
    A data frame with the canonical genes, transcripts, 
    exons or 3' UTRs of the requested organism. When
    \code{type="genes"}, the data frame has the following
    columns: chromosome, start, end, gene_id, gc_content,
    strand, gene_name, biotype. When \code{type="exon"} and
    \code{type="transexon"} the data frame has the following 
    columns: chromosome, start, end, exon_id, gene_id, strand, 
    gene_name, biotype. When \code{type="utr"} or  
    \code{type="transutr"}, the data frame has the following 
    columns: chromosome, start, end, transcript_id, gene_id, 
    strand, gene_name, biotype. The latter applies to when 
    \code{type="transcript"}. The gene_id and exon_id 
    correspond to \code{type="transcript"} Ensembl, UCSC 
    or RefSeq gene, transcript and exon accessions 
    respectively. The gene_name corresponds to HUGO 
    nomenclature gene names.
}
\description{
    For Ensembl based annotations, this function connects to the 
    EBI's Biomart service using the package biomaRt and downloads 
    annotation elements (gene co-ordinates, exon co-ordinates, 
    gene identifications, biotypes etc.) for each of the supported
    organisms. For UCSC/RefSeq annotations, it connects to the 
    respective UCSC SQL databases if the package \code{RMySQL} is 
    present, otherwise it downloads flat files and build a 
    temporary SQLite database to make the necessary build 
    queries. Gene and transcript versions can be attached (when
    available) using the \code{tv} argument. This is very useful
    when transcript versioning is required, such as several
    precision medicine applications.
}
\details{
    Regarding \code{org}, it can be, for human genomes 
    \code{"hg18"}, \code{"hg19"} or \code{"hg38"}, for mouse 
    genomes \code{"mm9"}, \code{"mm10"}, for rat genomes 
    \code{"rn5"} or \code{"rn6"}, for drosophila genome
    \code{"dm3"} or \code{"dm6"}, for zebrafish genome 
    \code{"danrer7"}, \code{"danrer10"} or \code{"danrer11"}, 
    for chimpanzee genome \code{"pantro4"}, \code{"pantro5"}, 
    for pig genome \code{"susscr3"}, \code{"susscr11"}, for 
    Arabidopsis thaliana genome \code{"tair10"} and for 
    Equus caballus genome \code{"equcab2"} and \code{"equcab3"}. 
    Finally, it can be \code{"USER_NAMED_ORG"} with a custom 
    organism which has been imported to the annotation database 
    by the user using a GTF/GFF file. For example 
    \code{org="mm10_p1"}.
    
    Regarding \code{type}, it defines the level of 
    transcriptional unit (gene, transcript, 3' UTR, exon)
    coordinates to be loaded or fetched if not present. The
    following types are supported:
    
    \itemize{
        \item \code{"gene"}: canonical gene coordinates are
        retrieved from the chosen database.
        \item \code{"transcript"}: all transcript 
        coordinates are retrieved from the chosen database.
        \item \code{"utr"}: all 3' UTR coordinates are
        retrieved from the chosen database, grouped per 
        gene.
        \item \code{"transutr"}: all 3' UTR coordinates are
        retrieved from the chosen database, grouped per \
        transcript.
        \item \code{"transexon"}: all exon coordinates are
        retrieved from the chosen database, grouped per 
        transcript.
        \item \code{"exon"}: all exon coordinates are
        retrieved from the chosen database.
    }
}

\note{
    The data frame that is returned contains only "canonical"
    chromosomes for each organism. It does not contain
    haplotypes or non-anchored sequences and does not contain
    mitochondrial chromosomes.
}
\examples{
mm10Genes <- getAnnotation("mm10","gene")
}
\author{
    Panagiotis Moulos
}

\name{importCustomAnnotation}
\alias{importCustomAnnotation}
\title{Import a metaseqR2 custom annotation element}
\usage{
    importCustomAnnotation(gtfFile, metadata,
        type = c("gene", "transcript", "utr",
            "transexon", "transutr", "exon"))
}
\arguments{
    \item{gtfFile}{a GTF file containing the gene structure
    of the organism to be imported.}

    \item{metadata}{a list with additional information about
    the annotation to be imported. The same as in the
    \code{\link{addCustomAnnotation}} man page.}
    
    \item{type}{same as the \code{type} in 
    \code{\link{loadAnnotation}}.}
}
\value{
    The function returns a \code{GenomicRanges} object with
    the requested annotation.
}
\description{
    This function imports a custom GTF/GFF file in a manner
    helpful for the addition of custom annotations to sitadela.
}
\examples{
# Dummy GTF as example
chromInfo <- data.frame(length=c(1000L,2000L,1500L),
    row.names=c("A","B","C"))

# Build with the metadata list filled (you can also provide a version)
if (.Platform$OS.type == "unix" && !grepl("^darwin",R.version$os)) {
    myGenes <- importCustomAnnotation(
        gtfFile=file.path(system.file(package="sitadela"),
            "dummy.gtf.gz"),
        metadata=list(
            organism="dummy",
            source="dummy_db",
            version=1,
            chromInfo=chromInfo
        ),
        type="gene"
    )
}

## Real data!
## Gene annotation dump from Ensembl
#download.file(paste0("ftp://ftp.ensembl.org/pub/release-98/gtf/",
#  "dasypus_novemcinctus/Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#  file.path(tempdir(),"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"))

## Build with the metadata list filled (you can also provide a version)
#dasGenes <- importCustomAnnotation(
#  gtfFile=file.path(tempdir(),"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#  metadata=list(
#    organism="dasNov3_test",
#    source="ensembl_test"
#  ),
#  type="gene"
#)
}
\author{
    Panagiotis Moulos
}

\name{addAnnotation}
\alias{addAnnotation}
\title{Build a local genomic regions annotation database}
\usage{
    addAnnotation(organisms, sources, db = getDbPath(),
        versioned = FALSE, forceDownload = TRUE, retries = 5,
        rc = NULL, stopIfNotBS = FALSE)
}
\arguments{
    \item{organisms}{a list of organisms and versions
    for which to download and build annotations. See 
    also Details.}

    \item{sources}{a character vector of public sources
    from which to download and build annotations. It can be
    one or more of \code{"ensembl"}, \code{"ucsc"}, 
    \code{"refseq"} or \code{"ncbi"}. See also Details.}
    
    \item{db}{a valid path (accessible at least by the
    current user) where the annotation database will be 
    set up. It defaults to 
    \code{system.file(package = "sitadela"),}
    \code{"annotation.sqlite")} that is, the installation
    path of sitadela package.}
    
    \item{versioned}{create an annotation database with
    versioned genes and transcripts, when possible.}
    
    \item{forceDownload}{by default, 
    \code{addAnnotation} will not download an
    existing annotation again (\code{FALSE}). Set to 
    \code{TRUE} if you wish to update the annotation 
    database for a particular version.}
    
    \item{retries}{how many times should the annotation
    worker try to re-connect to internet resources in case
    of a connection problem or failure.}

    \item{rc}{fraction (0-1) of cores to use in a multicore 
    system. It defaults to \code{NULL} (no parallelization).
    Sometimes used for building certain annotation types.}
    
    \item{stopIfNotBS}{stop or warn (default) if certain 
    \code{BSgenome} packages are not present. See also
    Details.}
}
\value{
    The function does not return anything. Only the SQLite 
    database is created or updated.
}
\description{
    This function is the main annotation database creator 
    of sitadela. It creates a local SQLite database for
    various organisms and categories of genomic regions.
    Annotations are retrieved in simple, tab-delimited or
    GRanges formats.
}
\details{
    Regarding the \code{organisms} argument, it is a list
    with specific format which instructs 
    \code{addAnnotation} on which organisms and
    versions to download from the respective sources. Such
    a list may have the format: 
    \code{organisms=list(hg19=75, mm9=67, mm10=96:97)}
    This is explained as follows:
    \itemize{
        \item A database comprising the human genome versions
        \code{hg19} and the mouse genome versions 
        \code{mm9, mm10} will be constructed.
        \item If \code{"ensembl"} is in \code{sources}, 
        version 75 is downloaded for \code{hg19} and versions 
        \code{67, 96, 97} for \code{mm9, mm10}. 
        \item If \code{"ucsc"} or \code{"refseq"} are in 
        \code{sources}, the latest versions are downloaded
        and marked by the download date. As UCSC and RefSeq
        versions are not accessible in the same way as
        Ensembl, this procedure cannot always be replicated.
    }
    \code{organisms} can also be a character vector with organism
    names/versions (e.g. \code{organisms = c("mm10","hg19")}),
    then the latest versions are downloaded in the case of 
    Ensembl.
    
    The supported supported organsisms are, for human genomes 
    \code{"hg18"}, \code{"hg19"} or \code{"hg38"}, for mouse 
    genomes \code{"mm9"}, \code{"mm10"}, for rat genomes 
    \code{"rn5"} or \code{"rn6"}, for drosophila genome
    \code{"dm3"} or \code{"dm6"}, for zebrafish genome 
    \code{"danrer7"}, \code{"danrer10"} or \code{"danrer11"}, 
    for chimpanzee genome \code{"pantro4"}, \code{"pantro5"}, 
    for pig genome \code{"susscr3"}, \code{"susscr11"}, for 
    Arabidopsis thaliana genome \code{"tair10"} and for 
    Equus caballus genome \code{"equcab2"} and \code{"equcab3"}. 
    Finally, it can be \code{"USER_NAMED_ORG"} with a custom 
    organism which has been imported to the annotation database 
    by the user using a GTF/GFF file. For example 
    \code{org="mm10_p1"}.
    
    Regarding \code{sources}, \code{"ucsc"} corresponds to 
    UCSC Genome Browser annotated transcripts, \code{"refseq"}
    corresponds to UCSC RefSeq maintained transcripts while 
    \code{"ncbi"} corresponds to NCBI RefSeq annotated and 
    maintained transcripts. UCSC, RefSeq and NCBI annotations 
    are constructed by querying the UCSC Genome Browser 
    database.
    
    Regarding \code{stopIfNotBS}, when \code{sources} 
    includes \code{"ucsc"}, \code{"refseq"} or \code{"ncbi"},
    the GC content of a gene is not available as a database
    attribute as with Ensembl and has to be calculated if to
    be included in the respective annotation. For this reason,
    sitadela uses `BSgenome` packages. If 
    \code{stopIfNotBS=FALSE} (default), then the annotation
    building continues and GC content is \code{NA} for the
    missing `BSgenome` packages.If \code{stopIfNotBS=FALSE},
    then building stops until all the required packages for
    the selected organisms become available (installed by
    the user).
}
\examples{
# Build a test database with one genome
myDb <- file.path(tempdir(),"testann.sqlite")

organisms <- list(mm10=100)
sources <- "ensembl"

# If the example is not running in a multicore system, rc is ignored
#addAnnotation(organisms,sources,db=myDb,rc=0.5)

# A more complete case, don't run as example
# Since we are using Ensembl, we can also ask for a version
#organisms <- list(
#    mm9=67,
#    mm10=96:97,
#    hg19=75,
#    hg38=96:97
#)
#sources <- c("ensembl", "refseq")

## Build on the default location (depending on package location, it may
## require root/sudo)
#addAnnotation(organisms,sources)

## Build on an alternative location
#myDb <- file.path(path.expand("~"),"my_ann.sqlite")
#addAnnotation(organisms,sources,db=myDb)
}
\author{
    Panagiotis Moulos
}

\name{removeAnnotation}
\alias{removeAnnotation}
\title{Remove an annotation from a sitadela database}
\usage{
    removeAnnotation(org, refdb, ver = NULL, db = NULL)
}
\arguments{
    \item{org}{an existing organism to remove from the
    database. See also \code{\link{addAnnotation}} and
    \code{\link{addCustomAnnotation}} for details.}

    \item{refdb}{an existing annotation source to remove 
    from the database. See also \code{\link{addAnnotation}} 
    and \code{\link{addCustomAnnotation}} for details.}
    
    \item{ver}{an existing annotation version to remove 
    from the database. See also \code{\link{addAnnotation}} 
    and \code{\link{addCustomAnnotation}} for details. If
    \code{NULL} (default), all versions corresponding to
    \code{org} and \code{refdb} will be removed.}
    
    \item{db}{the database to remove from, defaults to 
    \code{getDbPath()}.}
}
\value{
    The function return the number of rows removed from
    the database contents table.
}
\description{
    This function removes a specific annotation from a 
    sitadela database. It does not support multiple
    organism, resource and version removal for now.
}
\examples{
# Dummy database as example
customDir <- file.path(tempdir(),"test_remove")
dir.create(customDir)

myDb <- file.path(customDir,"testann.sqlite")
chromInfo <- data.frame(length=c(1000L,2000L,1500L),
    row.names=c("A","B","C"))

# Build with the metadata list filled (you can also provide a version)
if (.Platform$OS.type == "unix") {
    addCustomAnnotation(
        gtfFile=file.path(system.file(package="sitadela"),
            "dummy.gtf.gz"),
        metadata=list(
            organism="dummy",
            source="dummy_db",
            version=1,
            chromInfo=chromInfo
        ),
        db=myDb
    )

    # Try to retrieve some data
    myGenes <- loadAnnotation(genome="dummy",refdb="dummy_db",
        type="gene",db=myDb)
    myGenes
    
    # Now remove
    n <- removeAnnotation("dummy","dummy_db",1,myDb)
}
}
\author{
    Panagiotis Moulos
}

\name{testFuns}
\alias{testEnsembl}
\alias{testEnsemblSimple}
\alias{testUcsc}
\alias{testUcscAll}
\alias{testUcscUtr}
\alias{testUcscUtrAll}
\alias{testCustomGtf}
\alias{testKnownBuild}
\alias{testCustomBuild}
\title{Query and database build testing functions}
\usage{
    testEnsembl(level = c("normal", "long", "short"),
        versioned = FALSE)
    testEnsemblSimple(orgs, types, versioned = FALSE)

    testUcsc(orgs, refdbs, types, versioned = FALSE)
    testUcscAll()
    
    testUcscUtr(orgs, refdbs, versioned = FALSE)
    testUcscUtrAll()
    
    testCustomGtf(gtf)
    
    testKnownBuild(org, refdb, ver = NULL, tv = FALSE)
    testCustomBuild(gtf, metadata)
    
}
\arguments{
    \item{level}{how many Ensembl versions from the 
    supported organisms should be checked. It can be 
    \code{"normal"} (default), \code{"long"} or
    \code{"short"}. See also Details.}
    
    \item{orgs}{a vector of sitadela supported organisms.
    See also \code{\link{addAnnotation}}.}
    
    \item{refdbs}{a vector of sitadela supported annotation.
    sources. See also \code{\link{addAnnotation}}.}
    
    \item{versioned}{use versioned genes/transcripts where
    available.}
    
    \item{types}{a vector of sitadela annotation types.
    See also \code{\link{loadAnnotation}}.}
    
    \item{org}{as \code{orgs} above but only one organism.}
    
    \item{refdb}{as \code{refdbs} above but only one source.}
    
    \item{ver}{specific annotation version, see also 
    \code{\link{addAnnotation}}.}
    
    \item{tv}{retrieve versioned genes and transcripts when
    possible, see also \code{\link{addAnnotation}}.}
    
    \item{gtf}{a valid GTF or GFF file.}
    
    \item{metadata}{additional information on the contents
    of GTF/GFF file. See also 
    \code{\link{addCustomAnnotation}}.}
}
\value{
    This group of functions return either a vector of
    logical values showing success or failure of conducted
    tests, or a list of test failure reasons or \code{NULL}
    if all tests are successful. Specifically, 
    \code{testKnownBuild} and \code{testCustomBuild}) return
    logicals while all the rest return \code{NULL} if tests
    are successful or a list of failure reasons (and the
    respective test) otherwise.
}
\description{
    This group of testing fuctions can be used to test the
    entirety of sitadela annotation building capabilities
    from known resources or custom GTF/GFF files. They are
    useful for testing the particular annotation the user
    wishes to build prior to building the final database, 
    in order to avoid failures during the longer build. In
    all cases, useful messages are also displayed.
}
\details{
    Regarding \code{testEnsembl} and its arguments, when
    \code{level="normal"}, only the last one or two 
    (depending on availability with Biomart) supported 
    Ensembl versions are checked for fetching availability.
    If \code{level="long"}, all available versions are
    checked for fetching availability (use with care, it
    can run for some time!). If \code{level="short"}, only
    the last version of each supported organism is checked.
    Simpler tests with Ensembl (single organisms, types)
    can be performed with \code{testEnsemblSimple}. It will
    use only the latest version for the asked organism(s).
    
    Regarding \code{testUcsc}, it can be used to test the
    queries used with the UCSC databases for a given 
    organism and database. \code{testUcscAll} will test
    queries for all supported organisms and databases and
    may take a while to finish.
    
    Similarly, \code{testUcscUtr} and \code{testUcscUtrAll}
    will test the queries and building of 3' UTR regions
    form UCSC databases. 3' UTR constructing is not part
    of the other UCSC testing functions as the process is
    different and may be tested only in Unix/Linux machines.
    
    The function \code{testCustomGtf} will simply test
    whether the provided GTF/GFF file can be parsed and 
    used to extract the sitadela annotation types. If this
    is not possible (rarely), this test will fail. If you
    wish to test complete database building with a custom
    GTF/GFF file, use \code{testCustomBuild}.
    
    Finally, \code{testKnownBuild} will test database
    building and querying (add/remove annotation) for a
    single organism.
}
\examples{
    # Test a dummy GTF file
    gtf <- file.path(system.file(package="sitadela"),
        "dummy.gtf.gz")
    chromInfo <- data.frame(length=c(1000L,2000L,1500L),
        row.names=c("A","B","C"))
    metadata=list(
        organism="dummy",
        source="dummy_db",
        version=1,
        chromInfo=chromInfo
    )
    
    testResult <- testCustomBuild(gtf,metadata)
    # For this case, just testResult <- testCustomBuild()
    # would also work
    
    # More real tests
    if (require(RMySQL))
        f <- testUcsc("hg19","refseq","gene",TRUE)
    
    # Test a complete build for Ensembl mm9
    # testResult <- testKnownBuild()
    
    # Test a complete build for UCSC dm6
    # testResult <- testKnownBuild("dm6","ucsc")
}
\author{
    Panagiotis Moulos
}

\name{getUserAnnotations}
\alias{getUserAnnotations}
\title{List installed custom user-defined sitadela annotations}
\usage{
    getUserAnnotations(obj = NULL)
}
\arguments{
    \item{obj}{\code{NULL} or the path to a sitadela
    SQLite annotation database. If \code{NULL}, the
    function will try to guess the location of the 
    SQLite database.}
}
\value{
    The function returns a \code{data.frame} object with
    the installed, custom, user-defined local annotations
    only.
}
\description{
    This function returns a data frame with information
    on locally installed, custom user-defined annotations
    only. For a list of all annotations, see
    \code{\link{getInstalledAnnotations}}.
}
\examples{
db <- file.path(system.file(package="sitadela"),
    "annotation.sqlite")
if (file.exists(db))
    u <- getUserAnnotations(obj=db)
}
\author{
    Panagiotis Moulos
}
